/*
 * PCF8563 RTC
 *
 * From Phillips' datasheet:
 *
 * The PCF8563 is a CMOS real-time clock/calendar optimized for low power
 * consumption. A programmable clock output, interupt output and voltage
 * low detector are also provided. All address and data are transferred
 * serially via two-line bidirectional I2C-bus. Maximum bus speed is
 * 400 kbits/s. The built-in word address register is incremented
 * automatically after each written or read byte.
 *
 * Copyright (c) 2002-2003, Axis Communications AB
 * All rights reserved.
 *
 * Author: Tobias Anderberg <tobiasa@axis.com>.
 *
 */
 
//
// History :
//   V1.00  1/21/2005  Hank Chu 
//   1) Copied from uClinux's linux kernel v2.4.27 source codes :
//        arch/cris/drivers/pcf8563.c
//
//      And modified to meet Motorola PowerPC MPC8241 (Linux 2.4.21) platform.
//      Add a header file "ntx_rtc.h" for PCF8563.
//
//   2) It does not support the RTC calibration now.
//      Please refer to the pcf8563_calibrate_decr() routine for details.
//
//
//
//

#include <linux/config.h>
#include <linux/types.h>
#include <linux/sched.h>
#include <linux/init.h>
#include <linux/rtc.h>
#include <linux/bcd.h>

// 01/22/2005 Hank Chu : 
//  Because we use PCF8563 RTC chip for clock, we cannot
//  include the <asm/time.h> header file any more.
//  The asm/time.h (asm-ppc/time.h) will include <asm-ppc/mc146818rtc.h> file, and 
//  it define the CMOS_READ() and CMOS_WRITE() and etc. routines which are not usable now.
//  
// #include <asm/time.h>

#include "i2c_export.h"
#include "i2c.h"
#include "rtc_def.h"


#define PCF8563_MAJOR	121	/* Local major number. */
#define DEVICE_NAME	"rtc"	/* Name which is registered in /proc/devices. */
#define PCF8563_NAME	"PCF8563"
#define DRIVER_VERSION	"$Revision: 1.15 $"

//#ifdef M_NH230

  #define SLAVE_ADDRESS_OF_PCF8563    0x51
  
  #define M_MAX_I2C_LOOP   0x064
  
  
  rwlock_t pcf8563_lock;
  
  void to_tm(int tim, struct rtc_time * tm);
  long __init  pcf8563_rtc_init(void);
  u8 i2c_readreg(u8 c_slave_addr, u8 c_reg_addr);  
  int i2c_writereg(u8 c_slave_addr, u8 c_reg_addr, u8 c_value);
  
  //
  // Subject :
  //   The API routine to read 1-byte slave register of I2C bus.
  // Input :
  //   slave_addr : I2C bus slave address
  //   reg_addr   : register address of I2C slave device
  // Output :
  //   8-bit value of register
  //
  u8 i2c_readreg(u8 c_slave_addr, u8 c_reg_addr)
  {
    u8 c_value = 0;                    // 1-byte value to write to I2C slave's register
    u32 dw_loop = 0;                   // loop to access I2C slave until OK
    I2C_Status dw_status = 0;          // status of I2C slave access
    
    // 1) Loop to read access the I2C slave
    do {
      read_lock_irq(&pcf8563_lock);    // lock it
      
      // 1-1) Do I2C bus transaction to read register of slave, polling mode, retry 5 times.
      dw_status = I2C_do_transaction(0, I2C_MASTER_RCV, c_slave_addr, c_reg_addr, 1, &c_value, I2C_STOP, 5, 0);
      read_unlock_irq(&pcf8563_lock);  // unlock it
      dw_loop++;
    } while (dw_status != I2C_SUCCESS && dw_loop < M_MAX_I2C_LOOP);
    
    // 2) Print error message if failed
    if (dw_status != I2C_SUCCESS)
      printk("%s: error: failed to read I2C slave address=0x%X, register address=0x%X\n", 
               __FUNCTION__, c_slave_addr, c_reg_addr);
      
    return (c_value);
  }


  //
  // Subject :
  //   The API routine to write 1-byte slave register of I2C bus.
  // Input :
  //   slave_addr : I2C bus slave address
  //   reg_addr   : register address of I2C slave device
  //   c_value    : 1-byte value to write to slave register
  // Output :
  //   status of I2C access
  //  
  int i2c_writereg(u8 c_slave_addr, u8 c_reg_addr, u8 c_value)
  {
    u32 dw_loop = 0;                   // loop to access I2C slave until OK
    I2C_STATUS dw_status = 0;          // status of I2C slave access
    
    // 1) Loop to read access the I2C slave
    do {
      write_lock_irq(&pcf8563_lock);   // lock it
      
      // 1-1) Do I2C bus transaction to write register of slave, polling mode, retry 5 times.
      dw_status = I2C_do_transaction(0, I2C_MASTER_XMIT, c_slave_addr, c_reg_addr, 1, &c_value, I2C_STOP, 5, 0);
      write_unlock_irq(&pcf8563_lock); // unlock it
      dw_loop++;
    } while (dw_status != I2C_SUCCESS && dw_loop < M_MAX_I2C_LOOP);

    // 2) Print error message if failed
    if (dw_status != I2C_SUCCESS)
      printk("%s: error: failed to write I2C slave address=0x%X, register address=0x%X, value=0x%X\n", 
               __FUNCTION__, c_slave_addr, c_reg_addr, c_value);
    
    return ((int)dw_status);
  }
//#endif  // end of #ifdef {M_NH230}


/* Two simple wrapper macros, saves a few keystrokes. */
#define rtc_read(x) i2c_readreg(RTC_I2C_READ, x)
#define rtc_write(x,y) i2c_writereg(RTC_I2C_WRITE, x, y)

#if 0
unsigned char
pcf8563_readreg(int reg)
{
	unsigned char res = rtc_read(reg);

	/* The PCF8563 does not return 0 for unimplemented bits */
	switch (reg) {
		case RTC_SECONDS_REG:
		case RTC_MINUTES_REG:
			res &= 0x7F;
			break;
		case RTC_HOURS_REG:
		case RTC_DAY_OF_MONTH_REG:
			res &= 0x3F;
			break;
		case RTC_WEEKDAY_REG:
			res &= 0x07;
			break;
		case RTC_MONTH_REG:
			res &= 0x1F;
			break;
		case RTC_CONTROL1:
			res &= 0xA8;
			break;
		case RTC_CONTROL2:
			res &= 0x1F;
			break;
		case RTC_CLOCKOUT_FREQ:
		case RTC_TIMER_CONTROL:
			res &= 0x83;
			break;
	}
	return res;
}

void
pcf8563_writereg(int reg, unsigned char val)
{
#ifdef CONFIG_SANDPOINT_I2C_RTC_READONLY
	if (reg == RTC_CONTROL1 || (reg >= RTC_SECONDS_REG && reg <= RTC_YEAR_REG))
		return;
#endif

	rtc_write(reg, val);
}
#endif


u8 API_pcf8563_get_RTC_SECONDS_REG(void)
{
  u8 c_seconds = rtc_read(RTC_SECONDS_REG);
  
#ifndef M_NH230  // marked by ansel, we don't need this info
  if (c_seconds & 0x80) {              // bit[7] : 1 if low voltage ?
    printk(KERN_WARNING "%s: RTC Voltage Low - reliable date/time "
	    "information is no longer guaranteed!\n", PCF8563_NAME);    
  }
#endif
  BCD_TO_BIN(c_seconds);               // Convert it
    
  return (c_seconds);
}


int API_pcf8563_get_rtc_time(struct rtc_time *lpst_time)
{
  if (lpst_time == NULL)
    return (-1);           // NULL pointer found
	
  lpst_time->tm_sec  = rtc_read(RTC_SECONDS_REG);
  lpst_time->tm_min  = rtc_read(RTC_MINUTES_REG);
  lpst_time->tm_hour = rtc_read(RTC_HOURS_REG);
  lpst_time->tm_mday = rtc_read(RTC_DAY_OF_MONTH_REG);
  lpst_time->tm_wday = rtc_read(RTC_WEEKDAY_REG);
  lpst_time->tm_mon  = rtc_read(RTC_MONTH_REG);
  lpst_time->tm_year = rtc_read(RTC_YEAR_REG);
#ifndef M_NH230
  if (lpst_time->tm_sec & 0x80)     // bit[7]: 1 if low voltage ?
    printk(KERN_WARNING "%s: RTC Voltage Low - reliable date/time "
	    "information is no longer guaranteed!\n", PCF8563_NAME);
#endif
#ifdef M_NH230
//ansel 20050325 century bit = 1: 1900, 0: 2000
  lpst_time->tm_year = BCD_TO_BIN(lpst_time->tm_year) + ((lpst_time->tm_mon & 0x80) ? 1900 : 2000);
#else
  lpst_time->tm_year = BCD_TO_BIN(lpst_time->tm_year) + ((lpst_time->tm_mon & 0x80) ? 100 : 0);
#endif

  lpst_time->tm_sec  &= 0x7F;
  lpst_time->tm_min  &= 0x7F;
  lpst_time->tm_hour &= 0x3F;
  lpst_time->tm_mday &= 0x3F;
  lpst_time->tm_wday &= 0x07; /* Not coded in BCD. */
  lpst_time->tm_mon  &= 0x1F;

  BCD_TO_BIN(lpst_time->tm_sec);
  BCD_TO_BIN(lpst_time->tm_min);
  BCD_TO_BIN(lpst_time->tm_hour);
  BCD_TO_BIN(lpst_time->tm_mday);
  BCD_TO_BIN(lpst_time->tm_mon);

#ifndef M_NH230
//ansel 20050325 this is not right...
  lpst_time->tm_mon--; /* Month is 1..12 in RTC but 0..11 in linux */
#endif

  return (0);                               // OK
}


unsigned long pcf8563_get_rtc_time(void)
{
  struct rtc_time st_time;
  //memset(&tm, NULL, sizeof(tm));
	
  if (API_pcf8563_get_rtc_time(&st_time))   // failed to get RTC ?
    return (0);                             // return as error
	
  // Convert to total seconds
  unsigned long dw_total_seconds = 0;
  dw_total_seconds = mktime(st_time.tm_year, st_time.tm_mon, st_time.tm_mday, 
                            st_time.tm_hour, st_time.tm_min, st_time.tm_sec);
  return (dw_total_seconds);
}


int pcf8563_get_rtc_time_struct(struct rtc_time *lpst_time)
{
  return (API_pcf8563_get_rtc_time(lpst_time));
}



int pcf8563_set_rtc_time_struct(struct rtc_time *lpst_time)
{
#ifdef CONFIG_SANDPOINT_I2C_RTC_READONLY
  return -EPERM;

#else
  if (lpst_time == NULL)
    return (-EPERM);
    
  rtc_write(RTC_YEAR_REG, lpst_time->tm_year);
  rtc_write(RTC_MONTH_REG, lpst_time->tm_mon);
  rtc_write(RTC_WEEKDAY_REG, lpst_time->tm_wday); /* Not coded in BCD. */
  rtc_write(RTC_DAY_OF_MONTH_REG, lpst_time->tm_mday);
  rtc_write(RTC_HOURS_REG, lpst_time->tm_hour);
  rtc_write(RTC_MINUTES_REG, lpst_time->tm_min);
  rtc_write(RTC_SECONDS_REG, lpst_time->tm_sec);
  
  return 0;
#endif /* !CONFIG_SANDPOINT_I2C_RTC_READONLY */
}


//
// Subject :
//   Check whether the PCF8563 is under low voltage (clock information is not guaranteed) ?
// Output :
//   1 : if PCF8563 is under low voltage.
//   0 : not
//
int is_pcf8563_low_voltage(void)
{
 return (rtc_read(RTC_SECONDS_REG) & 0x80);
}


//
// Subject :
//   Check whether the PCF8563 is under low voltage (clock information is not guaranteed) ?
// Output :
//   1 : failed to write to RTC device's register
//   0 : OK
//
int pcf8563_reset_low_voltage(void)
{
  // Clear the VL bit in the seconds register
  int dw_value = rtc_read(RTC_SECONDS_REG);
  
  int dw_result = rtc_write(RTC_SECONDS_REG, (dw_value & 0x7F));   // Clean bit[7]
  return (dw_result == I2C_SUCCESS ? 0 : 1);
}


int pcf8563_set_rtc_time(unsigned long nowtime)
{
#ifdef CONFIG_SANDPOINT_I2C_RTC_READONLY
			return -EPERM;
#endif /* !CONFIG_SANDPOINT_I2C_RTC_READONLY */

  struct rtc_time tm;

  to_tm(nowtime, &tm);

  tm.tm_year = (tm.tm_year - 1900) % 100;

#if 0
  // 01/22/2005  Hank Chu :
  //  Should we use the following codes to convert the leap year ????
  //

  /* Convert from struct tm to struct rtc_time. */
  tm.tm_year += 1900;
  tm.tm_mon += 1;

  /*
   * Check if tm.tm_year is a leap year. A year is a leap
   * year if it is divisible by 4 but not 100, except
   * that years divisible by 400 _are_ leap years.
   */
  year = tm.tm_year;
  leap = (tm.tm_mon == 2) && ((year % 4 == 0 && year % 100 != 0) || year % 400 == 0);

  /* Perform some sanity checks. */
  if ((tm.tm_year < 1970) || (tm.tm_mon > 12) || (tm.tm_mday == 0) ||
      (tm.tm_mday > days_in_month[tm.tm_mon] + leap) ||
      (tm.tm_wday >= 7) || (tm.tm_hour >= 24) || (tm.tm_min >= 60) || (tm.tm_sec >= 60))
    return -EINVAL;

  century = (tm.tm_year >= 2000) ? 0x80 : 0;
  tm.tm_year = tm.tm_year % 100;

  BIN_TO_BCD(tm.tm_year);
  BIN_TO_BCD(tm.tm_mday);
  BIN_TO_BCD(tm.tm_hour);
  BIN_TO_BCD(tm.tm_min);
  BIN_TO_BCD(tm.tm_sec);
  tm.tm_mon |= century;
#else  
  
  BIN_TO_BCD(tm.tm_year);
  BIN_TO_BCD(tm.tm_mday);
  BIN_TO_BCD(tm.tm_hour);
  BIN_TO_BCD(tm.tm_min);
  BIN_TO_BCD(tm.tm_sec);
  //tm.tm_mon |= century;
#endif

  rtc_write(RTC_YEAR_REG, tm.tm_year);
  rtc_write(RTC_MONTH_REG, tm.tm_mon);
  rtc_write(RTC_WEEKDAY_REG, tm.tm_wday); /* Not coded in BCD. */
  rtc_write(RTC_DAY_OF_MONTH_REG, tm.tm_mday);
  rtc_write(RTC_HOURS_REG, tm.tm_hour);
  rtc_write(RTC_MINUTES_REG, tm.tm_min);
  rtc_write(RTC_SECONDS_REG, tm.tm_sec);

  return 0;
}


void pcf8563_calibrate_decr(void)
{
  // Do nothing as we do not support the RTC calibration now.
  return;
}


long __init  pcf8563_rtc_init(void)
{
  static int dw_init_flag = 0;   // 0 if not initialized  
  
  // Do nothing if already initialized ?
  if (dw_init_flag)
    return (0);
  
  // Initiate the I2C protocol
  I2C_Initialize(SLAVE_ADDRESS_OF_PCF8563, 0, NULL);   // ???? Is Slave Address 0x51 ?
  dw_init_flag = 1;                  // set as initialized
  

  //
  // First of all we need to reset the chip. This is done by
  // clearing control1, control2 and clk freq and resetting
  // all alarms.
  //
  if (rtc_write(RTC_CONTROL1, 0x00) != I2C_SUCCESS)
    goto err;

  if (rtc_write(RTC_CONTROL2, 0x00) != I2C_SUCCESS)
    goto err;

  if (rtc_write(RTC_CLOCKOUT_FREQ, 0x00) != I2C_SUCCESS)
    goto err;

  if (rtc_write(RTC_TIMER_CONTROL, 0x03) != I2C_SUCCESS)
    goto err;

  /* Reset the alarms. */
  if (rtc_write(RTC_MINUTE_ALARM_REG, 0x80) != I2C_SUCCESS)
    goto err;

  if (rtc_write(RTC_HOUR_ALARM_REG, 0x80) != I2C_SUCCESS)
    goto err;

  if (rtc_write(RTC_DAY_ALARM_REG, 0x80) != I2C_SUCCESS)
    goto err;

  if (rtc_write(RTC_WEEKDAY_REG_ALARM_REG, 0x80) != I2C_SUCCESS)
    goto err;

  printk(KERN_INFO "%s Real-Time Clock Driver, %s\n", PCF8563_NAME, DRIVER_VERSION);

  /* Check for low voltage, and warn about it.. */
  if (rtc_read(RTC_SECONDS_REG) & 0x80)    // bit[7]: 1 if integrity of clock not guaranteed
#ifdef M_NH230  //ansel: here I mark the warning and start with 2005/01/01 12:00:00
    {
        printk(KERN_WARNING "%s: RTC Voltage Low - reliable date/time "
            "information is no longer guaranteed!\n", PCF8563_NAME);
        printk(KERN_WARNING "%s: RTC Voltage Low - try to set date/time"
            "to default 2005/01/01 12:00:00\n", PCF8563_NAME);
        // if century bit = 0 then year = 2000 + RTC_YEAR_REG
        if (rtc_write(RTC_YEAR_REG, 0x05) != I2C_SUCCESS)
            goto err;
        // set month to 1 and century bit(bit 7) to 0 for 2xxx
        if (rtc_write(RTC_MONTH_REG, 0x01) != I2C_SUCCESS)
            goto err;
        if (rtc_write(RTC_DAY_OF_MONTH_REG, 0x01) != I2C_SUCCESS)
            goto err;
        if (rtc_write(RTC_HOURS_REG, 0x12) != I2C_SUCCESS)
            goto err;
        if (rtc_write(RTC_MINUTES_REG, 0x00) != I2C_SUCCESS)
            goto err;
        // set second to 0 and clear VL bit(bit 7)
        if (rtc_write(RTC_SECONDS_REG, 0x00) != I2C_SUCCESS)
            goto err;
    }
#else
        printk(KERN_WARNING "%s: RTC Voltage Low - reliable date/time "
            "information is no longer guaranteed!\n", PCF8563_NAME);
#endif

  return 0;

err:
  printk(KERN_INFO "%s: Error initializing chip.\n", PCF8563_NAME);
  return -1;
}

