/*
 * arch/ppc/platforms/sandpoint_setup.c
 * 
 * Board setup routines for the Motorola SPS Sandpoint Test Platform.
 *
 * Author: Mark A. Greer
 *         mgreer@mvista.com
 *
 * 2000-2002 (c) MontaVista, Software, Inc.  This file is licensed under
 * the terms of the GNU General Public License version 2.  This program
 * is licensed "as is" without any warranty of any kind, whether express
 * or implied.
 */

/*
 * This file adds support for the Motorola SPS Sandpoint Test Platform.
 * These boards have a PPMC slot for the processor so any combination
 * of cpu and host bridge can be attached.  This port is for an 8240 PPMC
 * module from Motorola SPS and other closely related cpu/host bridge
 * combinations (e.g., 750/755/7400 with MPC107 host bridge).
 * The sandpoint itself has a Windbond 83c553 (PCI-ISA bridge, 2 DMA ctlrs, 2
 * cascaded 8259 interrupt ctlrs, 8254 Timer/Counter, and an IDE ctlr), a
 * National 87308 (RTC, 2 UARTs, Keyboard & mouse ctlrs, and a floppy ctlr),
 * and 4 PCI slots (only 2 of which are usable; the other 2 are keyed for 3.3V
 * but are really 5V).
 *
 * The firmware on the sandpoint is called DINK (not my acronym :).  This port
 * depends on DINK to do some basic initialization (e.g., initialize the memory
 * ctlr) and to ensure that the processor is using MAP B (CHRP map).
 *
 * The switch settings for the Sandpoint board MUST be as follows:
 * 	S3: down
 * 	S4: up
 * 	S5: up
 * 	S6: down
 *
 * 'down' is in the direction from the PCI slots towards the PPMC slot;
 * 'up' is in the direction from the PPMC slot towards the PCI slots.
 * Be careful, the way the sandpoint board is installed in XT chasses will
 * make the directions reversed.
 *
 * Since Motorola listened to our suggestions for improvement, we now have
 * the Sandpoint X3 board.  All of the PCI slots are available, it uses
 * the serial interrupt interface (just a hardware thing we need to
 * configure properly).
 *
 * Use the default X3 switch settings.  The interrupts are then:
 *		EPIC	Source
 *		  0	SIOINT 		(8259, active low)
 *		  1	PCI #1
 *		  2	PCI #2
 *		  3	PCI #3
 *		  4	PCI #4
 *		  7	Winbond INTC	(IDE interrupt)
 *		  8	Winbond INTD	(IDE interrupt)
 *
 */
#include <linux/config.h>
#include <linux/stddef.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/errno.h>
#include <linux/reboot.h>
#include <linux/pci.h>
#include <linux/kdev_t.h>
#include <linux/major.h>
#include <linux/blk.h>
#include <linux/console.h>
#include <linux/delay.h>
#include <linux/irq.h>
#include <linux/ide.h>
#include <linux/seq_file.h>
#include <linux/root_dev.h>

#include <asm/system.h>
#include <asm/pgtable.h>
#include <asm/page.h>
#include <asm/time.h>
#include <asm/dma.h>
#include <asm/io.h>
#include <asm/machdep.h>
#include <asm/prom.h>
#include <asm/smp.h>
#include <asm/vga.h>
#include <asm/open_pic.h>
#include <asm/i8259.h>
#include <asm/todc.h>
#include <asm/bootinfo.h>
#include <asm/mpc10x.h>
#include <asm/pci-bridge.h>



//#define M_HANK_DEBUG
#ifdef M_HANK_DEBUG
  #define M_DBG(format, arg...)		printk(format , ## arg)
  #define M_FDBG(format, arg...)	printk("%s: " format, __FUNCTION__, ## arg)
#else
  #define M_DBG(format, arg...)
  #define M_FDBG(format, arg...)
#endif

#ifdef M_NH230
  #include <asm/ntx_rtc.h>
  #define SIZE_OF_ETH_ADDRESS  6
  u8 gsz_eth_address[SIZE_OF_ETH_ADDRESS+1];
  
  // For NH-230:
  //   IRQ 0     : reserved
  //   IRQ 1/2/3 : PCI devices : Level-trigger & low-active
  //   IRQ 4     : S/W button  : Level-trigger & high-active
  //          
  #define M_IRQ_TYPE  (IRQ_SENSE_LEVEL | IRQ_POLARITY_NEGATIVE)  
#endif

#include "sandpoint.h"

extern u_int openpic_irq(void);
extern void openpic_eoi(void);

static void	sandpoint_halt(void);
static struct rtc_time tm;//andyw


/*
 * *** IMPORTANT ***
 *
 * The first 16 entries of 'sandpoint_openpic_initsenses[]' are there and
 * initialized to 0 on purpose.  DO NOT REMOVE THEM as the 'offset' parameter
 * of 'openpic_init()' does not work for the sandpoint because the 8259
 * interrupt is NOT routed to the EPIC's IRQ 0 AND the EPIC's IRQ 0's offset is
 * the same as a normal openpic's IRQ 16 offset.
 */
static u_char sandpoint_openpic_initsenses[] __initdata = {
	0,	/* 0-15 not used by EPCI but by 8259 (std PC-type IRQs) */
	0,	/* 1 */
	0,	/* 2 */
	0,	/* 3 */
	0,	/* 4 */
	0,	/* 5 */
	0,	/* 6 */
	0,	/* 7 */
	0,	/* 8 */
	0,	/* 9 */
	0,	/* 10 */
	0,	/* 11 */
	0,	/* 12 */
	0,	/* 13 */
	0,	/* 14 */
	0,	/* 15 */
#ifdef CONFIG_SANDPOINT_X3
	1,	/* 16: EPIC IRQ 0: Active Low -- SIOINT (8259) */
 0, /* AACK!  Shouldn't need this.....see sandpoint_pci.c for more info */
	1,	/* 17: EPIC IRQ 1: Active Low -- PCI Slot 1 */
	1,	/* 18: EPIC IRQ 2: Active Low -- PCI Slot 2 */
	1,	/* 19: EPIC IRQ 3: Active Low -- PCI Slot 3 */
	1,	/* 20: EPIC IRQ 4: Active Low -- PCI Slot 4 */
	0,	/* 21 -- Unused */
	0,	/* 22 -- Unused */
	1,	/* 23 -- IDE (Winbond INT C)  */
	1,	/* 24 -- IDE (Winbond INT D)  */
		/* 35 - 31 (EPIC 9 - 15) Unused */
#else
  #ifdef M_NH230
    (IRQ_SENSE_LEVEL | IRQ_POLARITY_NEGATIVE),  // 16: EPIC IRQ 0: reserved (not-defined)
    (IRQ_SENSE_LEVEL | IRQ_POLARITY_NEGATIVE),  // 17: EPIC IRQ 1: Level-trigger Active Low : Gigabit Ethernet
    (IRQ_SENSE_LEVEL | IRQ_POLARITY_NEGATIVE),  // 18: EPIC IRQ 2: Level trigger Active Low : IDE
    (IRQ_SENSE_LEVEL | IRQ_POLARITY_NEGATIVE),  // 19: EPIC IRQ 3: Level-trigger Active Low : USB
    (IRQ_SENSE_LEVEL | IRQ_POLARITY_POSITIVE)   // 20: EPIC IRQ 4: Level-trigger Active High: S/W buttons
  #else
	1,	/* 16: EPIC IRQ 0: Active Low -- PCI intrs */
	1,	/* 17: EPIC IRQ 1: Active Low -- PCI (possibly 8259) intrs */
	1,	/* 18: EPIC IRQ 2: Active Low -- PCI (possibly 8259) intrs  */
	1	/* 19: EPIC IRQ 3: Active Low -- PCI intrs */
		/* 20: EPIC IRQ 4: Not used */
  #endif
#endif
};

static void __init
sandpoint_setup_arch(void)
{	
	// The CONFIG_SANDPOINT_MEM_CLK defined in Kernel configuration file is
	//   the maximum allowed in the Peripheral Logic/Mem Bus Clock Range (MHz).
	// Please refer to MPC8241 Integrated Processor Hardware Specifications, Table 19.
	//
	// Because we use SDRAM (PC133) 133MHz, and Mem Bus of CPU is configured to
	//   57 ~ 132 by PLL_CFG[4:0] pins. ==> It is 133000000
	//
	loops_per_jiffy = ( CONFIG_SANDPOINT_MEM_CLK  /  HZ );

#ifdef CONFIG_BLK_DEV_INITRD
	if (initrd_start)
		ROOT_DEV = Root_RAM0;
	else
#endif
#ifdef	CONFIG_ROOT_NFS
		ROOT_DEV = Root_NFS;
#else
		ROOT_DEV = Root_HDA1;
#endif

	/* Lookup PCI host bridges */
	sandpoint_find_bridges();

#ifdef CONFIG_DUMMY_CONSOLE
	conswitchp = &dummy_con;
#endif

	printk("Motorola  PMC8241 \n");
	printk("PORTING  by  Premier Microelectronics Development\n");

	/* The Sandpoint rom doesn't enable any caches.  Do that now.
	 * The 7450 portion will also set up the L3s once I get enough
	 * information do do so.  If the processor running doesn't have
	 * and L2, the _set_L2CR is a no-op.
	 */
	if (cur_cpu_spec[0]->cpu_features & CPU_FTR_SPEC7450) {
		/* Just enable L2, the bits are different from others.
		*/
		_set_L2CR(L2CR_L2E);
	}
	else {
		/* The magic number for Sandpoint/74xx PrPMCs.
		*/
		_set_L2CR(0xbd014000);
	}

#ifdef M_NH230

  #ifdef M_03_30_2005         // disable the codes below as no need to do it
    // The patches to NH-230 : Hank Chu : 03/17/2005
    // A-1) Assert the PCI Reset# 2ms
    #include <asm/nh230.h>
    printk("Hardware: assert PCI Reset# to reset PCI peripherals....\n");
    NH230_GPIO_PCI_RESET();
    mdelay(2);
  
    // A-2) De-assert the PCI Reset# 2ms to wait for working
    NH230_GPIO_LED_STATUS_ON();
    mdelay(2);
  #endif

  {
    // 1) Parse the kernel command line to retrieve MAC address
    #define STR_MAC_ADDRESS  "ethaddr="      // must same as boot-loader
    extern char saved_command_line[];
    M_DBG("%s: kernel command line from boot-loader: %s\n", __FUNCTION__, saved_command_line);
    
    // 1-1) Find the "ethaddr=" string in kernel command line
    char *lpsz_ptr = strstr(saved_command_line, STR_MAC_ADDRESS);
    if (lpsz_ptr == NULL) {
      // 1-1-1) Print error as not found
      printk("%s:error: cannot find %s in kernel command line\n", __FUNCTION__, STR_MAC_ADDRESS);     
    }
    else {
      // 1-1-2) Parse the MAC address as found
      int dw_MAC[SIZE_OF_ETH_ADDRESS+1];
      int w_qty = sscanf(lpsz_ptr, "ethaddr=%02X:%02X:%02X:%02X:%02X:%02X",
	                 &dw_MAC[0], &dw_MAC[1], &dw_MAC[2], &dw_MAC[3], &dw_MAC[4], &dw_MAC[5]);
      if (SIZE_OF_ETH_ADDRESS != w_qty) {
      	
      	// 1-1-2-1) Print error as invalid parameter
        printk("%s:error: kernel command line contain invalid %s parameter\n", __FUNCTION__, STR_MAC_ADDRESS);      
        memset(gsz_eth_address, '\0', sizeof(gsz_eth_address));  // NULL it as something wrong
      }
      else {
      	// 1-1-2-2) Duplicate the 6-byte MAC addresses parsed from kernel command line
      	int w_loop = 0;
      	for (w_loop = 0; w_loop < SIZE_OF_ETH_ADDRESS; w_loop++)
      	  gsz_eth_address[w_loop] = (u8)dw_MAC[w_loop];
      }
    }
  }
#endif  // end of #ifdef {M_NH230}
}

static void __init
sandpoint_init2(void)
{
	/* Do nothing  */
	return;
}

static void __init
sandpoint_init_IRQ(void)
{
#ifdef M_NH230
	// Initialize these global variables for arch/ppc/kernel/open_pic.c to use.
        OpenPIC_InitSenses = sandpoint_openpic_initsenses;
        OpenPIC_NumInitSenses = sizeof(sandpoint_openpic_initsenses);
	M_FDBG("OpenPIC_NumInitSenses = %d, IRQ type %d\n", OpenPIC_NumInitSenses, M_IRQ_TYPE);
	
	// Hank : Why must it be 138 ?
	//  The SANDPOINT_X3 seems that it has 138 interrupt sources ?
	openpic_set_sources( 0 , 138, NULL);
#else
	openpic_set_sources( 0 , 138 , NULL );
#endif
	openpic_init(0);
	
	
}

static int
sandpoint_get_irq(struct pt_regs *regs)
{
        int	irq, cascade_irq;

	// 1) Read the Processor Interrupt Acknowledge Register (IACK)
	//    to get the vector of highest pending interrupt in the PIC unit.
	irq = openpic_irq();

#ifdef M_NH230
	// 2) Reading IACK register when no interrupt is pending returns
	//    the spurious vector value.  So, we must check this here.
	if (irq == OPENPIC_VEC_SPURIOUS)
		irq = -1;
	
	// 3) MPC824x supports for connection of external interrupt 
	//    controller device such as an 8259 PIC. 
	//    It should be "cascade_irq = i8259_irq(smp_processor_id())";
	//
	//    But we do not use it, so do not (cascade it and issue EOI).
#else
	if (irq == SANDPOINT_SIO_IRQ) {

		if (cascade_irq != -1) {
			irq = cascade_irq;
			openpic_eoi();
		}
	}
	else if (irq == OPENPIC_VEC_SPURIOUS) {
		irq = -1;
	}
#endif
	return irq;
}

static u32
sandpoint_irq_cannonicalize(u32 irq)
{
	if (irq == 2)
	{
		return 9;
	}
	else
	{
		return irq;
	}
}

static ulong __init
sandpoint_find_end_of_memory(void)
{
	ulong	size = 0;

#if 0	/* Leave out until DINK sets mem ctlr correctly */
	size = mpc10x_get_mem_size(MPC10X_MEM_MAP_B);
#else
	size = ( CONFIG_SANDPOINT_MEM_SIZE*1024*1024 );  // Physical memory size
#endif

	return size;
}

static void __init
sandpoint_map_io(void)
{
	// Set up a mapping for a block of I/O.
	// ==>  io_block_mapping(virtual address, physical address, size, flag)
	// This should only be called before ioremap is called.
	//
	// We map a big area (0xf000.0000 ~ 0xffff.ffff , 256MB) for I/O access later.
	//
	io_block_mapping(0xf0000000, 0xf0000000, 0x10000000, _PAGE_IO);//andyw

#ifdef M_NH230
    // Set up a mapping for a block of I/O (for Extended ROM/Flash space, reserve 1MB here)
    // Must same as boot-loader.
    #include <asm/nh230.h>
    M_DBG("%s: mapping for a block of I/O for GPIO pins on 0x%X\n", __FUNCTION__, NH230_GPIO_ADDRESS);
    io_block_mapping(NH230_GPIO_ADDRESS, NH230_GPIO_ADDRESS, 0x100000, _PAGE_IO);
#endif
}

/*
 * Due to Sandpoint X2 errata, the Port 92 will not work.
 */
static void
_reset_board(void *addr)
{
	__cli();

	/* disable and invalidate the L2 cache */
       	_set_L2CR(0);
       	_set_L2CR(0x200000);

       	/* flush and disable L1 I/D cache */
       	__asm__ __volatile__
       	("\n\
       	mfspr   3,1008
       	ori     5,5,0xcc00
       	ori     4,3,0xc00
       	andc    5,3,5
       	sync
       	mtspr   1008,4
       	isync
       	sync
       	mtspr   1008,5
       	isync
       	sync	
       	");

#ifdef M_NH230
	// For NH-230, issue software reset via the GPIO pin.
	NH230_GPIO_SYS_RESET();
#else
	/* SRR0 has system reset vector, SRR1 has default MSR value */
       	/* rfi restores MSR from SRR1 and sets the PC to the SRR0 value */
       	/* NOTE: assumes reset vector is at 0xfff00100 */
       	__asm__ __volatile__
       	("\n\
       	mtspr   26, %0
       	li      4,(1<<6)	
       	mtspr   27,4
       	rfi
       	":: "r" (addr):"r4");
#endif
	
       	return;	
}

static void
sandpoint_restart(char *cmd)
{
	volatile ulong  i = 10000000;
    _reset_board((void *)0xfff00100);

	while (i-- > 0);
	panic("restart failed\n");
	
}

static void
sandpoint_power_off(void)
{
	unsigned long uhid0;
	local_irq_disable();
	printk("\n POWEROFF....\n");
	 	
	/* determine processor bus speed */
	uhid0 = mfspr(HID0);
	uhid0 &= 0xffff7ffff;
	mtspr( HID0 , uhid0 );
}

static void
sandpoint_halt(void)
{
	sandpoint_power_off();
	/* NOTREACHED */
}

static int
sandpoint_show_cpuinfo(struct seq_file *m)
{
	uint pvid;

	pvid = mfspr(PVR);

	seq_printf(m, "vendor\t\t: Motorola \n");
	seq_printf(m, "machine\t\t: MPC8241\n");
	seq_printf(m, "processor\t: PVID: 0x%x, vendor: %s\n",
			pvid, (pvid & (1<<15) ? "IBM" : "Motorola"));

	return 0;
}


#if defined(CONFIG_BLK_DEV_IDE) || defined(CONFIG_BLK_DEV_IDE_MODULE)
/*
 * IDE support.
 */
static int		sandpoint_ide_ports_known = 0;
static unsigned long	sandpoint_ide_regbase[MAX_HWIFS];
static unsigned long	sandpoint_ide_ctl_regbase[MAX_HWIFS];
static unsigned long	sandpoint_idedma_regbase;

static void
sandpoint_ide_probe(void)
{
#ifdef M_NH230
	// 1) Scan the ITE IT8212 in PCI bus firstly.
        struct pci_dev *pdev = NULL;
        M_DBG("%s: find PCI device with Vendor ID [0x%X] Device ID [0x%X]\n", 
        		__FUNCTION__, PCI_VENDOR_ID_ITE, PCI_DEVICE_ID_ITE_IT8212);
        pdev = pci_find_device(PCI_VENDOR_ID_ITE, PCI_DEVICE_ID_ITE_IT8212, NULL);
        
        // 2) If IT8212 not found, scan ITE IT8211 secondly.
        if (NULL == pdev) {              // Is IT8212 not found ?
          M_DBG("%s: find PCI device with Vendor ID [0x%X] Device ID [0x%X]\n", 
        		__FUNCTION__, PCI_VENDOR_ID_ITE, PCI_DEVICE_ID_ITE_IT8211);
          pdev = pci_find_device(PCI_VENDOR_ID_ITE, PCI_DEVICE_ID_ITE_IT8211, NULL);
        }
#else
        struct pci_dev *pdev = pci_find_device(PCI_VENDOR_ID_WINBOND,
					       PCI_DEVICE_ID_WINBOND_82C105,
					       NULL);
#endif  // end of #ifdef {M_NH230}

        if(pdev) {
                sandpoint_ide_regbase[0]=pdev->resource[0].start;
                sandpoint_ide_regbase[1]=pdev->resource[2].start;
                sandpoint_ide_ctl_regbase[0]=pdev->resource[1].start;
                sandpoint_ide_ctl_regbase[1]=pdev->resource[3].start;
                sandpoint_idedma_regbase=pdev->resource[4].start;
        }
#ifdef M_NH230
        else {
          printk("%s: error: cannot find PCI-to-IDE device\n", __FUNCTION__);
        }
#endif

        sandpoint_ide_ports_known = 1;
	return;
}

static int
sandpoint_ide_default_irq(unsigned long base)
{
        if (sandpoint_ide_ports_known == 0)
	        sandpoint_ide_probe();

	if (base == sandpoint_ide_regbase[0])
		return SANDPOINT_IDE_INT0;
	else if (base == sandpoint_ide_regbase[1])
		return SANDPOINT_IDE_INT1;
	else
		return 0;
}

static unsigned long
sandpoint_ide_default_io_base(int index)
{
        if (sandpoint_ide_ports_known == 0)
	        sandpoint_ide_probe();

	return sandpoint_ide_regbase[index];
}

static void __init
sandpoint_ide_init_hwif_ports(hw_regs_t *hw, unsigned long data_port,
			      unsigned long ctrl_port, int *irq)
{
	unsigned long reg = data_port;
	uint	alt_status_base;
	int	i;

	for (i = IDE_DATA_OFFSET; i <= IDE_STATUS_OFFSET; i++) {
		hw->io_ports[i] = reg++;
	}

	if (data_port == sandpoint_ide_regbase[0]) {
		alt_status_base = sandpoint_ide_ctl_regbase[0] + 2;
#ifdef M_NH230
		hw->irq = SANDPOINT_IDE_INT0;
#else
		hw->irq = 14;
#endif
	}
	else if (data_port == sandpoint_ide_regbase[1]) {
		alt_status_base = sandpoint_ide_ctl_regbase[1] + 2;
#ifdef M_NH230
		hw->irq = SANDPOINT_IDE_INT1;
#else
		hw->irq = 15;
#endif
	}
	else {
		alt_status_base = 0;
		hw->irq = 0;
	}

	if (ctrl_port) {
		hw->io_ports[IDE_CONTROL_OFFSET] = ctrl_port;
	} else {
		hw->io_ports[IDE_CONTROL_OFFSET] = alt_status_base;
	}

	if (irq != NULL) {
		*irq = hw->irq;
	}

	return;
}
#endif


/*
 * Set BAT 3 to map 0xf8000000 to end of physical memory space 1-to-1.
 */
static __inline__ void
sandpoint_set_bat(void)
{
	unsigned long   bat3u, bat3l;
	static int	mapping_set = 0;

	if (!mapping_set) {

		__asm__ __volatile__(
		" lis %0,0xf800\n \
		  ori %1,%0,0x002a\n \
		  ori %0,%0,0x0ffe\n \
		  mtspr 0x21e,%0\n \
		  mtspr 0x21f,%1\n \
		  isync\n \
		  sync "
		  : "=r" (bat3u), "=r" (bat3l));

		mapping_set = 1;
	}

	return;
}

#ifdef	CONFIG_SERIAL_TEXT_DEBUG
#include <linux/serialP.h>
#include <linux/serial_reg.h>
#include <asm/serial.h>

static struct serial_state rs_table[RS_TABLE_SIZE] = {
	SERIAL_PORT_DFNS	/* Defined in <asm/serial.h> */
};

static void
sandpoint_progress(char *s, unsigned short hex)
{
	volatile char c;
	volatile unsigned long com_port;
	u16 shift;

	com_port = rs_table[0].port;
	shift = rs_table[0].iomem_reg_shift;

	while ((c = *s++) != 0) {
		while ((*((volatile unsigned char *)com_port +
				(UART_LSR << shift)) & UART_LSR_THRE) == 0)
		                ;
	        *(volatile unsigned char *)com_port = c;
		
		if (c == '\n') {
			while ((*((volatile unsigned char *)com_port +
				(UART_LSR << shift)) & UART_LSR_THRE) == 0)
					;
	        	*(volatile unsigned char *)com_port = '\r';
		}
	}
	*(volatile unsigned char *)com_port = '\n';//andyw
	*(volatile unsigned char *)com_port = '\r';//andyw
}
#endif	/* CONFIG_SERIAL_TEXT_DEBUG */

__init void sandpoint_setup_pci_ptrs(void);

TODC_ALLOC();


void __init
platform_init(unsigned long r3, unsigned long r4, unsigned long r5,
	      unsigned long r6, unsigned long r7)
{
	parse_bootinfo(find_bootinfo());
	
	//andyw 
	if (r4) {
		initrd_start = r4 + KERNELBASE;
		initrd_end = r5 + KERNELBASE;
	}


	/* Copy the kernel command line arguments to a safe place. */

	if (r6) {
 		*(char *)(r7 + KERNELBASE) = 0;
		strcpy(cmd_line, (char *)(r6 + KERNELBASE));
	}
	//andyw

	/* Map in board regs, etc. */
	sandpoint_set_bat();

	isa_io_base = MPC10X_MAPB_ISA_IO_BASE;
	isa_mem_base = MPC10X_MAPB_ISA_MEM_BASE;
	pci_dram_offset = MPC10X_MAPB_DRAM_OFFSET;
	/*ISA_DMA_THRESHOLD = 0x00ffffff;
	DMA_MODE_READ = 0x44;
	DMA_MODE_WRITE = 0x48;*/

	ppc_md.setup_arch = sandpoint_setup_arch;
	ppc_md.show_cpuinfo = sandpoint_show_cpuinfo;
	ppc_md.irq_cannonicalize = sandpoint_irq_cannonicalize;
	ppc_md.init_IRQ = sandpoint_init_IRQ;
	ppc_md.get_irq = sandpoint_get_irq;
	ppc_md.init = sandpoint_init2;

	ppc_md.restart = sandpoint_restart;
	ppc_md.power_off = sandpoint_power_off;
	ppc_md.halt = sandpoint_halt;

	ppc_md.find_end_of_memory = sandpoint_find_end_of_memory;
	ppc_md.setup_io_mappings = sandpoint_map_io;

#ifdef M_NH230
  #ifdef CONFIG_NTX_RTC
        // M_NH230 : ported to support RTC on I2C bus
	ppc_md.time_init = todc_time_init;
	ppc_md.set_rtc_time = todc_set_rtc_time;
	ppc_md.get_rtc_time = todc_get_rtc_time;
	ppc_md.calibrate_decr = todc_calibrate_decr;
  #else
  	ppc_md.time_init = todc_time_init;
	ppc_md.set_rtc_time = todc_set_rtc_time;
	ppc_md.get_rtc_time = todc_get_rtc_time;
	ppc_md.calibrate_decr = todc_calibrate_decr;
  #endif
#else
	//TODC_INIT(TODC_TYPE_PC97307, 0x70, 0x00, 0x71, 8);
	ppc_md.time_init = todc_time_init;
	ppc_md.set_rtc_time = todc_set_rtc_time;
	ppc_md.get_rtc_time = todc_get_rtc_time;
	ppc_md.calibrate_decr = todc_calibrate_decr;

	//ppc_md.nvram_read_val = todc_mc146818_read_val;//andyw
	//ppc_md.nvram_write_val = todc_mc146818_write_val;//andyw
#endif

	ppc_md.heartbeat = NULL;
	ppc_md.heartbeat_reset = 0;
	ppc_md.heartbeat_count = 0;

#ifdef	CONFIG_SERIAL_TEXT_DEBUG
	ppc_md.progress = sandpoint_progress;
#else	/* !CONFIG_SERIAL_TEXT_DEBUG */
	ppc_md.progress = NULL;
#endif	/* CONFIG_SERIAL_TEXT_DEBUG */


#if defined(CONFIG_BLK_DEV_IDE) || defined(CONFIG_BLK_DEV_IDE_MODULE)
        ppc_ide_md.default_irq = sandpoint_ide_default_irq;
        ppc_ide_md.default_io_base = sandpoint_ide_default_io_base;
        ppc_ide_md.ide_init_hwif = sandpoint_ide_init_hwif_ports;
#endif

	return;
}

