/*
 * FILE NAME NH230_gpio.c
 *
 *  API for NETRONIX GPIO device.
 *  Driver for NETRONIX GPIO device.
 *
 *  Author: Hiram Lee  <hiram@netronixinc.com>
 *
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  either version 2 of the  License, or (at your
 *  option) any later version.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE	LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef __KERNEL__
#define __KERNEL__
#endif

#include <linux/kmod.h>
#include <linux/sched.h>
#include <linux/kernel.h> /* printk() */
#include <linux/fs.h>     /* everything... */
#include <linux/errno.h>  /* error codes */
#include <linux/delay.h>  /* udelay */
#include <linux/slab.h>
#include <linux/tqueue.h>
#include <linux/interrupt.h>
#include <linux/module.h>
#include <linux/config.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/miscdevice.h>
#include <linux/init.h>
#include <linux/ioport.h>
#include <asm/uaccess.h>
#include <linux/sched.h>
#include <linux/ite_gpio.h>

// == Hiram ==
#include <asm/nh230.h>


#define TIMEOUT_OF_LED_TIMER     HZ    // 1 second timer for HeartBeat LED ON/OFF
#define TIMEOUT_OF_HB_TIMER	 HZ/2
#define TIMEOUT_OF_RESET_DEFAULT 4     // count 4 seconds to detect factory default
#define TIMEOUT_OF_POWER_DOWN 2     // count 2 seconds to detect factory default

#define IRQ_NUMBER 20

static struct timer_list gst_RD_timer,gst_PD_timer,flash_timer;
static u8 RD_default_counter; 
static u8 PD_default_counter; 
 
static u8 gb_reset_to_default_flag; 
static u8 gb_power_down_flag; 

void irq_handler(int irq, void *dev_id, struct pt_regs *regs);
static int LED_change_stat (void *data);


static void NH230_Heart_bit(int mdealy);

static int reset_default_thread (void *data);
static int power_down_thread (void *data);

// LED flash 20 times in boot time !!
static int STARTUP_HB_COUNTER = 30;
static int STARTUP_HB_TOGGLE = 0;

#ifdef M_NH230
  // The body of API routine to write a byte to GPIO pins.
  void NH230_GPIO_WRITEB(unsigned char c_byte)
  {
    // 1) Determine the value of GPIO pins according to the combination of USB LEDs.
    extern u8  gc_USB1_LED, gc_USB2_LED;
    c_byte |= (gc_USB1_LED ? 0:GPIO_PIN_WR_USB1_LED_DIS) | (gc_USB2_LED ? 0:GPIO_PIN_WR_USB2_LED_DIS);
    
    // 2) Write to GPIO pins
     __GPIO_WB(c_byte);
  }
#endif

static void ift_RD_timeout(unsigned long dw_param) 
{ 
	int error;		
		       
	if(NH230_GPIO_IS_RESET_DEFAULT())
	{  
		
        	RD_default_counter++;     // add 1 as button held more than 1 second
    	}
    	else
    	{

        	RD_default_counter = 0;   // reset as button released
        	del_timer(&gst_RD_timer); 
        
		while(1)
		{	
			free_irq(IRQ_NUMBER, NULL);	  	
			error = request_irq(IRQ_NUMBER, irq_handler,SA_INTERRUPT, "start_polling_irq_handler",NULL);
	    		if(error)
	    		{
	   			udelay(1000);
	    			printk("POWER_DOWN: error !! IRQ not get..\n");
				BUG();
	    		}
	    		else
	    			break;
	    			
	    	}
		    	
	        return;
    		
	}
	
	
    if (RD_default_counter >= TIMEOUT_OF_RESET_DEFAULT) 
    {	
        printk("System : restore to MFG default setting....\n");
        RD_default_counter = 0;   // reset as button released
        del_timer(&gst_RD_timer);               
	gb_reset_to_default_flag=1;
		
        return;             
    }
    

    mod_timer(&gst_RD_timer, jiffies + TIMEOUT_OF_LED_TIMER);
}

static void ift_PD_timeout(unsigned long dw_param) 
{ 
	int error;		

	if(NH230_GPIO_IS_POWERDOWN())
	{  
		
        PD_default_counter++;     // add 1 as button held more than 1 second
    	}
    	else
    	{

        PD_default_counter = 0;   // reset as button released
        del_timer(&gst_PD_timer); 
        
	while(1)
	{	
		free_irq(IRQ_NUMBER, NULL);	  	
		error = request_irq(IRQ_NUMBER, irq_handler,SA_INTERRUPT, "start_polling_irq_handler",NULL);
    		if(error)
    		{
   			udelay(1000);
    			printk("POWER_DOWN: error !! IRQ not get..\n");
    		}
    		else
    			break;
    			
    	}
	    	
	        return;
    		
	}
		       

  
    if (PD_default_counter >= TIMEOUT_OF_POWER_DOWN) 
    {
        printk("System : Power Down now ....\n");
        PD_default_counter = 0;   // reset as button released
        del_timer(&gst_PD_timer);         
	gb_power_down_flag=1;
		
        return;             
    }
    
    mod_timer(&gst_PD_timer, jiffies + TIMEOUT_OF_LED_TIMER);
}



static void RESET_DEFAULT(void *arg)
{
    init_timer(&gst_RD_timer);
  
    gst_RD_timer.expires = jiffies + (TIMEOUT_OF_LED_TIMER);  // timeout of timer  
    gst_RD_timer.data = 0;  
    gst_RD_timer.function = ift_RD_timeout;                   // timer handler if timeout
  
    add_timer(&gst_RD_timer);
}




static void POWER_DOWN(void *arg)
{
    init_timer(&gst_PD_timer);
  
    gst_PD_timer.expires = jiffies + (TIMEOUT_OF_LED_TIMER);  // timeout of timer  
    gst_PD_timer.data = 0;  
    gst_PD_timer.function = ift_PD_timeout;                   // timer handler if timeout
  
    add_timer(&gst_PD_timer);
}




void irq_handler(int irq, void *dev_id, struct pt_regs *regs)
{
	static struct tq_struct task; 
		
	//printk("GPIO irq_handler \n");
	
		if(NH230_GPIO_IS_RESET_DEFAULT()){  
			free_irq(IRQ_NUMBER, NULL);
			task.routine=(void(*)(void*))RESET_DEFAULT;
			task.data=NULL;
        		queue_task(&task, &tq_immediate);
        		mark_bh(IMMEDIATE_BH);							
			printk("== RESET TO DEFAULT !! ==\n");			
			return;
		}
		
			
					
		if(NH230_GPIO_IS_POWERDOWN()){  
			free_irq(IRQ_NUMBER, NULL);			
			task.routine=(void(*)(void*))POWER_DOWN;
			task.data=NULL;
        		queue_task(&task, &tq_immediate);
        		mark_bh(IMMEDIATE_BH);					
      			printk("== POWER DOWN !! ==\n");
			return;
		}
								
			return;
}

static int power_flash_thread (void *data)
{

    init_timer(&flash_timer);
  
    flash_timer.expires = jiffies + (HZ/2);  // timeout of timer  
    flash_timer.data = 0;  
    flash_timer.function = LED_change_stat;  // timer handler if timeout
  
    add_timer(&flash_timer);
	
}

static int LED_change_stat (void *data)
{
	int i;

	//printk("(hiram)STARTUP_HB_COUNTER(%d) \n",STARTUP_HB_COUNTER);
	
	
	if(STARTUP_HB_COUNTER){
		
		if(!STARTUP_HB_TOGGLE){
			NH230_GPIO_LED_STATUS_ON();
			STARTUP_HB_TOGGLE=1;		
		}else{
			NH230_GPIO_LED_STATUS_OFF();
			STARTUP_HB_TOGGLE=0;
		}
		
		STARTUP_HB_COUNTER-=1;
		mod_timer(&flash_timer, jiffies + (HZ/2));

		return 0;
		
	}

	NH230_GPIO_LED_STATUS_ON();
	del_timer(&flash_timer); 
	return 0;  	
	
}


static int reset_default_thread (void *data)
{
	int i;
	
	DECLARE_WAIT_QUEUE_HEAD(wq);
	
	while (1) {

	    if (gb_reset_to_default_flag)
	    {  // reset to default button pressed a period ?  
		printk("System : reset to default button was pressed more than %d seconds.\n", TIMEOUT_OF_RESET_DEFAULT);        
		printk("System : restore to MFG default setting....\n");
		call_usermodehelper("/etc/rc.d/set_to_default.sh", NULL, NULL);
		gb_reset_to_default_flag = 0;        // reset flag   
	
		// waiting for shell script finished !!    
		for(i=0;i<=20;i++){
			NH230_Heart_bit(300);
		}		    
	    }  
	    
	    interruptible_sleep_on_timeout (&wq, 3*HZ );
	    
	}
	
	return 0;  
}


static int power_down_thread (void *data)
{
  __u32 i;
  spinlock_t *power_down_lock;
  
  DECLARE_WAIT_QUEUE_HEAD(wq);
	
  while (1) {
	  
    if (gb_power_down_flag)
    {  // reset to default button pressed a period ?      	
        printk("System : power down button was pressed more than %d seconds.\n", TIMEOUT_OF_POWER_DOWN);        
        printk("System : power down ....\n");

	spin_lock(power_down_lock);  
	call_usermodehelper("/etc/rc.d/powerdown.sh", NULL, NULL);
#ifdef M_NH230
	// Sleep a moment to wait for user-mode shell script to complete execution
	interruptible_sleep_on_timeout (&wq, 2*HZ );
#endif
	
	// wait wait wait ...
	for(i=0;i<=20;i++){
		NH230_Heart_bit(300);
	}
	
	// Die	
	spin_unlock(power_down_lock);
		
	NH230_GPIO_SYS_SHUTDOWN();
	
	gb_power_down_flag = 0;        // reset flag                 
    
        printk("System : is powered down.\n");
    }  
   
    interruptible_sleep_on_timeout (&wq, 3*HZ );
    
  }	
  
}


static void NH230_Heart_bit(int mdelay)
{
	NH230_GPIO_LED_STATUS_OFF();
	mdelay(mdelay);
	NH230_GPIO_LED_STATUS_ON();
	mdelay(mdelay);

	return;
}


int __init NH230_gpio_init(void)
{	
		
	int rc;
    	RD_default_counter = 0; 
    	PD_default_counter = 0;
	int error = 0;
	
	gb_reset_to_default_flag=0;
	gb_power_down_flag=0;
	printk("NH230_gpio: NH200 GPIO setting ...\n");

#ifdef M_NH230
	// Hank Chu  08/04/2005 :  FAN control for NH-230 
	//   1) Default to set FAN under low speed via GPIO bit.
	NH230_GPIO_FAN_CTRL_LOW();
#endif
    	
    	free_irq(IRQ_NUMBER, NULL);

	error = request_irq(IRQ_NUMBER, irq_handler,SA_INTERRUPT, "start_polling_irq_handler",NULL);

 
   	if(error){
    		printk("NH200_gpio: error !! IRQ not get..\n");
    		BUG();
    	}
	printk("NH230_gpio: Get the IRQ SUCCESS !! \n");
 
   	// power_flash_thread ==========================================================================
    	rc = kernel_thread (power_flash_thread, 0, CLONE_FS | CLONE_FILES);
    	if (rc < 0){
      		printk ("unable to create kernel thread for reset default\n");
		BUG();
	}
 	// ===============================================================================================
	
	
    	// reset_default_thread ==========================================================================
    	rc = kernel_thread (reset_default_thread, 0, CLONE_FS | CLONE_FILES);
    	if (rc < 0){
      		printk ("unable to create kernel thread for reset default\n");
		BUG();
	}  
 	// ===============================================================================================
 
 
     	// power_down_thread =============================================================================
     	rc = kernel_thread (power_down_thread, 0, CLONE_FS | CLONE_FILES);
    	if (rc < 0){
      		printk ("unable to create kernel thread for power down\n");
		BUG();
	}   
      	// ===============================================================================================	     		
     	
	return 0;

}	

void __exit NH230_gpio_exit(void)
{
    free_irq(IRQ_NUMBER, NULL);
}

module_init(NH230_gpio_init);
module_exit(NH230_gpio_exit);

MODULE_LICENSE("GPL");
