//
// Subject :
//   The character device driver for the RTC driver used in NH-230 {M_NH230}.
//   Let Userspace applications access RTC via the character device file (dev/rtc).
//
//   The /dev/rtc device file :
//     1) Major Number of Device is defined by "MAJOR_NO_OF_RTC_DEVICE"
//     2) Name of Device file is defined by "NAME_OF_RTC_DEVICE"
//
//
// History :
//   V1.00  1/21/2005  Hank Chu 
//   1) Create to support a character device driver (/dev/rtc) to let Userspace applications
//      access the RTC device on I2C bus.
//   2) It supports on Philips PCF8563 RTC chip only now.
//      The RTC device driver is located on arch/ppc/drivers/pcf8563.c file.
//
//
//
//

#include <linux/config.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/init.h>
#include <linux/proc_fs.h>
#include <linux/ioctl.h>
#include <linux/rtc.h>                 // for RTC_xxx_xxx commands

#include <asm/uaccess.h>               // for copy_to_user()
#include <asm/rtc.h>
#include <asm/ntx_rtc.h>


  // The definitions for the character device file of RTC driver
#define MAJOR_NO_OF_RTC_DEVICE	121           // The major number of character device.
#define NAME_OF_RTC_DEVICE	"rtc"         // Name which is registered in /proc/devices.
#define NAME_OF_PROC_FILE       "driver/rtc"  // file name under /proc/ directory to read status
#define VERSION_OF_RTC_DEVICE   "v1.00 (01/22/2005)"


  //**********************************************************
  // The IOCTL command for RTC device file
  // Please refer to include/linux/rtc.h for details.
  //**********************************************************
#define RTC_MAGIC 'p'
#define RTC_MAX_IOCTL 0x12

#ifdef CONFIG_SANDPOINT_I2C_RTC_PCF8563
  // Philips PCF8563 RTC chip: specific commands for its RTC device
  #define RTC_SET_CHARGE  _IOW(RTC_MAGIC, 0x0b, int)
  #define RTC_VLOW_RD     _IOR(RTC_MAGIC, 0x11, int)  /* Voltage Low detector */
  #define RTC_VLOW_SET    _IO(RTC_MAGIC, 0x12)        /* Clear voltage low information */
#endif

#ifdef M_NH230
// ansel 20050305 define ascii to integer, declare pcf8563 init, date input format length
#define	ATOI(x)	((x) - '0')
extern long __init  pcf8563_rtc_init(void);
#define	TIME_FORMAT_LEN	13
#endif


static const unsigned char days_in_month[] =
	{ 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };
static spinlock_t ntx_rtc_lock = SPIN_LOCK_UNLOCKED;


int ntx_rtc_ioctl(struct inode *, struct file *, unsigned int, unsigned long);
int ntx_rtc_open(struct inode *, struct file *);
int ntx_rtc_release(struct inode *, struct file *);
#ifdef M_NH230
int ntx_rtc_read(struct file *file, char *buf, size_t count, 
			loff_t *ppos);
int ntx_rtc_write(struct file *file, char *buf, size_t count, 
			loff_t *ppos);
#endif

static struct file_operations pcf8563_fops = {
  owner: THIS_MODULE,
  ioctl: ntx_rtc_ioctl,
  open: ntx_rtc_open,
#ifdef M_NH230
// ansel 20050325 add for /dev/rtc read and write function
  read: ntx_rtc_read,
  write: ntx_rtc_write,
#endif
  release: ntx_rtc_release,
};



//
// Subject :
//   The interface routine for Linux character device to do I/O control.
//   So, the Userspace applications can issue command to it.
//
// Notes :
//   The ioctl calls for this driver. Why return -ENOTTY upon error? Because POSIX says so!
//
int ntx_rtc_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
{
  // 1) Check the given parameters ?
  if (_IOC_TYPE(cmd) != RTC_MAGIC)     // Is IOCTL type not for RTC device ?
    return -ENOTTY;

  if (_IOC_NR(cmd) > RTC_MAX_IOCTL)    // Exceeds maximum command ID ?
    return -ENOTTY;

  // 2) Parse the given command and take actions
  switch (cmd) {
    case RTC_RD_TIME:
    {
      struct rtc_time tm;
      memset(&tm, 0, sizeof (struct rtc_time));
      ntx_get_rtc_time_struct(&tm);

      if (copy_to_user((struct rtc_time *) arg, &tm, sizeof tm)) {
        return -EFAULT;
      }
      return 0;    
    }
    
    case RTC_SET_TIME:
    {
#ifdef CONFIG_SANDPOINT_I2C_RTC_READONLY
      return -EPERM;
#else
      int leap;
      int year;
      int century;
      struct rtc_time tm;

      if (!capable(CAP_SYS_TIME))
        return -EPERM;

      if (copy_from_user(&tm, (struct rtc_time *) arg, sizeof tm))
        return -EFAULT;

      /* Convert from struct tm to struct rtc_time. */
      tm.tm_year += 1900;
      tm.tm_mon += 1;

      /*
       * Check if tm.tm_year is a leap year. A year is a leap
       * year if it is divisible by 4 but not 100, except
       * that years divisible by 400 _are_ leap years.
       */
      year = tm.tm_year;
      leap = (tm.tm_mon == 2) && ((year % 4 == 0 && year % 100 != 0) || year % 400 == 0);

      /* Perform some sanity checks. */
      if ((tm.tm_year < 1970) || (tm.tm_mon > 12) || (tm.tm_mday == 0) ||
          (tm.tm_mday > days_in_month[tm.tm_mon] + leap) ||
          (tm.tm_wday >= 7) || (tm.tm_hour >= 24) || (tm.tm_min >= 60) || (tm.tm_sec >= 60))
        return -EINVAL;

      century = (tm.tm_year >= 2000) ? 0 : 0x80;
      tm.tm_year = tm.tm_year % 100;

      BIN_TO_BCD(tm.tm_year);
      BIN_TO_BCD(tm.tm_mday);
      BIN_TO_BCD(tm.tm_hour);
      BIN_TO_BCD(tm.tm_min);
      BIN_TO_BCD(tm.tm_sec);
      tm.tm_mon |= century;

      // Write to RTC device
      ntx_set_rtc_time_struct(&tm);
      
      return 0;
#endif /* !CONFIG_SANDPOINT_I2C_RTC_READONLY */
    }

#ifdef CONFIG_SANDPOINT_I2C_RTC_PCF8563    // specific to Philips PCF8653 RTC chip   
    case RTC_VLOW_RD:
    {
      int vl_bit = 0;                      // 0 if OK
      if (is_pcf8563_low_voltage()) {      // PCF8563 udner low voltage ?
	vl_bit = 1;
	printk(KERN_WARNING "RTC: RTC Voltage Low - reliable "
	       "date/time information is no longer guaranteed!\n");
      }
      if (copy_to_user((int *) arg, &vl_bit, sizeof(int)))
	return -EFAULT;
	
      return 0;
    }

    case RTC_VLOW_SET:
    {
      // Clear the VL bit in the seconds register
      return (pcf8563_reset_low_voltage());
    }
#endif // end of #ifdef {CONFIG_SANDPOINT_I2C_RTC_PCF8563}

    default:
      return -ENOTTY;
  }
  
  return 0;
}


//
// Subject :
//   The interface routine for Linux character device to open it.
//
int ntx_rtc_open(struct inode *inode, struct file *filp)
{
  MOD_INC_USE_COUNT;
  return 0;
}


//
// Subject :
//   The interface routine for Linux character device to release it.
//
int ntx_rtc_release(struct inode *inode, struct file *filp)
{
  MOD_DEC_USE_COUNT;
  return 0;
}

#ifdef M_NH230
// ansel 20050325 add
// this is used for read and write from /dev/rtc
int ntx_rtc_read(struct file *file, char *buf, size_t count, 
			loff_t *ppos)
{
  struct rtc_time tm;
  memset(&tm, 0, sizeof (struct rtc_time));
  ntx_get_rtc_time_struct(&tm);

//the rtc output format is MMDDHHmmYYYY
  printk("%02d%02d%02d%02d%04d\n", tm.tm_mon, tm.tm_mday,
							             tm.tm_hour, tm.tm_min, tm.tm_year);
  return 0;
}

int ntx_rtc_write(struct file *file, char *buf, size_t count, 
			loff_t *ppos)
{

  int leap;
  int year;
  int century;
  struct rtc_time tm;
  if (count != TIME_FORMAT_LEN) {
		printk("invalid format\n");
		return -EINVAL;
  }
// the rtc input format is MMDDHHmmYYYY
  
  tm.tm_mon = ATOI(buf[0]) * 10 + ATOI(buf[1]);
  tm.tm_mday = ATOI(buf[2]) * 10 + ATOI(buf[3]);
  tm.tm_hour = ATOI(buf[4]) * 10 + ATOI(buf[5]);
  tm.tm_min = ATOI(buf[6]) * 10 + ATOI(buf[7]);
  tm.tm_year = ATOI(buf[8]) * 1000 + ATOI(buf[9]) * 100 + ATOI(buf[10]) * 10 + ATOI(buf[11]);
//  tm.tm_sec = ATOI(buf[12])  * 10 + ATOI(buf[13]);

  /* Perform some sanity checks. */
  if ((tm.tm_year < 1970) || (tm.tm_mon > 12) || (tm.tm_mday == 0) ||
      (tm.tm_mday > days_in_month[tm.tm_mon] + leap) ||
      (tm.tm_wday >= 7) || (tm.tm_hour >= 24) || (tm.tm_min >= 60) || (tm.tm_sec >= 60))
    return -EINVAL;

  century = (tm.tm_year >= 2000) ? 0 : 0x80;
  tm.tm_year = tm.tm_year % 100;

printk("%02d/%02d %02d:%02d:%02d %04d\n", 
                      tm.tm_mon, tm.tm_mday, tm.tm_hour,
                      tm.tm_min, tm.tm_sec, tm.tm_year);
                      
  BIN_TO_BCD(tm.tm_year);
  BIN_TO_BCD(tm.tm_mday);
  BIN_TO_BCD(tm.tm_hour);
  BIN_TO_BCD(tm.tm_min);
  BIN_TO_BCD(tm.tm_sec);
  tm.tm_mon |= century;

  // Write to RTC device


  ntx_set_rtc_time_struct(&tm);
  return count;
}


#endif

//
// Subject :
//   When the Userspace applications use "cat /proc/driver/rtc" command to look
//   the status of RTC device, this routine will be called.
//
static int ntx_rtc_read_proc(char *page, char **start, off_t off, int count, int *eof, void *data)
{
  int len;
  char *lpsz_ptr = page;
  struct rtc_time st_time;
  unsigned long flags;

  // 1) Lock it
  spin_lock_irqsave(&ntx_rtc_lock, flags);
  
  // 1-1) Retrive the RTC information
  ntx_get_rtc_time_struct(&st_time);
#ifdef M_NH230
//ansel 20050325 output for /proc/drivers/rtc, this one include seconds
  lpsz_ptr += sprintf(lpsz_ptr, "%02d%02d%02d%02d%02d%04d\n", 
                      st_time.tm_mon, st_time.tm_mday, st_time.tm_hour,
                      st_time.tm_min, st_time.tm_sec, st_time.tm_year);
#else
  lpsz_ptr += sprintf(lpsz_ptr, "RTC: %04d/%2d/%2d %02d:%02d:%02d\n", 
                      st_time.tm_year, st_time.tm_mon, st_time.tm_mday, 
                      st_time.tm_hour, st_time.tm_min, st_time.tm_sec);
#endif
  // 2) Unlock it
  spin_unlock_irqrestore(&ntx_rtc_lock,flags);	

  // 3) Calculate the length
  len = (int) (lpsz_ptr - page);  
  if (len <= off+count) 
    *eof = 1;
    
  *start = page + off;
  len -= off;
  
  if (len>count) 
    len = count;
    
  if (len<0) 
    len = 0;
  
  return len;
}


int __init ntx_rtc_init(void)
{
  // 1) Register a character device driver with major number and device name
  int w_result = register_chrdev(MAJOR_NO_OF_RTC_DEVICE, NAME_OF_RTC_DEVICE, &pcf8563_fops);
  
  if (w_result < 0) {                  // failed to register ?
    printk(KERN_ERR "RTC: cannot get register character device (/dev/%s) with major number %d.\n", 
                    NAME_OF_RTC_DEVICE, MAJOR_NO_OF_RTC_DEVICE);
    return (w_result);
  }

  // 2) Create read entry under /proc/ directory to let Userspace applications access RTC status
  struct proc_dir_entry  *st_dir;
  st_dir = create_proc_read_entry(NAME_OF_PROC_FILE, 0, NULL, ntx_rtc_read_proc, NULL);
  if (st_dir == NULL) {
    printk(KERN_ERR "RTC: cannot create /proc/%s for character device (/dev/%s).\n", 
                    NAME_OF_PROC_FILE, NAME_OF_RTC_DEVICE);
  }
  
  printk(KERN_INFO "RTC: character device (/dev/%s) driver %s is registered\n", 
          NAME_OF_RTC_DEVICE,  VERSION_OF_RTC_DEVICE);

#ifdef M_NH230
//ansel 20050325 here we init pcf8563, early init may cause boot failure
  pcf8563_rtc_init();
#endif
  return (0);
}


void __exit ntx_rtc_exit(void)
{
  // 1) Remove the entry under /proc/ directory
  remove_proc_entry(NAME_OF_PROC_FILE, NULL);
  
  // 2) Un-register it
  if (unregister_chrdev(MAJOR_NO_OF_RTC_DEVICE, NAME_OF_RTC_DEVICE) < 0) {
    printk(KERN_ERR "RTC: cannot unregister character device (/dev/%s).\n", NAME_OF_RTC_DEVICE);
  }
}


EXPORT_NO_SYMBOLS;
module_init(ntx_rtc_init);
module_exit(ntx_rtc_exit);
