/*
 *  linux/drivers/ide/it8212.c	Version 0.01	January 25, 2003
 *
 *  Copyright (C) 1998-2000	Andre Hedrick <andre@linux-ide.org>
 *  May be copied or modified under the terms of the GNU General Public License
 *
 *  Portions Copyright (C) 2003 Integrated Technology Express , Inc.
 *  Author: Liangyao Li <liangyao.li@iteusa.com>
 *  Released under terms of General Public License
 *
 *  Jan 25, 2003 Liangyao Li
 *		initial version.
 */

//
// History :
//   V1.00  02/01/2005  Hank Chu
//     1) Ported the Linux driver for ITE IT8212/8211 PCI-to-IDE chips.
//          From :  
//		MIPS CPU + Linux kernel v2.4.20
//          To   :  
//		Motorola MPC8241 + Linux kernel v2.4.21.
//
//     2) Support ITE IT8212 and IT8211 both.
//
//   V1.01  03/18/2005  Hank Chu
//     1) Add the interface routines for PCI-to-IDE driver according to Linux 2.4.21 OS.
//
//   V1.02  06/14/2005  Hank Chu
//     1) Bug fixed in NH-230 :
//          Problem :
//            In NH-230 (Motorola MPC8241 + RTL8169S + IT8212 + Genesys GL880S),
//            when we test with the following test environment, USB1.1 failed
//            but file copying to USB 2.0 and IDE HD still works fine.
//              In PC with Windows, copy files through the following FTP/Samba connections.
//                1 FTP connection to shared folder of IDE H.D. in NAS
//                1 Samba connection to shared folder of USB 2.0 external HD in NAS
//                1 Samba conenction to shared folder of USB 1.1 flash disk in NAS
//              In fact, the 2 Samba connections use the same Samba daemon in NAS.
//
//              After a moment (within 10 minutes), the Samba connection to USB 1.1 
//              failed as RS-232 console display the following error :
//                 I/O error: dev 08:11, sector xxxx
//
//          Reason :
//            If do not run the FTP connection to shared folder of IDE H.D.,
//            Samba connection to USB 2.0 and 1.1 storage devices works well.
//
//            It seems the PCI behavior of IDE controller causes this problem.
//
//          Solution :
//            The original driver source codes write the PCI Configuration Register
//            "PCI_LATENCY_TIMER" to 0x00.
//            After change to 0x40, this problem is fixed.
//
//            Please refer to the following routine :
//                void it8212_initPciRegisters(struct pci_dev *dev)
//
//
//
//
//
//


#include <linux/config.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/timer.h>
#include <linux/mm.h>
#include <linux/ioport.h>
#include <linux/blkdev.h>
#include <linux/hdreg.h>
#include <linux/interrupt.h>
#include <linux/pci.h>
#include <linux/init.h>
#include <linux/ide.h>

#include <linux/stat.h>
#include <linux/proc_fs.h>

#include <asm/io.h>
#include <asm/irq.h>

#include "ide_modes.h"

//#define IT8212_DEBUG	               // define it to enable debug messages
#include "it8212.h"


//#define M_HANK_DEBUG                 // define it to enable debugging messages
#ifdef M_HANK_DEBUG
  #define M_DBG(format, arg...)		printk(format , ## arg)
  #define M_FDBG(format, arg...)	printk("%s: " format, __FUNCTION__, ## arg)
#else
  #define M_DBG(format, arg...)
  #define M_FDBG(format, arg...)
#endif

#define M_HANK_PATCH   // define it to enable the patches


/************************************************************************
 * Global data and function prototype definitions
 ************************************************************************/

byte it8212_proc = 0;	// refered in ide-proc.c
extern int (*it8212_display_info)(char *, char **, off_t, int);	// from ide-proc.c

#ifdef M_NH230
  #include <linux/module.h>
  #include <ide-timing.h>
  #include <asm/nh230.h>          // APIs to read/write the GPIO : pins NH230_GPIO_READB()
  
  #define IN_BYTE(reg)          ( HWIF(drive)->INB(reg))
  #define OUT_BYTE(value, reg)  ( HWIF(drive)->OUTB(value, reg) )
  
  unsigned int __init ata66_it8212(ide_hwif_t *hwif);
  int it8212_getHsbIndex(u8 value);
  void it8212_clearDriveIntStatus(ide_drive_t *drive);
  void it8212_setDriveTiming(int channel, int drive);  
#endif


/************************************************************************
 * Local data and function prototype definitions
 ************************************************************************/

typedef struct
{
	u8 type;
	u8 mode;
	u8 clock;
	u8 udmaTiming;
	u8 pioDmaTiming;
} drive_config_t;


struct
{
	struct pci_dev *device;
	ide_hwif_t *hwif[NofCHANNEL];
	ide_drive_t *drive[NofCHANNEL][NofDRIVE];
	drive_config_t config[NofCHANNEL][NofDRIVE];
} it8212;




#ifdef M_NH230
  #define M_IDE_SLEEP_MODE             // define it to enable IDE sleep mode.
#endif

#ifdef M_IDE_SLEEP_MODE

  #define M_DEBUG_SLEEP_MODE         // define it to enable debugging messages for SLEEP mode
  #ifdef M_DEBUG_SLEEP_MODE
    #define M_MSG(format, arg...)	printk(format , ## arg)
    #define M_FMSG(format, arg...)	printk("%s: " format, __FUNCTION__, ## arg)
  #else
    #define M_MSG(format, arg...)
    #define M_FMSG(format, arg...)
  #endif  


  #define TIMEOUT_OF_IDE_SLEEP_MODE  (5 * 60 * HZ)  // Idle timeout 5 minutes
  #define TIMER_OF_IDE_SLEEP_MODE    (30 * HZ)      // Timer triggered per 30 seconds
  unsigned long gdw_ide_last_access_time = 0;       // Last access time of IDE drive
  u8  gc_ide_sleep_flag = 0;                        // 1 if IDE drives under sleep mode
  u8  gc_IDE_sleep_mode_enable = 0;                 // 1 if IDE sleep mode is enabled
  
  #define GET_STAT()   HWIF(drive)->INB(IDE_STATUS_REG)
  
  //
  // Subject :
  //   Wait until IDE command completion or timeout.
  // Input : 
  //   *drive : pointer to IDE drive
  // Output :
  //   0 : OK
  //
  int fn_IDE_wait_cmd(ide_drive_t *drive)
  {
    if (!drive) {
      printk("%s: NULL pointer found.\n", __FUNCTION__);
      return (-EINVAL);
    }
  
    // 1-4) Wait Busy bit is cleaned after SRST bit is reset (up to 31 seconds)
    #define MAX_OF_RESET_BUSY  40 * 1000     // 40 seconds
    u16  dw_loop = 0;
    for (dw_loop = 0; dw_loop < MAX_OF_RESET_BUSY; dw_loop++) {
      mdelay(1);
      if (!(GET_STAT() & BUSY_STAT))
        break;
    }
    
    // 1-4-1) Check whether it is timeout or not ?
    u8 c_busy_timeout_flag = 0;
    if (dw_loop >= MAX_OF_RESET_BUSY) {
      printk("IDE: error: timeout to wait for Busy bit cleaned.\n");
      c_busy_timeout_flag = 1;               // set as timeout
    }
          
    // 1-5) Wait for Device Ready bit is set (no maximum time is specified in spec.)
    for (dw_loop = 0; dw_loop < MAX_OF_RESET_BUSY; dw_loop++) {
      mdelay(1);
      if (GET_STAT() & READY_STAT)
        break;            
    }
    
    // 1-5-1) Check whether it is timeout or not ?
    u8 c_ready_timeout_flag = 0;
    if (dw_loop >= MAX_OF_RESET_BUSY) {
      printk("IDE: error: timeout to wait for Ready bit asserted.\n");
      c_ready_timeout_flag = 1;
    }
    
    if (c_busy_timeout_flag || c_ready_timeout_flag)
      return (-EIO);
    else
      return (0);
  }
  
  
  //
  // Subject :
  //   Flush the IDE cache before enter sleep mode.
  // Input :
  //   *drive : pointer to IDE drive.
  // Output :
  //   0 : OK
  //
  int fn_idedisk_flushcache(ide_drive_t *drive)
  {   
    if (!drive) {
      printk("%s: NULL pointer found.\n", __FUNCTION__);
      return (-EINVAL);
    }

    if (drive->id->cfs_enable_2 & 0x2400) {
    	OUT_BYTE(WIN_FLUSH_CACHE_EXT, IDE_COMMAND_REG);
    } else {
    	OUT_BYTE(WIN_FLUSH_CACHE, IDE_COMMAND_REG);
    }
    
    if (fn_IDE_wait_cmd(drive))
      return (-EIO);
    else
      return (0);
  }
  
  
  //
  // Subject :
  //   This routine will issue the command to IDE drives connected to enter sleep mode.
  // Input :
  //   None
  // Output :
  //   None
  //
  void fn_disks_sleep(void) 
  {
    int i, d;
    ide_drive_t *drive;

    // 1) Loop to let all drives of all channels enter the sleep mode.
    for ( i=0 ; i<NofCHANNEL ; i++ ) {
      
      // 1-1) If this IDE channel is not present, skip it
      if ( !it8212.hwif[i] || !it8212.hwif[i]->present )
        continue;
      
      // 1-2) Loop to go through the drives in this channel.
      ////M_FMSG("HD channel[%d]: name %s\n", i, it8212.hwif[i]->name);
      for ( d=0 ; d<NofDRIVE ; d++ ) {
      	
      	// 1-2-1) If this IDE drive is not present, skip it.
        if ( !it8212.hwif[i]->drives[d].present )
          continue;

        ////M_FMSG("HD channel[%d]: drive[%d]: drive %s\n", i, d, it8212.hwif[i]->drives[d].name);
        drive = &it8212.hwif[i]->drives[d];        
        
        // 1-2-2) Flush the Write Cache of IDE drive first if existed.
        if (drive->wcache) {
          if (fn_idedisk_flushcache(drive))
            printk(KERN_NOTICE "IDE: error: unable to flush writecache of %s.\n", drive->name);
        }
        
	OUT_BYTE(WIN_SLEEPNOW1, IDE_COMMAND_REG);
	OUT_BYTE(WIN_SLEEPNOW2, IDE_COMMAND_REG);	
      }
    }
  }
   
  
  //
  // Subject :
  //   The kernel thread will monitor the IDE idle in a fixed period.
  // Input :
  //   *data :  pointer to data passed by caller
  // Output :
  //
  static int fn_IDE_sleep_mode_thread (void *data)
  {
    // 1) Declare a wait queue.
    DECLARE_WAIT_QUEUE_HEAD(wq);
	
    // 2) Loop to check whether the drivers should enter sleep mode or not ?
    while (1) {
      sleep_on_timeout (&wq, TIMER_OF_IDE_SLEEP_MODE );
      
      if (!gc_IDE_sleep_mode_enable)    // Skip if IDE sleep mode is disabled ?
        continue;

      // 2-1) If the last access time of IDE had been initialized, calcualte the offset.
      unsigned long dw_offset = 0;
      if (gdw_ide_last_access_time)
        dw_offset = (jiffies >= gdw_ide_last_access_time) ? \
                    (jiffies - gdw_ide_last_access_time) : \
                    (ULONG_MAX - gdw_ide_last_access_time) + jiffies;
    
      // 2-2) If IDE not under sleep mode and offset more than idle timeout, enter sleep mode.
      if (!gc_ide_sleep_flag && dw_offset >= TIMEOUT_OF_IDE_SLEEP_MODE) {
      	
      	// 2-2-1) Put all disks into sleep mode and set flag
    	fn_disks_sleep();
    	gc_ide_sleep_flag = 1;
    	M_MSG("IDE: enter sleep mode on jiffies[%lu], last access[%lu]\n", jiffies, gdw_ide_last_access_time);
      }
    }   // end of while{}  
    
    return (0);
  }


  int fn_IDE_wake_up(ide_drive_t *drive)
  {
    if (!drive) {
      printk("%s: NULL pointer found.\n", __FUNCTION__);
      return (-EINVAL);
    }

    // 1-1) Set SRST bit (bit[2]) of IDE Device Control Register
    M_MSG("IDE: wake up on jiffies[%lu], last access[%lu]\n", jiffies, gdw_ide_last_access_time);
    OUT_BYTE( IN_BYTE(IDE_CONTROL_REG) | 0x04, IDE_CONTROL_REG);
	  
    // Busy bit asserted no later than 400ns after SRST bit set for device 0
    // Device 1 negates PDIAG- no later than 1 ms after detecting SRST bit set
    mdelay(2);                               // delay 2 ms
          
    // 1-3) Clear the SRST bit          
    OUT_BYTE( IN_BYTE(IDE_CONTROL_REG) & (~0x04), IDE_CONTROL_REG);
	  
    // 1-4) Wait Busy bit is cleaned after SRST bit is reset (up to 31 seconds)
    #define MAX_OF_RESET_BUSY  40 * 1000     // 40 seconds
    u16  dw_loop = 0;
    for (dw_loop = 0; dw_loop < MAX_OF_RESET_BUSY; dw_loop++) {
      mdelay(1);
      if (!(GET_STAT() & BUSY_STAT))
        break;
    }
    
    // 1-4-1) Check whether it is timeout or not ?
    if (dw_loop >= MAX_OF_RESET_BUSY)
      printk("IDE: error: timeout to wait for Busy bit cleaned.\n");
          
    // 1-5) Wait for Device Ready bit is set (no maximum time is specified in spec.)
    for (dw_loop = 0; dw_loop < MAX_OF_RESET_BUSY; dw_loop++) {
      mdelay(1);
      if (GET_STAT() & READY_STAT)
        break;            
    }
    
    // 1-5-1) Check whether it is timeout or not ?
    if (dw_loop >= MAX_OF_RESET_BUSY) {
      printk("IDE: error: timeout to wait for Ready bit asserted.\n");
      return (-EIO);
    }
    else    
      return (0);
  }
#endif  // end of #ifdef {M_IDE_SLEEP_MODE}


#ifdef M_HANK_PATCH
static int it8212_find_mapping(u8 c_mode, u8 xferspeed);

//
// Subject ;
//   Loop to look up the given array to find out according index.
// Input :
//   c_mode : UDMA_MODE, DMA_MODE, or PIO_MODE
//   xferspeed : XFER_UDMA_6, .., XFER_MW_DMA_2, .., XFER_PIO_4, .. and etc.
// Output :
//   Must be a signed integer
//
//     >= 0 : index in array
//     < 0  : error
//
static int it8212_find_mapping(u8 c_mode, u8 xferspeed)
{
  int w_ID = 0, w_speed = 0, w_found_flag = 0;
  M_FDBG("called: c_mode = %d, xferspeed = %d\n", c_mode, xferspeed);
  M_FDBG("M_ELEMENTS_OF_UDMA=%d, M_ELEMENTS_OF_DMA=%d, M_ELEMENTS_OF_PIO=%d\n", 
          M_ELEMENTS_OF_UDMA, M_ELEMENTS_OF_DMA, M_ELEMENTS_OF_PIO);
 
  // 1) Loop to look up the given array
  do {
  
    // 1-1) Map to speed in array according to the given mode
    switch (c_mode) {
      case UDMA_MODE :
        if (M_ELEMENTS_OF_UDMA <= w_ID)
	  return (-EINVAL);
	
        w_speed = udmaConfigValue[w_ID][ModeINDEX];
        break;
      
      case DMA_MODE :
        if (M_ELEMENTS_OF_DMA <= w_ID)
	  return (-EINVAL);
	
        w_speed = dmaConfigValue[w_ID][ModeINDEX];
        break;
      
      case PIO_MODE :
        if (M_ELEMENTS_OF_PIO <= w_ID)
	  return (-EINVAL);
	
        w_speed = pioConfigValue[w_ID][ModeINDEX];
        break;

      default :
        return (-EINVAL);
    }
   
     // 1-2) Check the speed in array with given one
    M_FDBG("ID[%d]: given speed=%d, speed in array=%d\n", w_ID, xferspeed, w_speed);
    if ( w_speed == xferspeed ) {
      w_found_flag = 1;                 // match, then found
      break;
    }

    w_ID++;                             // next one
    
  } while (1);
 
  return (w_found_flag ? w_ID : -EINVAL);
}

//
// Subject :
//   Apply the transfer speed to IT8212 configuration registers and etc.
// Input :
//   xferspeed : XFER_UDMA_6, ...., XFER_PIO_2, ...., and etc.
// Output :
//   0      : OK
//   Others : cannot convert it
// Modified :
//   driveConfig->type
//   driveConfig->mode
//   driveConfig->clock
//   ....
//   and etc.
//
static u8 it8212_fn_set_speed (ide_drive_t *drive, u8 xferspeed)
{
  int w_ID = 0;
  int driveNum, channelNum;
  struct hd_driveid *id;
  drive_config_t *driveConfig;

  id = drive->id;
  driveNum = drive->dn % NofDRIVE;
  channelNum = drive->dn / NofDRIVE;
  driveConfig = &(it8212.config[channelNum][driveNum]);
  
  M_FDBG("called: xferspeed = %u, drive NO#[%d], channel %d\n", 
          xferspeed, driveNum, channelNum);
  
  // 1) Convert the given speed into the mapping of IT8212 configuration
  switch(xferspeed) {
    case XFER_UDMA_6:                  // IT8212 supports Ultra DMA 0/1/2/3/4/5/6
    case XFER_UDMA_5:
    case XFER_UDMA_4:
    case XFER_UDMA_3:
    case XFER_UDMA_2:
    case XFER_UDMA_1:
    case XFER_UDMA_0:      
      // 1-1) Find it in Ultra DMA array
      if ( (w_ID = it8212_find_mapping(UDMA_MODE, xferspeed)) < 0) {
        printk("%s: given speed [%d] is not supported.\n", __FUNCTION__, xferspeed);
	return (-EINVAL);
      }
      
      driveConfig->type       = UDMA_MODE;
      driveConfig->mode       = udmaConfigValue[w_ID][ModeINDEX];
      driveConfig->clock      = udmaConfigValue[w_ID][ClockINDEX];
      driveConfig->udmaTiming = udmaConfigValue[w_ID][TimingINDEX];
      break;
    
    case XFER_MW_DMA_2:                // IT8212 supports MultiWord DMA 0/1/2
    case XFER_MW_DMA_1:
    case XFER_MW_DMA_0:
      // 1-2) Find it in MultiWord DMA array
      if ( (w_ID = it8212_find_mapping(DMA_MODE, xferspeed)) < 0) {
        printk("%s: given speed [%d] is not supported.\n", __FUNCTION__, xferspeed);
	return (-EINVAL);
      }
      
      driveConfig->type       = DMA_MODE;
      driveConfig->mode       = dmaConfigValue[w_ID][ModeINDEX];
      driveConfig->clock      = dmaConfigValue[w_ID][ClockINDEX];
      driveConfig->udmaTiming = dmaConfigValue[w_ID][TimingINDEX];
      break;
	
    case XFER_PIO_4:                   // IT8212 supports PIO mode 0/1/2/3/4
    case XFER_PIO_3:
    case XFER_PIO_2:
    case XFER_PIO_1:    
    case XFER_PIO_0:
      // 1-3) Find it in PIO array
      if ( (w_ID = it8212_find_mapping(PIO_MODE, xferspeed)) < 0) {
        printk("IT8212: given speed [%d] is not supported.\n", xferspeed);
	return (-EINVAL);
      }
      
      driveConfig->type       = PIO_MODE;
      driveConfig->mode       = pioConfigValue[w_ID][ModeINDEX];
      driveConfig->clock      = pioConfigValue[w_ID][ClockINDEX];
      driveConfig->udmaTiming = pioConfigValue[w_ID][TimingINDEX];

      break;

    case XFER_SW_DMA_2:
    case XFER_SW_DMA_1:
    case XFER_SW_DMA_0:
    case XFER_PIO_SLOW:
    default:
      // 1-4) the given speed is not supported
      printk("IT8212: the given speed [%d] is not supported.\n", xferspeed);
      return (-EINVAL);
  }

  // 2) Update these parameters of speed to registers  
  M_FDBG("given speed[%d]: found in array index[%d], type[%d], mode[%d]\n", 
         xferspeed, w_ID, driveConfig->type, driveConfig->mode);

  it8212_setDriveTiming(channelNum, driveNum);
////  ide_config_drive_speed(drive, driveConfig->mode);
  it8212_clearDriveIntStatus(drive);

  return (ide_config_drive_speed(drive, xferspeed));
}


//
// Subject :
//   This routine is called by Linux IDE layer to tune PIO mode for drivers.
//   Called by upper layer (IDE layer).
// Input :
//   *drive     : pointer to ide_drive_t structure
//   c_pio_mode : 0/1/2/3/4 for PIO 0/1/2/3/4
// Output :
//   None
//
static void it8212_tuneproc (ide_drive_t *drive, u8 c_pio_mode)
{
  // 1) Adjust the hardware according to the given transfer speed
  u8 xferspeed = XFER_PIO_4;
  M_FDBG("called: re-tune to PIO mode: PIO-%d\n", c_pio_mode);
  
  // 1-1) Convert the PIO mode to speed (XFER_PIO_4, .... and etc.)
  if (c_pio_mode == 255)                // Auto-tune to best PIO mode ?
    xferspeed = ide_find_best_mode(drive, XFER_PIO);
  else {
    
    // 1-1-1) Validate whether the given PIO mode is supported by IT8212 ?
    if (M_ELEMENTS_OF_PIO <= c_pio_mode) {
      printk("IT8212: the give PIO-%d mode is not supported.\n", c_pio_mode);
      return;
    }
    else {
      // 1-1-2) Convert from PIO (0/1/2/3/4) to speed (XFER_PIO_4, .... , and etc.)
      xferspeed = XFER_PIO_0 + c_pio_mode;
    }
  }
  
  // 2) Apply it to hardware
  it8212_fn_set_speed (drive, xferspeed);
}


//
// Subject :
//   This routine is called by Linux IDE layer to tune DMA mode for drivers.
//   Called by upper layer (IDE layer).
// Input :
//   *drive    : pointer to ide_drive_t structure
//   xferspeed : XFER_UDMA_6, .... and etc.
// Output :
//   0 : OK
//
static int it8212_speedproc (ide_drive_t *drive, u8 xferspeed)
{
  // 1) Adjust the hardware according to the given transfer speed
  M_FDBG("called: re-tune to DMA mode: xferspeed = %d\n", xferspeed);
  
  return ( it8212_fn_set_speed (drive, xferspeed) );
}


#if 0
static int it8172_config_chipset_for_dma (ide_drive_t *drive)
{
  u8 speed = ide_dma_speed(drive, it8212_ratemask(drive));

  if (!(speed)) {
    u8 tspeed = ide_get_best_pio_mode(drive, 255, 4, NULL);
    speed = it8212_dma_2_pio(XFER_PIO_0 + tspeed);
  }

  (void) it8212_speedproc(drive, speed);
   return ide_dma_enable(drive);
}
#endif

#endif // end of #ifdef {M_HANK_PATCH}

/************************************************************************
 *
 ************************************************************************/

void it8212_showDriveTiming(int channel)
{
	u8 modeControl;
	u8 udmaTiming;
	u8 pioDmaTiming;
	int channelOffset;
	struct pci_dev *dev;

	ITE_KDBG("(channel %d)\n", channel);

	dev = it8212.device;
	channelOffset = 4 * channel;

	pci_read_config_byte(dev, IT8212_PCI_PciModeCONTROL, &modeControl);
	pci_read_config_byte(dev, IT8212_PCI_PciModePrimPioDmaTIMING + channelOffset, &pioDmaTiming);
	pci_read_config_byte(dev, IT8212_PCI_PciModePrimDrive0UdmaTIMING + channelOffset, &udmaTiming);

	ITE_DBG("modeControl	%02x\n", modeControl);
	ITE_DBG("udmaTiming	%02x\n", udmaTiming);
	ITE_DBG("pioDmaTiming	%02x\n", pioDmaTiming);
}


/************************************************************************
 *
 ************************************************************************/

void it8212_showIdeRegisters(ide_hwif_t *hwif)
{
	u32 dword;
	u8 byte1, byte2;
	ide_drive_t *drive;
	unsigned long busMasterBase;

	ITE_KDBG("(%s)\n", hwif->name);

	drive = &(hwif->drives[0]);
	busMasterBase = hwif->dma_base;

	ITE_DBG("%s\n", hwif->name);
	ITE_DBG("data	%08lx  %04x\n", IDE_DATA_REG, inw(IDE_DATA_REG));
	ITE_DBG("error	%08lx  %02x\n", IDE_ERROR_REG, IN_BYTE(IDE_ERROR_REG));
	ITE_DBG("nsector	%08lx  %02x\n", IDE_NSECTOR_REG, IN_BYTE(IDE_NSECTOR_REG));
	ITE_DBG("sector	%08lx  %02x\n", IDE_SECTOR_REG, IN_BYTE(IDE_SECTOR_REG));
	ITE_DBG("lcyl	%08lx  %02x\n", IDE_LCYL_REG, IN_BYTE(IDE_LCYL_REG));
	ITE_DBG("hcyl	%08lx  %02x\n", IDE_HCYL_REG, IN_BYTE(IDE_HCYL_REG));
	ITE_DBG("select	%08lx  %02x\n", IDE_SELECT_REG, IN_BYTE(IDE_SELECT_REG));
	ITE_DBG("status	%08lx  %02x\n", IDE_STATUS_REG, IN_BYTE(IDE_STATUS_REG));
	ITE_DBG("control	%08lx  %02x\n", IDE_CONTROL_REG, IN_BYTE(IDE_CONTROL_REG));
	ITE_DBG("irq 	%08lx  %02x\n", IDE_IRQ_REG, IN_BYTE(IDE_IRQ_REG));

	ITE_DBG("BusMaster IDE IORegs\n");

	byte1 = inb(busMasterBase);
	byte2 = inb(busMasterBase + 2);
	dword = inl(busMasterBase + 4);

	ITE_DBG("command	%08lx  %02x \n", busMasterBase, byte1);
	ITE_DBG("status	%08lx  %02x\n", busMasterBase + 2, byte2);
	ITE_DBG("table	%08lx  %08x\n", busMasterBase + 4, dword);
}


/************************************************************************
 *
 ************************************************************************/

void it8212_showPciRegisters(struct pci_dev *dev)
{
	u8 byte1, byte2, byte3, byte4;
	u16 word1, word2;
	u32 dword;

	ITE_KDBG("(%s)\n", dev->name);

	pci_read_config_word(dev, PCI_VENDOR_ID, &word1);	
	pci_read_config_word(dev, PCI_DEVICE_ID, &word2);
	ITE_DBG("00h   %04x %04x\n", word1, word2);

	pci_read_config_word(dev, PCI_COMMAND, &word1);
	pci_read_config_word(dev, PCI_STATUS, &word2);
	ITE_DBG("04h   %04x %04x\n", word1, word2);

	pci_read_config_byte(dev, PCI_CLASS_REVISION, &byte1);	
	pci_read_config_byte(dev, PCI_CLASS_PROG, &byte2);
	pci_read_config_byte(dev, PCI_CLASS_DEVICE, &byte3);	
	pci_read_config_byte(dev, IT8212_PCI_BaseClassCode, &byte4);
	ITE_DBG("08h   %02x %02x %02x %02x\n", byte1, byte2, byte3, byte4);

	pci_read_config_byte(dev, PCI_CACHE_LINE_SIZE, &byte1);	
	pci_read_config_byte(dev, PCI_LATENCY_TIMER, &byte2);
	pci_read_config_byte(dev, PCI_HEADER_TYPE, &byte3);	
	pci_read_config_byte(dev, PCI_BIST, &byte4);
	ITE_DBG("0Ch   %02x %02x %02x %02x\n", byte1, byte2, byte3, byte4);

	pci_read_config_dword(dev, PCI_BASE_ADDRESS_0, &dword);
	ITE_DBG("10h   %08x\n", dword);
	
	pci_read_config_dword(dev, PCI_BASE_ADDRESS_1, &dword);
	ITE_DBG("14h   %08x\n", dword);

	pci_read_config_dword(dev, PCI_BASE_ADDRESS_2, &dword);
	ITE_DBG("18h   %08x\n", dword);
	
	pci_read_config_dword(dev, PCI_BASE_ADDRESS_3, &dword);
	ITE_DBG("1Ch   %08x\n", dword);

	pci_read_config_dword(dev, PCI_BASE_ADDRESS_4, &dword);
	ITE_DBG("20h   %08x\n", dword);

	ITE_DBG("24h   --------\n");
	ITE_DBG("28h   --------\n");
	
	pci_read_config_word(dev, PCI_SUBSYSTEM_VENDOR_ID, &word1);
	pci_read_config_word(dev, PCI_SUBSYSTEM_ID, &word2);
	ITE_DBG("2Ch   %04x %04x\n", word1, word2);

	pci_read_config_dword(dev, PCI_ROM_ADDRESS, &dword);
	ITE_DBG("30h   %08x\n", dword);

	pci_read_config_byte(dev, PCI_CAPABILITY_LIST, &byte1);
	ITE_DBG("34h   %02x -- -- --\n", byte1);

	ITE_DBG("38h   --------\n");

	pci_read_config_byte(dev, PCI_INTERRUPT_LINE, &byte1);
	pci_read_config_byte(dev, PCI_INTERRUPT_PIN, &byte2);	
	pci_read_config_byte(dev, PCI_MIN_GNT, &byte3);
	pci_read_config_byte(dev, PCI_MAX_LAT, &byte4);
	ITE_DBG("3Ch   %02x %02x %02x %02x\n", byte1, byte2, byte3, byte4);

	pci_read_config_word(dev, IT8212_PCI_IdeIoCONFIG, &word1);
	pci_read_config_byte(dev, IT8212_PCI_IdeDrivingCURRENT, &byte1);	
	pci_read_config_byte(dev, IT8212_PCI_IdeVirtualChannelEXIST, &byte2);
	ITE_DBG("40h   %04x %02x %02x\n", word1, byte1, byte2);

	pci_read_config_dword(dev, IT8212_PCI_PciBurstTHRESHOLD, &dword);
	ITE_DBG("44h   %08x\n", dword);

	pci_read_config_byte(dev, IT8212_PCI_Pll1CONTROL, &byte1);
	pci_read_config_byte(dev, IT8212_PCI_Pll2CONTROL, &byte2);
	ITE_DBG("48h   %02x %02x -- --\n", byte1, byte2);

	pci_read_config_dword(dev, IT8212_PCI_IdeBusSkewCONTROL, &dword);
	ITE_DBG("4Ch   %08x\n", dword);

	pci_read_config_byte(dev, IT8212_PCI_PciModeCONTROL, &byte1);
	pci_read_config_byte(dev, IT8212_PCI_RaidCONTROL, &byte2);
	pci_read_config_word(dev, IT8212_PCI_RaidTransferSectorCOUNT, &word1);
	ITE_DBG("50h   %02x %02x %04x\n", byte1, byte2, word1);

	pci_read_config_byte(dev, IT8212_PCI_PciModePrimPioDmaTIMING, &byte1);
	pci_read_config_byte(dev, IT8212_PCI_PciModePrimDrive0UdmaTIMING, &byte2);	
	pci_read_config_byte(dev, IT8212_PCI_PciModePrimDrive1UdmaTIMING, &byte3);
	ITE_DBG("54h   %02x -- %02x %02x\n", byte1, byte2, byte3);

	pci_read_config_byte(dev, IT8212_PCI_PciModeSecPioDmaTIMING, &byte1);	
	pci_read_config_byte(dev, IT8212_PCI_PciModeSecDrive0UdmaTIMING, &byte2);
	pci_read_config_byte(dev, IT8212_PCI_PciModeSecDrive1UdmaTIMING, &byte3);
	ITE_DBG("58h   %02x -- %02x %02x\n", byte1, byte2, byte3);

	pci_read_config_byte(dev, IT8212_PCI_TestMODE, &byte1);
	pci_read_config_byte(dev, IT8212_PCI_CpuNmiCONTROL, &byte2);	
	pci_read_config_byte(dev, IT8212_PCI_CpuCONTROL, &byte3);
	ITE_DBG("5Ch   %02x %02x %02x --\n", byte1, byte2, byte3);

	ITE_DBG("60h   --------\n");
	ITE_DBG("7Fh   --------\n");

	pci_read_config_byte(dev, IT8212_PCI_CapIDENTIFIER, &byte1);
	pci_read_config_byte(dev, IT8212_PCI_NextItemPOINTER, &byte2);
	pci_read_config_word(dev, IT8212_PCI_PowerManageCAP, &word1);
	ITE_DBG("80h   %02x %02x %04x\n", byte1, byte2, word1);

	pci_read_config_word(dev, IT8212_PCI_PowerManageControlSTATUS, &word1);
	pci_read_config_byte(dev, IT8212_PCI_BridgeSUPPORT, &byte2);
	pci_read_config_byte(dev, IT8212_PCI_DATA, &byte2);
	ITE_DBG("84h   %04x %02x %02x\n", word1, byte1, byte2);
}


/************************************************************************
 * Show it8212 device information
 ************************************************************************/

static char *it8212_deviceInfo(char *buffer, struct pci_dev *dev)
{
	u8 	pcicr;
	u16	iocfg;
	char *p = buffer;

	ITE_KDBG("(%s)\n", dev->name);
 
	pci_read_config_word(dev, IT8212_PCI_IdeIoCONFIG, &iocfg);
	pci_read_config_byte(dev, IT8212_PCI_PciModeCONTROL, &pcicr);

	p += sprintf(p, "-----------------------------------  IT8212  -----------------------------------\n");
	p += sprintf(p, "--------------- Primary Channel ---------------- Secondary Channel -------------\n");
	p += sprintf(p, "                %s                         %s\n",
		(iocfg&0x8000)?"enabled ":"disabled",
		(iocfg&0x4000)?"enabled ":"disabled");
	p += sprintf(p, "Clock :         %s                            %s\n",
		(pcicr&0x02)?  "50Mhz":"66Mhz",
		(pcicr&0x04)?  "50Mhz":"66Mhz");
	p += sprintf(p, "--------------- drive0 --------- drive1 -------- drive0 ---------- drive1 ------\n");
	p += sprintf(p, "Cable Report:    %s              %s             %s               %s\n",					 
		(iocfg&0x0010)?"80p":"40p",(iocfg&0x0020)?"80p":"40p",
		(iocfg&0x0040)?"80p":"40p",(iocfg&0x0080)?"80p":"40p");
	p += sprintf(p, "TransferMode:   %s       %s      %s        %s\n",					 
		(pcicr&0x08)?"MW DMA   ":"Ultra DMA",(pcicr&0x10)?"MW DMA   ":"Ultra DMA",
		(pcicr&0x20)?"MW DMA   ":"Ultra DMA",(pcicr&0x40)?"MW DMA   ":"Ultra DMA");

	return (char *)p;
}


/************************************************************************
 *
 ************************************************************************/

void it8212_showIdeDrive(ide_drive_t *drive)
{
#ifdef IT8212_DEBUG
	struct hd_driveid *id = drive->id;

	ITE_KDBG("(%s)\n", drive->name);
	
	ITE_DBG("disk drive %s info\n", drive->name);
	ITE_DBG("  capability     %02x\n", id->capability);
	ITE_DBG("  tPIO           %02x\n", id->tPIO);
	ITE_DBG("  tDMA           %02x\n", id->tDMA);
	ITE_DBG("  field_valid    %04x\n", id->field_valid);
	ITE_DBG("  dma_1word      %04x\n", id->dma_1word);
	ITE_DBG("  dma_mword      %04x\n", id->dma_mword);
	ITE_DBG("  eide_pio_modes %04x\n", id->eide_pio_modes);
	ITE_DBG("  dma_ultra      %04x\n", id->dma_ultra);
	ITE_DBG("  hw_config      %04x\n", id->hw_config);
#endif
}


/************************************************************************
 *
 ************************************************************************/

void it8212_showIdeHwif(ide_hwif_t *hwif)
{
	ITE_KDBG("(%s)\n", hwif->name);
}


/************************************************************************
 *
 ************************************************************************/

void it8212_showPciDev(struct pci_dev *dev)
{
	ITE_KDBG("(%s)\n", dev->name);
}


/************************************************************************
 *
 ************************************************************************/

void it8212_enableDriveInterrupt(ide_drive_t *drive, byte driveDn)
{
	ITE_KDBG("(%s, %d)\n", drive->name, driveDn);

	OUT_BYTE((driveDn << 4) | 0xa0, IDE_SELECT_REG);
	OUT_BYTE(0x00, IDE_CONTROL_REG);
}


/************************************************************************
 *
 ************************************************************************/

void it8212_disableDriveInterrupt(ide_drive_t *drive, byte driveDn)
{
	ITE_KDBG("(%s, %d)\n", drive->name, driveDn);

	OUT_BYTE((driveDn << 4) | 0xa0, IDE_SELECT_REG);
	OUT_BYTE(0x02, IDE_CONTROL_REG);
}


/************************************************************************
 *
 ************************************************************************/

void it8212_clearDriveIntStatus(ide_drive_t *drive)
{
	u32 busMasterBase = ((ide_hwif_t *)(drive->hwif))->dma_base;

	ITE_KDBG("(%s)\n", drive->name);

	if (busMasterBase)
	{
		outb(inb(busMasterBase + 0x2) | 0x04, busMasterBase + 0x2);

		it8212_enableDriveInterrupt(drive, drive->dn);
	}
}


/************************************************************************
 *
 ************************************************************************/

int it8212_getHsbIndex(u8 value)
{
	int index = (value) ? 7 : 0;

	ITE_KDBG("(%02x)\n", value);

	if (value)
	{
		while ((value & 0x80) == 0)
		{
			value <<= 1;
			index--;
		}
	}

	ITE_DBG(" HSB index %d\n", index);

	return index;
}


/************************************************************************
 *
 ************************************************************************/

void it8212_setDriveTiming(int channel, int drive)
{
	u8 modeControl;
	u8 udmaTiming;
	u8 pioDmaTiming;
	int channelOffset;
	struct pci_dev *dev;
	drive_config_t *driveConfig;

	M_FDBG("ide%d: channel %d\n", drive, channel);

	dev = it8212.device;
	driveConfig = &(it8212.config[channel][drive]);
	channelOffset = 4 * channel;

	pci_read_config_byte(dev, IT8212_PCI_PciModeCONTROL, &modeControl);
	pci_read_config_byte(dev, IT8212_PCI_PciModePrimPioDmaTIMING + channelOffset, &pioDmaTiming);
	pci_read_config_byte(dev, IT8212_PCI_PciModePrimDrive0UdmaTIMING + channelOffset, &udmaTiming);

	if (driveConfig->type == UDMA_MODE)
	{
		udmaTiming = driveConfig->udmaTiming;
		modeControl &= modeControlMaskVerBX[channel][drive][UdmaMaskINDEX];
	}
	else if (driveConfig->type == DMA_MODE)
	{
		pioDmaTiming = driveConfig->pioDmaTiming;
		modeControl |= modeControlMaskVerBX[channel][drive][DmaMaskINDEX];
	}
	else
	{
		pioDmaTiming = driveConfig->pioDmaTiming;
	}

	modeControl &= modeControlMaskVerBX[channel][drive][ClockMaskINDEX];
	modeControl |= driveConfig->clock << (channel + 1);

	pci_write_config_byte(dev, IT8212_PCI_PciModeCONTROL, modeControl);
	pci_write_config_byte(dev, IT8212_PCI_PciModePrimPioDmaTIMING + channelOffset, pioDmaTiming);
	pci_write_config_byte(dev, IT8212_PCI_PciModePrimDrive0UdmaTIMING + channelOffset, udmaTiming);
	pci_write_config_byte(dev, IT8212_PCI_PciModePrimDrive1UdmaTIMING + channelOffset, udmaTiming);

#ifdef IT8212_DEBUG
	it8212_showDriveTiming(channel);
#endif
}


/************************************************************************
 *
 ************************************************************************/

#ifdef M_NH230
int it8212_dma_check(ide_drive_t *drive)
#else
void it8212_dma_check(ide_drive_t *drive)
#endif
{
	int index;
	int driveNum;
	int channelNum;
	struct hd_driveid *id;
	drive_config_t *driveConfig;

	M_FDBG("%s: dn = %d\n", drive->name, drive->dn);

#ifdef IT8212_DEBUG
	it8212_showIdeDrive(drive);
	it8212_showIdeRegisters(drive->hwif);
#endif

	id = drive->id;
	driveNum = drive->dn % NofDRIVE;
	channelNum = drive->dn / NofDRIVE;
	driveConfig = &(it8212.config[channelNum][driveNum]);

	if (id->field_valid & 0x04)		// DMA capable
	{
		if (id->dma_ultra)			// UDMA mode
		{
			index = it8212_getHsbIndex(id->dma_ultra);

#ifdef M_NH230
			// Do not limit the Ultra DMA mode (up to Ultra DMA 6 by IT8212)
  M_FDBG("index=%d: driveconfig: mode = 0x%X, clock = 0x%X, udmaTiming = 0x%X\n", 
  	index, udmaConfigValue[index][ModeINDEX], udmaConfigValue[index][ClockINDEX], 
  	udmaConfigValue[index][TimingINDEX]);
#else
			if (udmaConfigValue[index][ModeINDEX] > XFER_UDMA_2)
			{
				index = XFER_UDMA_2_INDEX;
			}
#endif

			driveConfig->type = UDMA_MODE;
			driveConfig->mode = udmaConfigValue[index][ModeINDEX];
			driveConfig->clock = udmaConfigValue[index][ClockINDEX];
			driveConfig->udmaTiming = udmaConfigValue[index][TimingINDEX];
		}
		else						// MW DMA mode
		{
			index = it8212_getHsbIndex(id->dma_mword);

			driveConfig->type = DMA_MODE;
			driveConfig->mode = dmaConfigValue[index][ModeINDEX];
			driveConfig->clock = dmaConfigValue[index][ClockINDEX];
			driveConfig->pioDmaTiming = dmaConfigValue[index][TimingINDEX];
		}
	}
	else if (id->field_valid & 0x02)	// PIO only
	{
		if (id->eide_pio_modes)
		{
			index = it8212_getHsbIndex(id->eide_pio_modes) + XFER_PIO_3_INDEX;
		}
		else
		{
			index = XFER_PIO_2_INDEX;
		}

		driveConfig->type = PIO_MODE;
		driveConfig->mode = pioConfigValue[index][ModeINDEX];
		driveConfig->clock = pioConfigValue[index][ClockINDEX];
		driveConfig->pioDmaTiming = pioConfigValue[index][TimingINDEX];
	}

#if 0	// force to PIO4 mode for testing
	index = XFER_PIO_4_INDEX;

	driveConfig->type = PIO_MODE;
	driveConfig->mode = pioConfigValue[index][ModeINDEX];
	driveConfig->clock = pioConfigValue[index][ClockINDEX];
	driveConfig->pioDmaTiming = pioConfigValue[index][TimingINDEX];
#endif

#if 0	// force to MWDMA mode for testing
	index = XFER_MW_DMA_2_INDEX;

	driveConfig->type = DMA_MODE;
	driveConfig->mode = dmaConfigValue[index][ModeINDEX];
	driveConfig->clock = dmaConfigValue[index][ClockINDEX];
	driveConfig->pioDmaTiming = dmaConfigValue[index][TimingINDEX];
#endif

#if 0	// force to UDMA mode for testing
	index = XFER_UDMA_2_INDEX;

	driveConfig->type = UDMA_MODE;
	driveConfig->mode = udmaConfigValue[index][ModeINDEX];
	driveConfig->clock = udmaConfigValue[index][ClockINDEX];
	driveConfig->udmaTiming = udmaConfigValue[index][TimingINDEX];
#endif

	ITE_DBG("type		%02x\n", driveConfig->type);
	ITE_DBG("mode		%02x\n", driveConfig->mode);
	ITE_DBG("clock		%02x\n", driveConfig->clock);
	ITE_DBG("udmaTiming	%02x\n", driveConfig->udmaTiming);
	ITE_DBG("pioDmaTiming	%02x\n", driveConfig->pioDmaTiming);

	it8212_setDriveTiming(channelNum, driveNum);

	ide_config_drive_speed(drive, driveConfig->mode);

	it8212_clearDriveIntStatus(drive);
	
#ifdef M_NH230
	if ( !(HWIF(drive)->udma_four) )
	  HWIF(drive)->udma_four = ata66_it8212(HWIF(drive));
	
	M_FDBG("HWIF(drive)->udma_four = %d\n", HWIF(drive)->udma_four);
#endif	
	
#ifdef M_NH230
	if ( drive->autodma && (driveConfig->type != PIO_MODE) ) {
	  M_FDBG("IDE is set to DMA mode\n");
	  return HWIF(drive)->ide_dma_on(drive);           // enable DMA on an IDE controller
	}
	else {
	  M_FDBG("IDE is set to PIO mode\n");
	  return HWIF(drive)->ide_dma_off_quietly(drive);  // disable DMA as PIO mode
	}
#endif
}


/************************************************************************
 *
 ************************************************************************/

void it8212_initIdeRegisters(ide_hwif_t *hwif)
{
	u8 status;
	u32 busMasterBase = hwif->dma_base;

	ITE_KDBG("(%s)\n", hwif->name);

	if (busMasterBase)
	{

//		outb(0x60, busMasterBase + 0x2);
		outb(0x64, busMasterBase + 0x2);

//		outb(0x01, busMasterBase + 0x0);

		if ( (status = inb(busMasterBase + 2)) != 0x60)
		{
			ITE_DBG("Can't set Drive0/1 DMA -- 02h  %02x\n\n", status);
		}

//		it8212_disableDriveInterrupt(&(hwif->drives[1]), 1);
//		it8212_disableDriveInterrupt(&(hwif->drives[0]), 0);
	}
	else
	{
		ITE_DBG("Can't turn on DMA transfer mode\n");
	}
}


/************************************************************************
 *
 ************************************************************************/

void it8212_initPciRegisters(struct pci_dev *dev)
{
	ITE_KDBG("(%s)\n", dev->name);

	pci_write_config_byte(dev, IT8212_PCI_CpuCONTROL, 0x01);
	pci_write_config_byte(dev, IT8212_PCI_PciModeCONTROL, 0x00);
	pci_write_config_word(dev, PCI_COMMAND, 0x0047);
	pci_write_config_word(dev, IT8212_PCI_IdeIoCONFIG, 0xA0F3);
	pci_write_config_dword(dev, IT8212_PCI_IdeBusSkewCONTROL, 0x02040204);
	pci_write_config_byte(dev, IT8212_PCI_IdeDrivingCURRENT, 0x36);
#ifdef M_NH230
	// Hank Chu : 06/14/2005
	//   For NH-230, do not set PCI_LATENCY_TIMER to 0x00.
	//   Otherwise, Genesys GL880S UHCI (USB 1.1) will fail if IDE H.D active.
	////pci_write_config_byte(dev, PCI_LATENCY_TIMER, 0x40);
	pci_write_config_byte(dev, PCI_LATENCY_TIMER, 0x80);
#else
	pci_write_config_byte(dev, PCI_LATENCY_TIMER, 0x00);
#endif
}


/************************************************************************
 *
 ************************************************************************/

void it8212_init8212Channel(ide_hwif_t *hwif)
{
	byte channel = hwif->channel;

	ITE_KDBG("(%s) ch%d\n", hwif->name, channel);

	it8212.hwif[channel] = hwif;
	it8212.drive[channel][0] = &(hwif->drives[0]);
	it8212.drive[channel][1] = &(hwif->drives[1]);

	it8212_initIdeRegisters(hwif);

	it8212_setDriveTiming(channel, DRIVE1);
	it8212_setDriveTiming(channel, DRIVE0);

#ifdef IT8212_DEBUG
	it8212_showIdeRegisters(hwif);
	it8212_showIdeHwif(hwif);
#endif
}


/************************************************************************
 *
 ************************************************************************/

void it8212_init8212(struct pci_dev *dev)
{
  int i, j;
  drive_config_t *drive;

  M_FDBG("%s: initiating parameters of drive and PCI Configuration Registers....\n", dev->name);

  it8212.device = dev;

  // Loop to initialize the parameters of drives
  for (i = 0; i < NofCHANNEL; i++) {
    for (j = 0; j < NofDRIVE; j++) {
      drive = &(it8212.config[i][j]);
		
      drive->type = UDMA_MODE;
      drive->mode = XFER_UDMA_2;
      drive->clock = CLOCK66;
      drive->udmaTiming = udmaConfigValue[XFER_UDMA_2_INDEX][TimingINDEX];
      drive->pioDmaTiming = pioConfigValue[XFER_PIO_3_INDEX][TimingINDEX];
    }
  }

  // Initialize the PCI Configuration Registers
  it8212_initPciRegisters(dev);

#ifdef IT8212_DEBUG
  it8212_showPciRegisters(dev);
  it8212_showPciDev(dev);
#endif
}


/************************************************************************
 * IT8212 get driver info function
 ************************************************************************/

int it8212_get_info(char *buffer, char **addr, off_t offset, int count)
{
	char *p;

	p = it8212_deviceInfo(buffer, it8212.device);

	return (p - buffer);	/* => must be less than 4k! */
}


/************************************************************************
 * 
 ************************************************************************/
#ifdef M_NH230
#else
int it8212_dmaproc(ide_dma_action_t func, ide_drive_t *drive)
{
	ITE_KDBG("(%s, %d)\n", drive->name, func);

	switch (func) 
    {
		case ide_dma_check:
			it8212_dma_check(drive);
			break;
     
		case ide_dma_timeout:
			if (HWIF(drive)->resetproc != NULL)
			{
				HWIF(drive)->resetproc(drive);
			}
			break;

		default:
			break;
	}

	return ide_dmaproc(func, drive);	/* Use standard DMA stuff	*/
}
#endif


#ifdef M_NH230
int it8212_ide_dma_timeout (ide_drive_t *drive)
{
  if (HWIF(drive)->resetproc != NULL)
    HWIF(drive)->resetproc(drive);
    
  return (0);
}
#endif


/************************************************************************
 * IT8212 reset drive function
 ************************************************************************/

void it8212_reset(ide_drive_t *drive)
{
	ITE_KDBG("(%s)\n", drive->name);

	it8212_initIdeRegisters(drive->hwif);
}


/************************************************************************
 * IDE init function
 ************************************************************************/

void __init ide_init_it8212(ide_hwif_t *hwif)
{
	ITE_KDBG("(%s)\n", hwif->name);

#ifdef M_HANK_PATCH	
  hwif->tuneproc = &it8212_tuneproc;      // routine to tune PIO mode for drives  
  hwif->speedproc = &it8212_speedproc; // routine to retune DMA modes for drives
#endif

	hwif->chipset = ide_it8212;
	hwif->resetproc = &it8212_reset;

	if (hwif->dma_base) {
#ifdef M_NH230
  #ifdef M_HANK_PATCH
                // Please refer to the ide_config_drive_speed() in ide-iops.c file.
  		M_DBG("%s: DMA is supported\n", __FUNCTION__);
		hwif->atapi_dma = 1;        // host support atapi dma
		hwif->ultra_mask = 0x7f;    // IT8212 up to Ultra DMA 6 (0x40 | 0x20 | ....)
		hwif->mwdma_mask = 0x07;
		hwif->swdma_mask = 0x07;
		
		if (!noautodma)
			hwif->autodma = 1;
		hwif->drives[0].autodma = hwif->autodma;
		hwif->drives[1].autodma = hwif->autodma;
  #endif
		hwif->ide_dma_check = &it8212_dma_check;
		hwif->ide_dma_timeout = &it8212_ide_dma_timeout;
#else
 		hwif->dmaproc = &it8212_dmaproc;
#endif
		hwif->autodma = 1;
	}
	else {
		hwif->drives[0].autotune = 1;
		hwif->drives[1].autotune = 1;
		hwif->autodma = 0;
    	}

	it8212_init8212Channel(hwif);
}


/************************************************************************
 * ATA init function
 ************************************************************************/

//
// Subject :
//   Determine whether is it under ATA-66 or higher transfer rate
//
unsigned int __init ata66_it8212(ide_hwif_t *hwif)
{
  M_FDBG("called\n");
	
#ifdef M_NH230
  #define M_UDMA56_SELECT_BITMASK  0x080  // bit[7] : 1 if UDMA5/6 selected
  u8 c_byte = 0;
  pci_read_config_byte(hwif->pci_dev, IT8212_PCI_PciModePrimDrive0UdmaTIMING, &c_byte);

  M_FDBG("UDMA5/6 is %s for %s, PCI Register=0x%X\n",
    (c_byte & M_UDMA56_SELECT_BITMASK) ? "enabled" : "disabled", hwif->name, c_byte);
		
  return ((c_byte & M_UDMA56_SELECT_BITMASK) ? 1 : 0);
#else
	u16 mask;
	u16 iocfg;;

	ITE_KDBG("(%s)\n", hwif->name);

	mask = (hwif->channel) ? (1 << 3) : (1 << 2);
	pci_read_config_word(hwif->pci_dev, 0x40, &iocfg);
	
	M_FDBG("PCI : IDE I/O Configuration Register = 0x%X\n", iocfg);
	iocfg &= mask;

	ITE_DBG("CBLID signal %d\n", iocfg);
 
	return ((iocfg) ? 0 : 1);
#endif
}


/************************************************************************
 * PCI init function
 ************************************************************************/

unsigned int __init pci_init_it8212(struct pci_dev *dev, const char *name)
{
	
	DECLARE_WAIT_QUEUE_HEAD(wq);

	ITE_KDBG("(%s %02x:%02x)\n", name, dev->bus->number, dev->devfn);

#ifdef M_NH230
	// 04/15/2004  Hank Chu :
	//    Because hard disk is slow to ready for access after cold boot, 
	//    we need to delay 4 ~ 6 seconds (said by ITE) to wait for it ready, and then start to initialize IT8212.
	//    In order to make sure it is ready, we delay 8 seconds here.
	//
	u16 w_loop = 0;
	printk("IT8212: wait 8 seconds for hard disk ready before initialize IT8212 after cold boot.\n");
	for (w_loop = 0; w_loop < 8; w_loop++) {   // loop 8 times (1000ms per loop) == 8 seconds
	  printk(".");
	  
	  interruptible_sleep_on_timeout (&wq, HZ);	
	  // LED no flash !!	
	  //NH230_GPIO_LED_STATUS_ON();              // turn on status LED
	  //mdelay(500);
	  //NH230_GPIO_LED_STATUS_OFF();             // turn off status LED
	  //mdelay(500);
	}
	printk(" Done\n");
  
  #ifdef M_IDE_SLEEP_MODE
  	// Create a kernel thread to monitor the IDE idle to sleep mode.
  	if (kernel_thread (fn_IDE_sleep_mode_thread, 0, CLONE_FS | CLONE_FILES) < 0)
          printk (KERN_WARNING "IDE: unable to create kernel thread for sleep mode.\n");     		
  #endif
#endif
	
	it8212_proc = 1;

#ifdef M_NH230
	// For Linux kernel v2.4.21, no need to use it.
#else
	it8212_display_info = &it8212_get_info;
#endif

	it8212_init8212(dev);

	return dev->irq;
}

#ifdef M_NH230
static void __init init_setup_it8212 (struct pci_dev *dev, ide_pci_device_t *d)
{
  ide_setup_pci_device(dev, d);
}


static void __init init_dma_it8212 (ide_hwif_t *hwif, unsigned long dmabase)
{
	ide_setup_dma(hwif, dmabase, 8);
}


extern void ide_setup_pci_device(struct pci_dev *, ide_pci_device_t *);

static int __devinit it8212_init_one(struct pci_dev *dev, const struct pci_device_id *id)
{
  ide_pci_device_t *d = &it8212_chipsets[id->driver_data];
  
#ifdef M_NH230
  if (dev->device != d->device) {       // PCI Device ID match ?
    printk("%s: dev->vendor/device = 0x%X/0x%X, d->vendor/device = 0x%X/0x%X, id->driver_data =%d\n",
            __FUNCTION__, dev->vendor, dev->device, d->vendor, d->device, id->driver_data);
    BUG();                              // make system crash as not match
  }
#else
  if ((!(PCI_FUNC(dev->devfn) & 1) || (!((dev->class >> 8) == PCI_CLASS_STORAGE_IDE)))) {
    printk("IT821x: error: PCI device is not IDE.\n");
    printk("IT821x: error: PCI device : Device ID[0x%X], Vendor ID[0x%X], devfn[0x%X], class[0x%X]\n",
             dev->vendor, dev->device, dev->devfn, dev->class);
    return 1;                          // error
  }
#endif
                
  d->init_setup(dev, d);
  MOD_INC_USE_COUNT;
  return 0;                            // OK
}

//
// The last parameter : pci_dev_id.driver_data : 
//  It mean the index of it8212_pci_tbl[] & it8212_chipsets[] arrays.
//
static struct pci_device_id it8212_pci_tbl[] __devinitdata = {
	{ PCI_VENDOR_ID_ITE, PCI_DEVICE_ID_ITE_IT8212, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0},
	{ PCI_VENDOR_ID_ITE, PCI_DEVICE_ID_ITE_IT8211, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 1},
	{ 0, },
};


static struct pci_driver driver = {
	.name		= "IT8212/8211 IDE",
	.id_table	= it8212_pci_tbl,
	.probe		= it8212_init_one,
};


static int it8212_ide_init(void)
{
	return ide_pci_register_driver(&driver);
}


static void it8212_ide_exit(void)
{
	ide_pci_unregister_driver(&driver);
}

module_init(it8212_ide_init);
module_exit(it8212_ide_exit);

MODULE_AUTHOR("liangyao.li@iteusa.com");
MODULE_DESCRIPTION("PCI driver module for ITE IT8212/8211 IDE");
MODULE_LICENSE("GPL");

EXPORT_NO_SYMBOLS;


#endif  // end of #ifdef {M_NH230}
