/*
=========================================================================
 r8169.c: A RealTek RTL8169s/8110s Gigabit Ethernet driver for Linux kernel 2.4.x.
 --------------------------------------------------------------------

 History:
 Feb  4 2002	- created initially by ShuChen <shuchen@realtek.com.tw>.
 May 20 2002	- Add link status force-mode and TBI mode support.
=========================================================================

RTL8169_VERSION "1.1"	<2002/10/4>

	The bit4:0 of MII register 4 is called "selector field", and have to be
	00001b to indicate support of IEEE std 802.3 during NWay process of
	exchanging Link Code Word (FLP).

RTL8169_VERSION "1.2"	<2003/6/17>
	Update driver module name.
	Modify ISR.
        Add chip mcfg.

RTL8169_VERSION "1.3"	<2003/6/20>
        Add chip pcfg.
	Add priv->phy_timer_t, rtl8169_phy_timer_t_handler()
	Add rtl8169_hw_PHY_config()
	Add rtl8169_hw_PHY_reset()

RTL8169_VERSION "1.4"	<2003/7/14>
	Add tx_bytes, rx_bytes.

RTL8169_VERSION "1.5"	<2003/7/18>
	Set 0x0000 to PHY at offset 0x0b.
	Modify chip mcfg, pcfg
	Force media for multiple card.
RTL8169_VERSION "1.6"	<2003/8/25>
	Modify receive data buffer.

RTL8169_VERSION "1.7"	<2003/9/18>
	Add Jumbo Frame support.

RTL8169_VERSION "1.8"	<2003/10/21>
	Performance and CPU Utilizaion Enhancement.

RTL8169_VERSION "1.9"	<2003/12/29>
	Enable Tx/Rx flow control.

RTL8169_VERSION "2.0"	<2004/03/26>
	Beta version.
	Support for linux 2.6.x

RTL8169_VERSION "2.1"	<2004/07/05>
	Modify parameters.

RTL8169_VERSION "2.2"	<2004/08/09>
	Add.pci_dma_sync_single.
	Add pci_alloc_consistent()/pci_free_consistent().
	Revise parameters.
	Recognize our interrupt for linux 2.6.x.
*/
//
// History :
//   v2.2A  10/15/2004  Hank Chu
//     1) Modify from Realtek RTL8169S Linux driver (v2.2) for Intel IXP42x
//        under Snapgear uClinux.
//     2) Use the macro M_NH230 to enable the patches.
//     3) Modify the "rtl8169_pci_driver" structure to fix linker error.
//    
//   v2.2B  01/14/2005  Hank Chu
//     1) Modify the display of LEDs according to NH-230 H/W Design Specification.
//        1-1) In the rtl8169_hw_start() routine, read/write the Config1 register.
//
//   v2.2C  01/24/2005  Hank Chu
//     1) Add the "r8169_ioctl.c" file to support IOCTL for Ethernet driver.
//        Turn on the RTL8169_IOCTL_SUPPORT macro to enable it.
//     2) Add the "RTL8169_CHECKSUM_BY_HW" macro to support checksum by H/W.
//        But the "checksum by hardware" function cannot co-work with Jumbo frame.
//     3) Assign the MAC addresses (gsz_eth_address) parsed by the kernel command 
//        line into driver and write back to IDR0~5 register.
//
//   v2.2-R1 03/29/2005  Hank Chu
//     1) Add the M_HANK_PROFILE macro to use counters to tune performance.
//     2) Support the NAPI driver interface by the CONFIG_R8169_NAPI macro.
//     3) Optimize the performance of driver.
//
//   v2.2-R2 04/07/2005  Hank Chu
//     1) Realtek sent us a version of Linux driver for RTL8169S on 04/06/2005.
//           rtl8169_20050225_iperf.zip
//
//        And they said it can improve throughput.
//        After tested, it does not improve throughput, and same as 
//           linux-8169(220).zip
//
//        But anyway, we add the pacthes into this driver but disable them.
//        The codes of patches are separated by M_REALTEK_20050225_IPERF macro.
//        Realtek insists to enable this patches.
//
//
//     2) About the "Checksum by hardware" function : (RTL8169_CHECKSUM_BY_HW)
//        Dynamically enable/disable it by :
//          - If MTU size is <= 1500, enable the "Checksum by hardware" function.
//          - If MTU size > 1500 (Jumbo frame enabled), disable it.
//
//     3) Add the M_RX_SKB_POOL macro :
//          Pre-allocate a SKB pool with SKBs for rtl8169_rx_interrupt() to use.
//          It save the time to memory allocation of SKBs in runtime.
//          ==> Not completed yet as it caused system crashed.
//
//   v2.2-R3  04/13/2005  Hank Chu
//     1) Increase the Tx queue length of driver. 
//           dev->tx_queue_len = 1024;
//
//        Otherwise, under heavy Tx traffic (like FTP get file),
//        the Tx descriptor ring will be full and stop queue temporarily.
//
//        The rtl8169_start_xmit() routine check whether Tx descriptor ring 
//        is full or not ? 
//        If yes, it will call "netif_stop_queue()" to sto queue.
//
//     2) Add the M_TX_POLLING_MODE macro.
//
//   v2.2-R4  05/02/2005  Hank Chu
//     1) Add the M_DYNAMICALLY_TUNE_INTRMITIGATE_REGISTER macro to 
//        tune the Intrmitigate Register dynamically according to the speed.
//        For 10Mbps,  up to 14880 packets/second
//        For 100Mbps, up to 148800 packets/second
//        For 1Gbps,   up to 1488000 packets/second
//
//        * Do not define this macro :
//        *   If defined, sometimes, "FTP get file" will fail as
//              "Can't build data connection" error.
//
//   v2.2-R5  09/05/2005  Hank Chu
//     1) Bug fixed :
//        Problem :
//          If LAN cable disconnects and power on NH-230, and plug in LAN cable
//          when smbd is decompressed during system boot up, 
//          sometimes, the RTL8169 cannot link up (LAN LED is off).          
//        Reason :
//          If RTL8169S and Gigabit capable :
//            If link down more than 12 by timer checking, the PHY must be reset.
//        Solution :
//          In the rtl8169_timer_handler() routine :
//            Should not set "priv->phy_link_down_cnt" to 0 here.
//            It should be reset in rtl8169_phy_timer_t_handler() routine.
//
//
//
//
//  Notes :
//    1) Currently, this driver support only 1 driver instaqnce.
//       It means it support only 1 RTL8169S PCI controller.
//       Because the following limitation :
//        1-1) extern u8 gsz_eth_address[];
//        1-2) u32 gdw_mtu = DEFAULT_MTU;
//        1-3) and etc.
//
//       If you want to support multiple RTL8169S PCI controllers, please
//       modify the global variables above to be an array for all controllers.
//
//    2) According to Realtek comment, the Jumbo frame function cannot co-work
//       with Checksum by hardware function.
// 
//    3) We use the M_ORIGINA_CODES macro to disable the original codes as we
//       do not use it.
//
//    4) The memory used by the driver :
//       If 1 driver instance, it majorly takes around :
//        A) ( NUM_TX_DESC * sizeof(TxDesc) ) + 256
//        B) ( NUM_RX_DESC * sizeof(RxDesc) ) + 256
//        C) ( NUM_RX_DESC * MAX_RX_SKBDATA_SIZE )
//
//       So,for the maximum :
//       4-1) Jumbo frame 7KB, and NUM_TX_DESC/NUM_RX_DESC is 64/64 :       
//            It takes around : around 441 KB or more
//
//       4-2) Jumbo frame 7KB, and NUM_TX_DESC/NUM_RX_DESC is 96/96 :
//            It takes around : around 661 KB or more
//
//       4-3) Jumbo frame 7KB, and NUM_TX_DESC/NUM_RX_DESC is 256/256 :
//            It takes around : around 1.722 MB or more
//
//       The comparison of performance :
//          256/256 is more than 64/64 around several Mbps only for TCP/IP.
//          96/96 is more than 256/256 very few Mbps only for TCP/IP.
//
//       Hank Chu : 04/08/2005
//          So, we choose 96/96 for the NUM_TX_DESC and NUM_RX_DESC.
//
//
//
//
//
//


#include <linux/module.h>
#include <linux/pci.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/delay.h>
#include <linux/version.h>

#include <linux/timer.h>
#include <linux/init.h>
#include <linux/in.h>                  // for IPPROTO_TCP & IPPROTO_UDP
#include <linux/ip.h>                  // for the "struct iphdr{}"
#include <linux/tcp.h>

//sp/
//#define SP_R8169_N_DEBUG

#if      defined(SP_R8169_N_DEBUG)
	#ifndef	DEBUG
	//# define DEBUG
	#endif
	# define  sp_r8169Trace(format,arg...)		printk(KERN_INFO __FILE__": "__FUNCTION__ ": " format "\n" , ## arg)
#else
	# define  sp_r8169Trace(format, args...)
#endif

#ifdef M_NH230
//#undef M_NH230

unsigned long g_ioaddr;
struct net_device *g_dev;
struct rtl8169_private *g_priv;

#endif
#ifdef M_NH230
  #define RTL8169_VERSION "2.2-R5 (09/05/2005)"
#else
  #define RTL8169_VERSION "2.2"
#endif

#define MODULENAME "RTL8169s/8110s"
#define RTL8169_DRIVER_NAME   MODULENAME " Gigabit Ethernet driver " RTL8169_VERSION
#define PFX MODULENAME ": "


#undef RTL8169_DEBUG             // define it to enable debugging messages by Realtek

//#define M_HANK_DEBUG           // define it to enable debugging messages
//#define M_HANK_PROFILE         // define it to use counters to tune performance

#ifdef M_HANK_DEBUG
  #define M_DBG(format, arg...)		printk(format , ## arg)
  #define M_FDBG(format, arg...)	printk("%s: " format, __FUNCTION__, ## arg)
  
  void dbg_dump_CPU_reg (void);
  void dbg_dump_PCI_reg (struct net_device *dev);
  void dbg_dump_reg (struct net_device *dev);
  
  #define M_DUMP_CPU_REG()      dbg_dump_CPU_reg()
  #define M_DUMP_PCI_REG(dev)   dbg_dump_PCI_reg(dev)
  #define M_DUMP_REG(dev)       dbg_dump_reg(dev)
#else
  #define M_DBG(format, arg...)
  #define M_FDBG(format, arg...)
  
  #define M_DUMP_CPU_REG()
  #define M_DUMP_PCI_REG(dev)
  #define M_DUMP_REG(dev)
#endif

#ifdef M_HANK_PROFILE
  u32 gdw_eth_tick = 0;                 // previous time of Eth interrupt triggered
  u32 gdw_eth_rx_bytes = 0;             // previous total bytes of packets received
  u32 gdw_eth_tx_bytes = 0;             // previous total bytes of packets sent
  
  u32 gdw_eth_rx_pkts = 0;              // previous total NO# of packets received
  u32 gdw_eth_tx_pkts = 0;              // previous total NO# of packets sent

  #define M_INC_COUNTER(n) (n += 1)     // add profile counter with 1
 
  #define M_PRINT(format, arg...)	printk(format , ## arg)
  #define M_FPRINT(format, arg...)	printk("%s: " format, __FUNCTION__, ## arg)
#else
  #define M_INC_COUNTER(n)
  
  #define M_PRINT(format, arg...)
  #define M_FPRINT(format, arg...)
#endif


#ifdef M_NH230
	// 6-byte Ethernet address parsed from kernel command line
        extern u8 gsz_eth_address[];
	
        // Larger MAX_JUMBO_FRAME_MTU will reduce around 1/4 throughput as 
        //   allocate bigger memory runtime.  So, we set it exactly to :
        //     MTU + Ethernet Header size + (VLAN header size ?)
        //
        u32 gdw_mtu = 0;
        
        // Enable the NAPI for driver
        #define CONFIG_R8169_NAPI
        
	// Enable the Jumbo frame and H/W flow control.
	#define RTL8169_JUMBO_FRAME_SUPPORT
        #define RTL8169_HW_FLOW_CONTROL_SUPPORT

	#define RTL8169_IOCTL_SUPPORT     // define it to use I/O control of driver
        #undef  RTL8169_DYNAMIC_CONTROL   // undefine it to remove dynamic control
        #undef  RTL8169_USE_IO            // undefine it to use PCI Memory Space
        
        // Define it to support Checksum by hardware and enable/disable runtime.
	//  (But it cannot co-exist with Jumbo Frame : MTU size > 1500)
        #define RTL8169_CHECKSUM_BY_HW

	// Define it to handle Tx by polling timer, not by interrupt
	#define M_TX_POLLING_MODE
		
	// Realtek sent us patches on 04/06/2005 and said it can improve throughput.
	// After tested, it seems that it does not improve anything.
	// But Realtek insists to enable the patches.
	#define  M_REALTEK_20050225_IPERF

	// Dynamically tuning the Intrmitigate Register according to the speed.
	#undef  M_DYNAMICALLY_TUNE_INTRMITIGATE_REGISTER
	
	// Undefine it to disable the SKB pool for Rx as system crashed
	#undef  M_RX_SKB_POOL
#else
  #undef RTL8169_JUMBO_FRAME_SUPPORT
  #undef	RTL8169_HW_FLOW_CONTROL_SUPPORT


  #undef RTL8169_IOCTL_SUPPORT
  #undef RTL8169_DYNAMIC_CONTROL
  #define RTL8169_USE_IO
#endif

#ifdef CONFIG_R8169_NAPI
  #include <linux/tqueue.h>
  #define work_struct		tq_struct
  #define INIT_WORK		INIT_TQUEUE
  #define schedule_work		schedule_task
  #define flush_scheduled_work	flush_scheduled_tasks
  #define PREPARE_WORK		PREPARE_TQUEUE

  #define DMA_64BIT_MASK     0xffffffffffffffffULL
  #define DMA_32BIT_MASK     0x00000000ffffffffULL
#endif

#define TX_BUFFS_AVAIL(tp) \
	(tp->dirty_tx + NUM_TX_DESC - tp->cur_tx - 1)

#ifdef CONFIG_R8169_NAPI
  #define rtl8169_rx_skb			netif_receive_skb
  #define rtl8169_rx_hwaccel_skb		vlan_hwaccel_rx
  #define rtl8169_rx_quota(count, quota)	min(count, quota)
#else
  #define rtl8169_rx_skb			netif_rx
  #define rtl8169_rx_hwaccel_skb		vlan_hwaccel_receive_skb
  #define rtl8169_rx_quota(count, quota)	count
#endif


#ifdef RTL8169_DEBUG
	#define assert(expr) \
        	if(!(expr)) { printk( "Assertion failed! %s,%s,%s,line=%d\n", #expr,__FILE__,__FUNCTION__,__LINE__); }
	#define DBG_PRINT( fmt, args...)   printk("r8169: " fmt, ## args);
#else
	#define assert(expr) do {} while (0)
	#define DBG_PRINT( fmt, args...)   ;
#endif	// end of #ifdef RTL8169_DEBUG


/* media options */
#define MAX_UNITS 8
static int media[MAX_UNITS] = {-1, -1, -1, -1, -1, -1, -1, -1};

/* Maximum events (Rx packets, etc.) to handle at each interrupt. */
static int max_interrupt_work = 20;

/* Maximum number of multicast addresses to filter (vs. Rx-all-multicast).
   The RTL chips use a 64 element hash table based on the Ethernet CRC.  */
static int multicast_filter_limit = 32;

/* MAC address length*/
#define MAC_ADDR_LEN        6

#define RX_FIFO_THRESH      7       /* 7 means NO threshold, Rx buffer level before first PCI xfer.  */

#define RX_DMA_BURST        7       /* Maximum PCI burst, '6' is 1024 */
#define TX_DMA_BURST        7       /* Maximum PCI burst, '6' is 1024 */

#define ETTh                0x3F    /* 0x3F means NO threshold */

#define ETH_HDR_LEN         14
#define DEFAULT_MTU         1500
#define DEFAULT_RX_BUF_LEN  1536


#ifdef RTL8169_JUMBO_FRAME_SUPPORT
  #ifdef M_NH230
    #define MAX_JUMBO_FRAME_MTU	  ( gdw_mtu + ETH_HDR_LEN + 8)
    
    // Maximum MTU of Jumbo frame allowed for RTL8169S
    #define MAX_JUMBO_FRAME_MTU_ALLOWED       ( 10000 )
  #else
    #define MAX_JUMBO_FRAME_MTU	( 10000 )
  #endif
  
  #define MAX_RX_SKBDATA_SIZE	( MAX_JUMBO_FRAME_MTU + ETH_HDR_LEN )
#else
  #define MAX_RX_SKBDATA_SIZE 1600
#endif //end #ifdef RTL8169_JUMBO_FRAME_SUPPORT


#define InterFrameGap       0x03    /* 3 means InterFrameGap = the shortest one */


// Be careful as more Tx/Rx descriptors will take more free memory.
#ifdef CONFIG_R8169_NAPI
  #ifdef M_REALTEK_20050225_IPERF
      #define NUM_TX_DESC   256 //1024     /* Number of Tx descriptor registers*/
      #define NUM_RX_DESC   96  //1024     /* Number of Rx descriptor registers*/
  #else  
      #define NUM_TX_DESC   256      /* Number of Tx descriptor registers*/
      #define NUM_RX_DESC   96       /* Number of Rx descriptor registers*/
  #endif
  
  #define R8169_NAPI_WEIGHT   NUM_RX_DESC
#else
  #ifdef M_REALTEK_20050225_IPERF
      #define NUM_TX_DESC   256 //1024     /* Number of Tx descriptor registers*/
      #define NUM_RX_DESC   96  //1024     /* Number of Rx descriptor registers*/
  #else
      #define NUM_TX_DESC   64       /* Number of Tx descriptor registers*/
      #define NUM_RX_DESC   64       /* Number of Rx descriptor registers*/
  #endif
#endif


#define RTL_MIN_IO_SIZE     0x80
#define TX_TIMEOUT          (6*HZ)
#define RTL8169_TIMER_EXPIRE_TIME 100  // 100 Jiffies == 1 seconds


#ifdef RTL8169_USE_IO
#define RTL_W8(reg, val8)   outb ((val8), ioaddr + (reg))
#define RTL_W16(reg, val16) outw ((val16), ioaddr + (reg))
#define RTL_W32(reg, val32) outl ((val32), ioaddr + (reg))
#define RTL_R8(reg)         inb (ioaddr + (reg))
#define RTL_R16(reg)        inw (ioaddr + (reg))
#define RTL_R32(reg)        ((unsigned long) inl (ioaddr + (reg)))
#else
/* write/read MMIO register */
#define RTL_W8(reg, val8)   writeb ((val8), ioaddr + (reg))
#define RTL_W16(reg, val16) writew ((val16), ioaddr + (reg))
#define RTL_W32(reg, val32) writel ((val32), ioaddr + (reg))
#define RTL_R8(reg)         readb (ioaddr + (reg))
#define RTL_R16(reg)        readw (ioaddr + (reg))
#define RTL_R32(reg)        ((unsigned long) readl (ioaddr + (reg)))
#endif

#define MCFG_METHOD_1		0x01
#define MCFG_METHOD_2		0x02
#define MCFG_METHOD_3		0x03
#define MCFG_METHOD_4		0x04

#define PCFG_METHOD_1		0x01	//PHY Reg 0x03 bit0-3 == 0x0000
#define PCFG_METHOD_2		0x02	//PHY Reg 0x03 bit0-3 == 0x0001
#define PCFG_METHOD_3		0x03	//PHY Reg 0x03 bit0-3 == 0x0002


#ifdef RTL8169_DYNAMIC_CONTROL
#include "r8169_callback.h"
#endif  //end #ifdef RTL8169_DYNAMIC_CONTROL


const static struct {
	const char *name;
	u8 mcfg;		 /* depend on RTL8169 docs */
	u32 RxConfigMask; 	/* should clear the bits supported by this chip */
} rtl_chip_info[] = {
	{ "RTL8169",  MCFG_METHOD_1,  0xff7e1880 },
	{ "RTL8169s/8110s",  MCFG_METHOD_2,  0xff7e1880 },
	{ "RTL8169s/8110s",  MCFG_METHOD_3,  0xff7e1880 },
};


static struct pci_device_id rtl8169_pci_tbl[] __devinitdata = {
	{ 0x10ec, 0x8169, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0 },
	{0,},
};


MODULE_DEVICE_TABLE (pci, rtl8169_pci_tbl);


enum RTL8169_registers {
	MAC0 = 0x0,
	MAR0 = 0x8,
	TxDescStartAddr	= 0x20,
	TxHDescStartAddr= 0x28,
	FLASH	= 0x30,
	ERSR	= 0x36,
	ChipCmd	= 0x37,
	TxPoll	= 0x38,
	IntrMask = 0x3C,
	IntrStatus = 0x3E,
	TxConfig = 0x40,
	RxConfig = 0x44,
	RxMissed = 0x4C,
	Cfg9346 = 0x50,
	Config0	= 0x51,
	Config1	= 0x52,
	Config2	= 0x53,
	Config3	= 0x54,
	Config4	= 0x55,
	Config5	= 0x56,
	MultiIntr = 0x5C,
	PHYAR	= 0x60,
	TBICSR	= 0x64,
	TBI_ANAR = 0x68,
	TBI_LPAR = 0x6A,
	PHYstatus = 0x6C,
	RxMaxSize = 0xDA,
	CPlusCmd = 0xE0,
#ifdef M_DYNAMICALLY_TUNE_INTRMITIGATE_REGISTER
	// IntrMitigate register :
	//   bit[0:3] : RxPackets
	//   bit[4:7] : RxTimer
	//   bit[8:11] : TxPackets
	//   bit[12:15] : TxTimer
	//
	IntrMitigate = 0xE2,
#endif
	RxDescStartAddr	= 0xE4,
	ETThReg	= 0xEC,
	FuncEvent	= 0xF0,
	FuncEventMask	= 0xF4,
	FuncPresetState	= 0xF8,
	FuncForceEvent	= 0xFC,
};

enum RTL8169_register_content {
	/*InterruptStatusBits*/
	SYSErr 		= 0x8000,
	PCSTimeout	= 0x4000,
	SWInt		= 0x0100,
	TxDescUnavail	= 0x80,
	RxFIFOOver 	= 0x40,
	LinkChg 	= 0x20,
	RxOverflow 	= 0x10,
	TxErr 	= 0x08,
	TxOK 	= 0x04,
	RxErr 	= 0x02,
	RxOK 	= 0x01,

	/*RxStatusDesc*/
	RxRES = 0x00200000,
	RxCRC = 0x00080000,
	RxRUNT= 0x00100000,
	RxRWT = 0x00400000,

	/*ChipCmdBits*/
	CmdReset = 0x10,
	CmdRxEnb = 0x08,
	CmdTxEnb = 0x04,
	RxBufEmpty = 0x01,

	/*Cfg9346Bits*/
	Cfg9346_Lock = 0x00,
	Cfg9346_Unlock = 0xC0,

	/*rx_mode_bits*/
	AcceptErr = 0x20,
	AcceptRunt = 0x10,
	AcceptBroadcast = 0x08,
	AcceptMulticast = 0x04,
	AcceptMyPhys = 0x02,
	AcceptAllPhys = 0x01,

	/*RxConfigBits*/
	RxCfgFIFOShift = 13,
	RxCfgDMAShift = 8,

	/*TxConfigBits*/
	TxInterFrameGapShift = 24,
	TxDMAShift = 8,

	/* CPlusCmd p.31 */
	RxVlan		= (1 << 6),
	RxChkSum	= (1 << 5),
	PCIDAC		= (1 << 4),
	PCIMulRW	= (1 << 3),
	
	/*rtl8169_PHYstatus*/
	TBI_Enable	= 0x80,
	TxFlowCtrl	= 0x40,
	RxFlowCtrl	= 0x20,
	_1000bpsF	= 0x10,
	_100bps		= 0x08,
	_10bps		= 0x04,
	LinkStatus	= 0x02,
	FullDup		= 0x01,

	/*GIGABIT_PHY_registers*/
	PHY_CTRL_REG = 0,
	PHY_STAT_REG = 1,
	PHY_AUTO_NEGO_REG = 4,
	PHY_1000_CTRL_REG = 9,

	/*GIGABIT_PHY_REG_BIT*/
	PHY_Restart_Auto_Nego	= 0x0200,
	PHY_Enable_Auto_Nego	= 0x1000,

	//PHY_STAT_REG = 1;
	PHY_Auto_Neco_Comp	= 0x0020,

	//PHY_AUTO_NEGO_REG = 4;
	PHY_Cap_10_Half		= 0x0020,
	PHY_Cap_10_Full		= 0x0040,
	PHY_Cap_100_Half	= 0x0080,
	PHY_Cap_100_Full	= 0x0100,

	//PHY_1000_CTRL_REG = 9;
#ifdef M_REALTEK_20050225_IPERF
	PHY_Cap_1000_Half       = 0x0100,
#endif
	PHY_Cap_1000_Full	= 0x0200,
	PHY_Cap_PAUSE		= 0x0400,
	PHY_Cap_ASYM_PAUSE	= 0x0800,

	PHY_Cap_Null		= 0x0,

	/*_MediaType*/
	_10_Half	= 0x01,
	_10_Full	= 0x02,
	_100_Half	= 0x04,
	_100_Full	= 0x08,
	_1000_Full	= 0x10,

	/*_TBICSRBit*/
	TBILinkOK 	= 0x02000000,
};

#ifdef M_DYNAMICALLY_TUNE_INTRMITIGATE_REGISTER
  #ifdef M_TX_POLLING_MODE
    #define INTRMITIGATE_REG_1Gbps    0x0001F
    #define INTRMITIGATE_REG_100Mbps  0x0001F
  #else
    #define INTRMITIGATE_REG_1Gbps    0x01F1F 
    #define INTRMITIGATE_REG_100Mbps  0x01F1F
  #endif
  
  // 10Mbps, at least 1 packet/interrupt
  #define INTRMITIGATE_REG_10Mbps   INTRMITIGATE_REG_DEFAULT  
  #define INTRMITIGATE_REG_DEFAULT  0x00000
#endif

enum _DescStatusBit {
	OWNbit	= 0x80000000,
	EORbit	= 0x40000000,
	FSbit	= 0x20000000,
	LSbit	= 0x10000000,
};


#ifdef M_NH230
  #ifdef RTL8169_CHECKSUM_BY_HW
    // C+ Command Register
    #define RxChkSum     0x0020        // Rx Checksum Offload Enable
  
    // Normal Tx Descriptor : the bit definitions
    #define IPCS         0x00040000    // IP checksum offlocad
    #define UDPCS        0x00020000    // UDP checksum offload
    #define TCPCS        0x00010000    // TCP checksum offlocad
  
    // Rx Status Descriptor : the bit definitions
    #define IPFail       0x00010000    // IP checksum failure
    #define UDPFail      0x00008000    // UDP checksum failure
    #define TCPFail      0x00004000    // TCP checksum failure
  
    #define PID1         0x00040000    // Protocol ID1
    #define PID0         0x00020000    // Protocol ID0
    
    #define RX_STATUS_DESC_PID(status)  ( (status >> 17) & 3 )
  
    #define RxProtoTCP   1             // protocol type of packet received is TCP
    #define RxProtoUDP   2             // protocol type of packet received is UDP
    #define RxProtoIP    3             // protocol type of packet received is IP    
  
  
    // 
    // Subject :
    //   Verify whether the packet received with valid TCP/IP or UDP/IP chcksum ?
    // Input :
    //   status  : 32-bit value of Rx Status Descriptor
    // Output :
    //   1  : checksum OK
    //   0  : not
    // 
    static inline unsigned int cp_rx_csum_ok (u32 status)
    {
      unsigned int protocol = RX_STATUS_DESC_PID(status);
	
      if (likely((protocol == RxProtoTCP) && (!(status & TCPFail))))
        return 1;
      else if ((protocol == RxProtoUDP) && (!(status & UDPFail)))
	return 1;
      else if ((protocol == RxProtoIP) && (!(status & IPFail)))
	return 1;
	
      return 0;                         // checksum failure
    }
    
  #endif  // end of #ifdef {RTL8169_CHECKSUM_BY_HW}
#endif    // end of #ifdef {M_NH230}

struct TxDesc {
	u32		status;
	u32		vlan_tag;
	u32		buf_addr;
	u32		buf_Haddr;
};

struct RxDesc {
	u32		status;
	u32		vlan_tag;
	u32		buf_addr;
	u32		buf_Haddr;
};


typedef struct timer_list rt_timer_t;


#ifdef M_RX_SKB_POOL
  #define MAX_ENTRIES_OF_SKB_POOL     (NUM_RX_DESC)
  #define CMD_SKB_POOL_ALLOCATE      0
  #define CMD_SKB_POOL_FREE          1
  
  // The structure deinfes the entry in SKB pool for Rx ( a single-linked list)
  struct st_skb_pool_entry {
    struct st_skb_pool_entry *lpst_next;  // pointer to next entry
    struct sk_buff *lpst_skb;             // pointer to SKB allocated
  };
  
  typedef struct st_skb_pool_entry   ST_SKB_POOL_ENTRY;

  // The SKB pool for Rx is a single-linked list
  typedef struct _tag_skb_pool {
    u32    dw_qty;                        // NO# of available SKBs in pool
    ST_SKB_POOL_ENTRY  *lpst_head;        // pointer to linked list for SKBs
  } ST_SKB_POOL;
  
  int fn_skb_pool_handler(struct net_device *dev, u8 c_cmd);
  
  static void	RXbh(void *data);
#endif

#ifdef M_HANK_PROFILE
  typedef struct _tag_counter_struct {
    u32 dw_int_in;                      // NO# of ISR called
    u32 dw_int_out;                     // NO# of ISR quit
    u32 dw_inpoll_int;                  // NO# of ISR but in polling
    u32 dw_other_int;                   // NO# of other interupts not related to NAPI
    u32 dw_RxOk_int;                    // NO# of Rx OK interrupts
    u32 dw_tx_int_handled;              // NO# of Tx interrupts handled
    u32 dw_RxOverFlow_int;              // NO# of Rx overflow interrupts
    u32 dw_RxFifoOver_int;              // NO# of Rx FIFO Overrun interrupts
    u32 dw_TxOk_int;                    // NO# of Tx OK interrupts
    u32 dw_TxErr_int;                   // NO# of Tx Error interrupts
    u32 dw_poll;                        // NO# of dev->poll are called    
    u32 dw_redo_poll;                   // NO# of redo polling in dev->poll()
    u32 dw_tx_poll;                     // NO# of tx_interrupt() called
    u32 dw_tx_polling_timer;            // NO# of tx_interrupt() called in polling timer
    u32 dw_rx_poll;                     // NO# of rx_interrupt() called        
    u32 dw_rx_no_quota;                 // NO# of rx_interrupt() called but no quota
    u32 dw_poll_done;                   // NO# of dev->poll() and done
    u32 dw_poll_not_done;               // NO# of dev->poll() and not done
    u32 dw_tx_pkts;                     // NO# of packets sent
    u32 dw_rx_pkts;                     // NO# of packets received
    u32 dw_rx_dropped;                  // NO# of received packets dropped as NET_IF_RX() failed
    u32 dw_tx_started;                  // NO# of netif_start_queue() is called
    u32 dw_tx_stopped;                  // NO# of netif_stop_queue() is called
  } ST_COUNTER;
#endif

struct rtl8169_private {
	unsigned long ioaddr;                /* memory map physical address*/
	struct pci_dev *pci_dev;                /* Index of PCI device  */
	struct net_device_stats stats;          /* statistics of net device */
	spinlock_t lock;                        /* spin lock flag */
	int chipset;
	int mcfg;
	int pcfg;
	rt_timer_t r8169_timer;
	unsigned long expire_time;

	unsigned long phy_link_down_cnt;
	unsigned long cur_rx;                   /* Index into the Rx descriptor buffer of next Rx pkt. */
	unsigned long cur_tx;                   /* Index into the Tx descriptor buffer of next Rx pkt. */
	unsigned long dirty_tx;
	struct	TxDesc	*TxDescArray;           /* Index of 256-alignment Tx Descriptor buffer */
	struct	RxDesc	*RxDescArray;           /* Index of 256-alignment Rx Descriptor buffer */
	struct	sk_buff	*Tx_skbuff[NUM_TX_DESC];/* Index of Transmit data buffer */
	struct	sk_buff	*Rx_skbuff[NUM_RX_DESC];/* Receive data buffer */
	unsigned char   drvinit_fail;

	dma_addr_t txdesc_array_dma_addr[NUM_TX_DESC];
	dma_addr_t rxdesc_array_dma_addr[NUM_RX_DESC];
	dma_addr_t rx_skbuff_dma_addr[NUM_RX_DESC];

	void *txdesc_space;
	dma_addr_t txdesc_phy_dma_addr;
	int sizeof_txdesc_space;

	void *rxdesc_space;
	dma_addr_t rxdesc_phy_dma_addr;
	int sizeof_rxdesc_space;

	int curr_mtu_size;
	int tx_pkt_len;
	int rx_pkt_len;

#ifdef CONFIG_R8169_NAPI
	u16 intr_mask;
	u8  c_rx_int_done;  // 1 if Rx descriptors had been all processed
#endif

#ifdef RTL8169_CHECKSUM_BY_HW
	// 1 if Checksum by hardware is enabled, 0 if disabled
	u8  c_checksum_by_hw_enabled_flag;
#endif

#ifdef M_RX_SKB_POOL
	ST_SKB_POOL st_skb_pool;     // SKB pool for Rx
	struct tq_struct	rx_bh;
	u8 c_skb_pool_timer_active_flag;
#endif

#ifdef M_HANK_PROFILE
	ST_COUNTER st_counter;
#endif

#ifdef M_DYNAMICALLY_TUNE_INTRMITIGATE_REGISTER
	u8 c_speed;                     // Speed of controller (10/100/1000Mbps)
#endif

	int hw_rx_pkt_len;

#ifdef RTL8169_DYNAMIC_CONTROL
	struct r8169_cb_t rt;
#endif //end #ifdef RTL8169_DYNAMIC_CONTROL

	unsigned char   linkstatus;
};


#ifdef RTL8169_CHECKSUM_BY_HW
  //
  // These macro are used to enable/disable/check Checumsum by hardware function.
  //
  #define M_ENABLE_CHECKSUM_BY_HW(priv)     (priv->c_checksum_by_hw_enabled_flag = 1)
  #define M_DISABLE_CHECKSUM_BY_HW(priv)    (priv->c_checksum_by_hw_enabled_flag = 0)
  #define M_IS_CHECKSUM_BY_HW_ENABLED(priv) (priv->c_checksum_by_hw_enabled_flag)
#endif

MODULE_AUTHOR ("Realtek");
MODULE_DESCRIPTION ("RealTek RTL-8169 Gigabit Ethernet driver");
MODULE_PARM (media, "1-" __MODULE_STRING(MAX_UNITS) "i");
MODULE_LICENSE("GPL");


static int rtl8169_open (struct net_device *dev);
static int rtl8169_start_xmit (struct sk_buff *skb, struct net_device *dev);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
typedef	int				irqreturn_t;
#define	IRQ_NONE		0
#define	IRQ_HANDLED		1
static void rtl8169_interrupt (int irq, void *dev_instance, struct pt_regs *regs);
#else
static irqreturn_t rtl8169_interrupt (int irq, void *dev_instance, struct pt_regs *regs);
#endif

static void rtl8169_init_ring (struct net_device *dev);
static void rtl8169_hw_start (struct net_device *dev);
static int rtl8169_close (struct net_device *dev);
static inline u32 ether_crc (int length, unsigned char *data);
static void rtl8169_set_rx_mode (struct net_device *dev);
static void rtl8169_tx_timeout (struct net_device *dev);
static struct net_device_stats *rtl8169_get_stats(struct net_device *netdev);

#ifdef RTL8169_JUMBO_FRAME_SUPPORT
static int rtl8169_change_mtu(struct net_device *dev, int new_mtu);
#endif //end #ifdef RTL8169_JUMBO_FRAME_SUPPORT

static void rtl8169_hw_PHY_config (struct net_device *dev);
static void rtl8169_hw_PHY_reset(struct net_device *dev);
static void rtl8169_tx_interrupt (struct net_device *dev, struct rtl8169_private *priv, unsigned long ioaddr);


#ifdef M_TX_POLLING_MODE
  // If handle Tx by polling timer, disable the TxOK interrupt of RTL8169S
  static const u16 rtl8169_intr_mask = LinkChg | RxOverflow | RxFIFOOver | TxErr | RxErr | RxOK ;
#else
  static const u16 rtl8169_intr_mask = LinkChg | RxOverflow | RxFIFOOver | TxErr | TxOK | RxErr | RxOK ;
#endif

#ifdef M_NH230
  static const unsigned int rtl8169_rx_config = (RX_FIFO_THRESH << RxCfgFIFOShift) | (RX_DMA_BURST << RxCfgDMAShift) | 0x0000000A;
#else
  static const unsigned int rtl8169_rx_config = (RX_FIFO_THRESH << RxCfgFIFOShift) | (RX_DMA_BURST << RxCfgDMAShift) | 0x0000000E;
#endif

#ifdef CONFIG_R8169_NAPI
  static int rtl8169_poll(struct net_device *dev, int *budget);
  
  #ifdef M_TX_POLLING_MODE_NOT_ENABLED
    // If Tx polling mode, do not handle TxOK interrupt
    static const u16 rtl8169_napi_event = RxOK | RxOverflow | RxFIFOOver | TxErr;
  #else
    static const u16 rtl8169_napi_event = RxOK | RxOverflow | RxFIFOOver | TxOK | TxErr;
  #endif
#endif

#ifdef M_NH230
  // The macro determines whether any packets in Tx descriptor ring or not ?
  #define M_IS_PKTS_IN_TX_RING()  ( (priv->cur_tx - priv->dirty_tx) > 0 )
#endif


#define RTL8169_WRITE_GMII_REG_BIT( ioaddr, reg, bitnum, bitval )\
{ \
	int val; \
	if( bitval == 1 ){ val = ( RTL8169_READ_GMII_REG( ioaddr, reg ) | (bitval<<bitnum) ) & 0xffff ; } \
	else{ val = ( RTL8169_READ_GMII_REG( ioaddr, reg ) & (~(0x0001<<bitnum)) ) & 0xffff ; } \
	RTL8169_WRITE_GMII_REG( ioaddr, reg, val ); \
}



#ifdef RTL8169_DEBUG
unsigned alloc_rxskb_cnt = 0;
#define RTL8169_ALLOC_RXSKB(bufsize) 	dev_alloc_skb(bufsize); alloc_rxskb_cnt ++ ;
#define RTL8169_FREE_RXSKB(skb) 	kfree_skb(skb); alloc_rxskb_cnt -- ;
#define RTL8169_NETIF_RX(skb) 		netif_rx(skb); alloc_rxskb_cnt -- ;
#else
#define RTL8169_ALLOC_RXSKB(bufsize) 	dev_alloc_skb(bufsize);
#define RTL8169_FREE_RXSKB(skb) 	kfree_skb(skb);
#define RTL8169_NETIF_RX(skb) 		netif_rx(skb);
#endif //end #ifdef RTL8169_DEBUG






//=================================================================
//	PHYAR
//	bit		Symbol
//	31		Flag
//	30-21	reserved
//	20-16	5-bit GMII/MII register address
//	15-0	16-bit GMII/MII register data
//=================================================================
void RTL8169_WRITE_GMII_REG( unsigned long ioaddr, int RegAddr, int value )
{
	int	i;

	RTL_W32 ( PHYAR, 0x80000000 | (RegAddr&0xFF)<<16 | value);
	udelay(1000);

	for( i = 2000; i > 0 ; i -- ){
		// Check if the RTL8169 has completed writing to the specified MII register
		if( ! (RTL_R32(PHYAR)&0x80000000) ){
			break;
		}
		else{
			udelay(100);
		}// end of if( ! (RTL_R32(PHYAR)&0x80000000) )
	}// end of for() loop
}
//=================================================================
int RTL8169_READ_GMII_REG( unsigned long ioaddr, int RegAddr )
{
	int i, value = -1;

	RTL_W32 ( PHYAR, 0x0 | (RegAddr&0xFF)<<16 );
	udelay(1000);

	for( i = 2000; i > 0 ; i -- ){
		// Check if the RTL8169 has completed retrieving data from the specified MII register
		if( RTL_R32(PHYAR) & 0x80000000 ){
			value = (int)( RTL_R32(PHYAR)&0xFFFF );
			break;
		}
		else{
			udelay(100);
		}// end of if( RTL_R32(PHYAR) & 0x80000000 )
	}// end of for() loop
	return value;
}


#ifdef RTL8169_IOCTL_SUPPORT
#include "r8169_ioctl.c"
#endif //end #ifdef RTL8169_IOCTL_SUPPORT


#ifdef RTL8169_DYNAMIC_CONTROL
#include "r8169_callback.c"
#endif



#define rtl8169_request_timer( timer, timer_expires, timer_func, timer_data ) \
{ \
	init_timer(timer); \
	timer->expires = (unsigned long)(jiffies + timer_expires); \
	timer->data = (unsigned long)(timer_data); \
	timer->function = (void *)(timer_func); \
	add_timer(timer); \
	DBG_PRINT("request_timer at 0x%08lx\n", (unsigned long)timer); \
}

#define rtl8169_delete_timer( del_timer_t ) \
{ \
	del_timer(del_timer_t); \
	DBG_PRINT("delete_timer at 0x%08lx\n", (unsigned long)del_timer_t); \
}

#define rtl8169_mod_timer( timer, timer_expires ) \
{ \
	mod_timer( timer, jiffies + timer_expires ); \
}




//======================================================================================================
//======================================================================================================
void rtl8169_phy_timer_t_handler( void	*timer_data )
{
	struct net_device *dev = (struct net_device *)timer_data;
	struct rtl8169_private *priv = (struct rtl8169_private *) (dev->priv);
	unsigned long ioaddr = priv->ioaddr;

	//M_FDBG("called on jiffies = %d\n", jiffies);
	
	assert( priv->mcfg > MCFG_METHOD_1 );
	assert( priv->pcfg < PCFG_METHOD_3 );

	if( RTL_R8(PHYstatus) & LinkStatus ){    // Link up ?
		priv->phy_link_down_cnt = 0 ;
	}
	else{                                    // Link down ?
		priv->phy_link_down_cnt ++ ;
		if( priv->phy_link_down_cnt >= 12 ){
			// If link on 1000, perform phy reset.
			if( RTL8169_READ_GMII_REG( ioaddr, PHY_1000_CTRL_REG ) & PHY_Cap_1000_Full )
			{
				DBG_PRINT("rtl8169_hw_PHY_reset\n");
				rtl8169_hw_PHY_reset( dev );
			}

			priv->phy_link_down_cnt = 0 ;
		}
	}

	//---------------------------------------------------------------------------
	//mod_timer is a more efficient way to update the expire field of an active timer.
	//---------------------------------------------------------------------------
//	rtl8169_mod_timer( (&priv->phy_timer_t), 100 );
}



//======================================================================================================
//======================================================================================================
void rtl8169_timer_handler( void *timer_data )
{
	struct net_device *dev = (struct net_device *)timer_data;
	struct rtl8169_private *priv = (struct rtl8169_private *) (dev->priv);
	
	if( (priv->mcfg > MCFG_METHOD_1) && (priv->pcfg < PCFG_METHOD_3) ){
		DBG_PRINT("FIX PCS -> rtl8169_phy_timer_t_handler\n");
#ifdef M_NH230
		// Should not set "priv->phy_link_down_cnt" to 0 here.
		// It should be reset in rtl8169_phy_timer_t_handler() routine.
#else
		priv->phy_link_down_cnt = 0;
#endif
		rtl8169_phy_timer_t_handler( timer_data );
	}

#ifdef M_TX_POLLING_MODE
	// If handle Tx by polling timer, prcoess Tx descriptor ring here.
	if (M_IS_PKTS_IN_TX_RING()) {  // pkts in Tx descriptor ring ?
  #ifdef CONFIG_R8169_NAPI
	  if (!test_bit(__LINK_STATE_RX_SCHED, &dev->state)) {// dev->poll() in-active ?
  #else
          {
  #endif
		// Process the Tx descriptor ring only if not empty
	  	unsigned long ioaddr = priv->ioaddr;
	  	rtl8169_tx_interrupt(dev, priv, ioaddr);
		M_INC_COUNTER(priv->st_counter.dw_tx_polling_timer);
	  }
	}
#endif

#ifdef M_DYNAMICALLY_TUNE_INTRMITIGATE_REGISTER
	// Dynamically tuning the IntrMitigate register according to speed
	unsigned long ioaddr = priv->ioaddr;
	u8 c_PHY_status = RTL_R8(PHYstatus), c_curr_speed = 0;	
	u16 w_intrmitigate_reg = INTRMITIGATE_REG_DEFAULT;
	
	if ( c_PHY_status & LinkStatus ) {     // Link up ?
	  if ( c_PHY_status & _1000bpsF ) {    // 1Gbps ?
	    w_intrmitigate_reg = INTRMITIGATE_REG_1Gbps;
	    c_curr_speed = _1000bpsF;
	  }
	  else {
	    if ( c_PHY_status & _100bps ) {    // 100Mbps ?
	      w_intrmitigate_reg = INTRMITIGATE_REG_100Mbps;
	      c_curr_speed = _100bps;
	    }
	    else {
	      if ( c_PHY_status & _10bps ) {   // 10Mbps ?
	        w_intrmitigate_reg = INTRMITIGATE_REG_10Mbps;
		c_curr_speed = _10bps;
	      }
	    }
	  }	  
	  
	  // Write to IntrMitigate register only if speed changed.
	  if (c_curr_speed != priv->c_speed) { // Speed changed ?
	    RTL_W16(IntrMitigate, w_intrmitigate_reg);
	    M_DBG("%s: set IntrMitigate Register to 0x%X, speed curr/prev = 0x%X/0x%X\n",
    		__FUNCTION__, RTL_R16(IntrMitigate), c_curr_speed, priv->c_speed);
		
	    priv->c_speed = c_curr_speed;      // Duplicate current speed
	  }
	}
#endif


#ifdef RTL8169_DYNAMIC_CONTROL
	{
		struct r8169_cb_t *rt = &(priv->rt);
		if( priv->linkstatus == _1000_Full ){
			r8169_callback(rt);
		}
	}
#endif //end #ifdef RTL8169_DYNAMIC_CONTROL


	rtl8169_mod_timer( (&priv->r8169_timer), priv->expire_time );
}



//======================================================================================================
//======================================================================================================
static int __devinit rtl8169_init_board ( struct pci_dev *pdev, struct net_device **dev_out, unsigned long *ioaddr_out)
{
	unsigned long ioaddr = 0;
	struct net_device *dev;
	struct rtl8169_private *priv;
	int rc, i;
	unsigned long mmio_start, mmio_end, mmio_flags, mmio_len;


	assert (pdev != NULL);
	assert (ioaddr_out != NULL);

	*ioaddr_out = 0;
	*dev_out = NULL;

	// dev zeroed in init_etherdev
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	dev = init_etherdev (NULL, sizeof (*priv));
#else
	dev = alloc_etherdev (sizeof (*priv));
#endif

	if (dev == NULL) {
		printk (KERN_ERR PFX "unable to alloc new ethernet\n");
		return -ENOMEM;
	}

	SET_MODULE_OWNER(dev);

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,0)
	SET_NETDEV_DEV(dev, &pdev->dev);
#endif

	priv = dev->priv;

	// enable device (incl. PCI PM wakeup and hotplug setup)
	rc = pci_enable_device (pdev);
	if (rc)
		goto err_out;

	mmio_start = pci_resource_start (pdev, 1);
	mmio_end = pci_resource_end (pdev, 1);
	mmio_flags = pci_resource_flags (pdev, 1);
	mmio_len = pci_resource_len (pdev, 1);

	// make sure PCI base addr 1 is MMIO
	if (!(mmio_flags & IORESOURCE_MEM)) {
		printk (KERN_ERR PFX "region #1 not an MMIO resource, aborting\n");
		rc = -ENODEV;
		goto err_out;
	}

	// check for weird/broken PCI region reporting
	if ( mmio_len < RTL_MIN_IO_SIZE ) {
		printk (KERN_ERR PFX "Invalid PCI region size(s), aborting\n");
		rc = -ENODEV;
		goto err_out;
	}


	rc = pci_request_regions (pdev, dev->name);
	if (rc)
		goto err_out;

	// enable PCI bus-mastering
	pci_set_master (pdev);

#ifdef RTL8169_USE_IO
	ioaddr = pci_resource_start(pdev, 0);
#else
	// ioremap MMIO region
	ioaddr = (unsigned long)ioremap (mmio_start, mmio_len);
	if (ioaddr == 0) {
		printk (KERN_ERR PFX "cannot remap MMIO, aborting\n");
		rc = -EIO;
		goto err_out_free_res;
	}
#endif

	// Soft reset the chip.
	RTL_W8 ( ChipCmd, CmdReset);

	// Check that the chip has finished the reset.
	for (i = 1000; i > 0; i--){
		if ( (RTL_R8(ChipCmd) & CmdReset) == 0){
			break;
		}
		else{
			udelay (10);
		}
	}

	// identify config method
	{
		unsigned long val32 = (RTL_R32(TxConfig)&0x7c800000);

		if( val32 == (0x1<<28) ){
			priv->mcfg = MCFG_METHOD_4;
		}
		else if( val32 == (0x1<<26) ){
			priv->mcfg = MCFG_METHOD_3;
		}
		else if( val32 == (0x1<<23) ){
			priv->mcfg = MCFG_METHOD_2;
		}
		else if( val32 == 0x00000000 ){
			priv->mcfg = MCFG_METHOD_1;
		}
		else{
			priv->mcfg = MCFG_METHOD_1;
		}
	}
	{
		unsigned char val8 = (unsigned char)(RTL8169_READ_GMII_REG(ioaddr,3)&0x000f);
		if( val8 == 0x00 ){
			priv->pcfg = PCFG_METHOD_1;
		}
		else if( val8 == 0x01 ){
			priv->pcfg = PCFG_METHOD_2;
		}
		else if( val8 == 0x02 ){
			priv->pcfg = PCFG_METHOD_3;
		}
		else{
			priv->pcfg = PCFG_METHOD_3;
		}
	}


	for (i = ARRAY_SIZE (rtl_chip_info) - 1; i >= 0; i--){
		if (priv->mcfg == rtl_chip_info[i].mcfg) {
			priv->chipset = i;
			goto match;
		}
	}

	//if unknown chip, assume array element #0, original RTL-8169 in this case
	printk (KERN_DEBUG PFX "PCI device %s: unknown chip version, assuming RTL-8169\n", pdev->slot_name);
	priv->chipset = 0;

match:
	*ioaddr_out = ioaddr;
	*dev_out = dev;
	return 0;

#ifndef RTL8169_USE_IO
err_out_free_res:
	pci_release_regions (pdev);
#endif

err_out:
	unregister_netdev (dev);
	kfree (dev);
	return rc;
}







//======================================================================================================
static int __devinit rtl8169_init_one (struct pci_dev *pdev, const struct pci_device_id *ent)
{
	struct net_device *dev = NULL;
	struct rtl8169_private *priv = NULL;
	unsigned long ioaddr = 0;
	static int board_idx = -1;
	int i;
	int option = -1, Cap10_100 = 0, Cap1000 = 0;


	assert (pdev != NULL);
	assert (ent != NULL);

	board_idx++;


	i = rtl8169_init_board (pdev, &dev, &ioaddr);
	
	if (i < 0) {
		return i;
	}

	priv = dev->priv;

	assert (ioaddr != NULL);
	assert (dev != NULL);
	assert (priv != NULL);

#ifdef M_NH230
	// Duplicated MAC address parsed from kernel command line "ethaddr=xx:xx:xx:xx:xx:xx"
	memcpy(dev->dev_addr, gsz_eth_address, ETH_ALEN);
	
	g_ioaddr = ioaddr;
	g_dev = dev;
	g_priv = priv;
#else
	// Get MAC address //
	for (i = 0; i < MAC_ADDR_LEN ; i++){
		dev->dev_addr[i] = RTL_R8( MAC0 + i );
	}
#endif

	dev->open		= rtl8169_open;
	dev->hard_start_xmit 	= rtl8169_start_xmit;
	dev->get_stats    	= rtl8169_get_stats;
	dev->stop 		= rtl8169_close;
	dev->tx_timeout 	= rtl8169_tx_timeout;
	dev->set_multicast_list = rtl8169_set_rx_mode;
	dev->watchdog_timeo 	= TX_TIMEOUT;
	dev->irq 		= pdev->irq;
	dev->base_addr 		= (unsigned long) ioaddr;

#ifdef M_TX_POLLING_MODE
	dev->tx_queue_len = 1024; 	// Increase the Tx queue length
#endif
	
#ifdef RTL8169_JUMBO_FRAME_SUPPORT
	dev->change_mtu		= rtl8169_change_mtu;
	
  #ifdef M_NH230
        gdw_mtu = dev->mtu;             // duplicate the new MTU
  #endif

#endif //end #ifdef RTL8169_JUMBO_FRAME_SUPPORT

#ifdef RTL8169_CHECKSUM_BY_HW
        if (dev->mtu <= ETH_DATA_LEN)   // Is Jumbo frame disabled (normal MTU) ?
	  M_ENABLE_CHECKSUM_BY_HW(priv);
	else
	  M_DISABLE_CHECKSUM_BY_HW(priv);
#endif

#ifdef CONFIG_R8169_NAPI
	dev->poll = rtl8169_poll;
	dev->weight = R8169_NAPI_WEIGHT;
	
	priv->intr_mask = 0xffff;
	printk(KERN_INFO PFX "RTL8169S: NAPI enabled\n");
#endif

#ifdef M_HANK_PROFILE
	// reset counters for profile
	memset(&(priv->st_counter), 0, sizeof(ST_COUNTER));
#endif


#ifdef RTL8169_IOCTL_SUPPORT
	dev->do_ioctl 		= rtl8169_ioctl;
#endif //end #ifdef RTL8169_IOCTL_SUPPORT

#ifdef RTL8169_DYNAMIC_CONTROL
	priv->rt.dev = dev;
#endif //end #ifdef RTL8169_DYNAMIC_CONTROL

	priv = dev->priv;				// private data //
	priv->pci_dev 	= pdev;
	priv->ioaddr 	= ioaddr;

//#ifdef RTL8169_JUMBO_FRAME_SUPPORT
	priv->curr_mtu_size = dev->mtu;
	priv->tx_pkt_len = dev->mtu + ETH_HDR_LEN;
	priv->rx_pkt_len = dev->mtu + ETH_HDR_LEN;
	priv->hw_rx_pkt_len = priv->rx_pkt_len + 8;
//#endif //end #ifdef RTL8169_JUMBO_FRAME_SUPPORT

	DBG_PRINT("-------------------------- \n");
	DBG_PRINT("dev->mtu = %d \n", dev->mtu);
	DBG_PRINT("priv->curr_mtu_size = %d \n", priv->curr_mtu_size);
	DBG_PRINT("priv->tx_pkt_len = %d \n", priv->tx_pkt_len);
	DBG_PRINT("priv->rx_pkt_len = %d \n", priv->rx_pkt_len);
	DBG_PRINT("priv->hw_rx_pkt_len = %d \n", priv->hw_rx_pkt_len);
	DBG_PRINT("-------------------------- \n");

	spin_lock_init (&priv->lock);

	register_netdev (dev);

	pci_set_drvdata(pdev, dev);     //      pdev->driver_data = data;


	printk (KERN_DEBUG "%s: Identified chip type is '%s'.\n",dev->name,rtl_chip_info[priv->chipset].name);
	printk (KERN_INFO "%s: %s at 0x%lx, "
				"%2.2x:%2.2x:%2.2x:%2.2x:%2.2x:%2.2x, "
				"IRQ %d\n",
				dev->name,
				RTL8169_DRIVER_NAME,
				dev->base_addr,
				dev->dev_addr[0], dev->dev_addr[1],
				dev->dev_addr[2], dev->dev_addr[3],
				dev->dev_addr[4], dev->dev_addr[5],
				dev->irq);


	// Config PHY
	rtl8169_hw_PHY_config(dev);

	DBG_PRINT("Set MAC Reg C+CR Offset 0x82h = 0x01h\n");
	RTL_W8( 0x82, 0x01 );

	if( priv->mcfg < MCFG_METHOD_3 ){
		DBG_PRINT("Set PCI Latency=0x40\n");
		pci_write_config_byte(pdev, PCI_LATENCY_TIMER, 0x40);
	}

#ifdef M_NH230
	// increase PCI latency timer register
	M_DBG("%s: Initialize PCI Configuration Registers.\n", __FUNCTION__);
	////pci_write_config_byte(pdev, PCI_LATENCY_TIMER, 0x40);
	pci_write_config_byte(pdev, PCI_LATENCY_TIMER, 0x80);

	// Enable Memory Write & invalidate	
	u8 c_byte = 0;
	pci_read_config_byte(pdev, PCI_COMMAND, &c_byte);
	pci_write_config_byte(pdev, PCI_COMMAND, c_byte | PCI_COMMAND_INVALIDATE);
	
	// increase the PCI cache line size
	pci_write_config_byte(pdev, PCI_CACHE_LINE_SIZE, 0x08);   // 32 DWORDs	
#endif

	if( priv->mcfg == MCFG_METHOD_2 ){
		DBG_PRINT("Set MAC Reg C+CR Offset 0x82h = 0x01h\n");
		RTL_W8( 0x82, 0x01 );
		DBG_PRINT("Set PHY Reg 0x0bh = 0x00h\n");
		RTL8169_WRITE_GMII_REG( ioaddr, 0x0b, 0x0000 );	//w 0x0b 15 0 0
	}

	// if TBI is not endbled
	if( !(RTL_R8(PHYstatus) & TBI_Enable) ){
		int	val = RTL8169_READ_GMII_REG( ioaddr, PHY_AUTO_NEGO_REG );

#ifdef RTL8169_HW_FLOW_CONTROL_SUPPORT
		val |= PHY_Cap_PAUSE | PHY_Cap_ASYM_PAUSE ;
#endif //end #define RTL8169_HW_FLOW_CONTROL_SUPPORT

		option = (board_idx >= MAX_UNITS) ? 0 : media[board_idx];
		sp_r8169Trace("before Nway..RTL8169_READ_GMII_REG( ioaddr, 0 );=%08x", RTL8169_READ_GMII_REG( ioaddr, 0 ));
		sp_r8169Trace("before Nway..RTL8169_READ_GMII_REG( ioaddr, 1 );=%08x", RTL8169_READ_GMII_REG( ioaddr, 1 ));
		sp_r8169Trace("before Nway..RTL8169_READ_GMII_REG( ioaddr, 4 );=%08x", RTL8169_READ_GMII_REG( ioaddr, 4 ));
		sp_r8169Trace("before Nway..RTL_R8(PHYstatus);=%08x", RTL_R8(PHYstatus));
		// Force RTL8169 in 10/100/1000 Full/Half mode.
		if( option > 0 ){
			printk(KERN_INFO "%s: Force-mode Enabled. \n", dev->name);
			Cap10_100 = 0;
			Cap1000 = 0;
			switch( option ){
				case _10_Half:
						Cap10_100 = PHY_Cap_10_Half;
						Cap1000 = PHY_Cap_Null;
						break;
				case _10_Full:
						Cap10_100 = PHY_Cap_10_Full | PHY_Cap_10_Half;
						Cap1000 = PHY_Cap_Null;
						break;
				case _100_Half:
						Cap10_100 = PHY_Cap_100_Half | PHY_Cap_10_Full | PHY_Cap_10_Half;
						Cap1000 = PHY_Cap_Null;
						break;
				case _100_Full:
						Cap10_100 = PHY_Cap_100_Full | PHY_Cap_100_Half | PHY_Cap_10_Full | PHY_Cap_10_Half;
						Cap1000 = PHY_Cap_Null;
						break;
				case _1000_Full:
						Cap10_100 = PHY_Cap_100_Full | PHY_Cap_100_Half | PHY_Cap_10_Full | PHY_Cap_10_Half;
						Cap1000 = PHY_Cap_1000_Full;
						break;
				default:
						break;
			}
			RTL8169_WRITE_GMII_REG( ioaddr, PHY_AUTO_NEGO_REG, Cap10_100 | ( val&0xC1F ) );	//leave PHY_AUTO_NEGO_REG bit4:0 unchanged
			RTL8169_WRITE_GMII_REG( ioaddr, PHY_1000_CTRL_REG, Cap1000 );
		}
		else{
			printk(KERN_INFO "%s: Auto-negotiation Enabled.\n", dev->name);

			// enable 10/100 Full/Half Mode, leave PHY_AUTO_NEGO_REG bit4:0 unchanged
			RTL8169_WRITE_GMII_REG( ioaddr, PHY_AUTO_NEGO_REG,
				PHY_Cap_10_Half | PHY_Cap_10_Full | PHY_Cap_100_Half | PHY_Cap_100_Full | ( val&0xC1F ) );

			// enable 1000 Full Mode
#ifdef M_REALTEK_20050225_IPERF
			RTL8169_WRITE_GMII_REG( ioaddr, PHY_1000_CTRL_REG, PHY_Cap_1000_Full | PHY_Cap_1000_Half);
#else
			RTL8169_WRITE_GMII_REG( ioaddr, PHY_1000_CTRL_REG, PHY_Cap_1000_Full );
#endif
		}// end of if( option > 0 )


		// Enable auto-negotiation and restart auto-nigotiation
		RTL8169_WRITE_GMII_REG( ioaddr, PHY_CTRL_REG, PHY_Enable_Auto_Nego | PHY_Restart_Auto_Nego );
		udelay(100);

		// wait for auto-negotiation process
		for( i = 10000; i > 0; i-- ){
			//check if auto-negotiation complete
			if( RTL8169_READ_GMII_REG(ioaddr, PHY_STAT_REG) & PHY_Auto_Neco_Comp ){
				udelay(100);
				option = RTL_R8(PHYstatus);
				if( option & _1000bpsF ){
					printk(KERN_INFO "%s: 1000Mbps Full-duplex operation.\n", dev->name);
				}
				else{
					printk(KERN_INFO "%s: %sMbps %s-duplex operation.\n", dev->name,
							(option & _100bps) ? "100" : "10", (option & FullDup) ? "Full" : "Half" );
				}
				break;
			}
			else{
				udelay(100);
			}// end of if( RTL8169_READ_GMII_REG(ioaddr, 1) & 0x20 )
		}// end for-loop to wait for auto-negotiation process

		option = RTL_R8(PHYstatus);
		if( option & _1000bpsF ){
			priv->linkstatus = _1000_Full;
		}
		else{
			if(option & _100bps){
				priv->linkstatus = (option & FullDup) ? _100_Full : _100_Half;
			}
            else{
				priv->linkstatus = (option & FullDup) ? _10_Full : _10_Half;
			}
		}
		DBG_PRINT("priv->linkstatus = 0x%02x\n", priv->linkstatus);

	}// end of TBI is not enabled
	else{
		udelay(100);
		DBG_PRINT("1000Mbps Full-duplex operation, TBI Link %s!\n",(RTL_R32(TBICSR) & TBILinkOK) ? "OK" : "Failed" );

	}// end of TBI is not enabled

	sp_r8169Trace("after Nway..RTL8169_READ_GMII_REG( ioaddr, 0 );=%08x", RTL8169_READ_GMII_REG( ioaddr, 0 ));
	sp_r8169Trace("after Nway..RTL8169_READ_GMII_REG( ioaddr, 1 );=%08x", RTL8169_READ_GMII_REG( ioaddr, 1 ));
	sp_r8169Trace("after Nway..RTL8169_READ_GMII_REG( ioaddr, 4 );=%08x", RTL8169_READ_GMII_REG( ioaddr, 4 ));
	sp_r8169Trace("before Nway..RTL_R8(PHYstatus);=%08x", RTL_R8(PHYstatus));
	return 0;
}







//======================================================================================================
static void __devexit rtl8169_remove_one (struct pci_dev *pdev)
{
	struct net_device *dev = pci_get_drvdata(pdev);

	assert (dev != NULL);
#ifdef M_NH230
#else
	assert (priv != NULL);
#endif

	unregister_netdev (dev);

#ifdef RTL8169_USE_IO
#else
	iounmap ((void *)(dev->base_addr));
#endif
	pci_release_regions (pdev);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	kfree (dev);
#else
	free_netdev(dev);
#endif

	pci_set_drvdata (pdev, NULL);
}







//======================================================================================================
static int rtl8169_open (struct net_device *dev)
{
	struct rtl8169_private *priv = dev->priv;
	struct pci_dev *pdev = priv->pci_dev;
	int retval;
//	u8 diff;
//	u32 TxPhyAddr, RxPhyAddr;


	if( priv->drvinit_fail == 1 ){
		printk("%s: Gigabit driver open failed.\n", dev->name );
		return -ENOMEM;
	}

	retval = request_irq (dev->irq, rtl8169_interrupt, SA_SHIRQ, dev->name, dev);
	if (retval) {
		return retval;
	}

	//2004-05-11
	// Allocate tx/rx descriptor space
	priv->sizeof_txdesc_space = NUM_TX_DESC * sizeof(struct TxDesc)+256;
	priv->txdesc_space = pci_alloc_consistent( pdev, priv->sizeof_txdesc_space, &priv->txdesc_phy_dma_addr );
	if( priv->txdesc_space == NULL ){
		printk("%s: Gigabit driver alloc txdesc_space failed.\n", dev->name );
		return -ENOMEM;
	}
	priv->sizeof_rxdesc_space = NUM_RX_DESC * sizeof(struct RxDesc)+256;
	priv->rxdesc_space = pci_alloc_consistent( pdev, priv->sizeof_rxdesc_space, &priv->rxdesc_phy_dma_addr );
	if( priv->rxdesc_space == NULL ){
		printk("%s: Gigabit driver alloc rxdesc_space failed.\n", dev->name );
		return -ENOMEM;
	}

	if(priv->txdesc_phy_dma_addr & 0xff){
		printk("%s: Gigabit driver txdesc_phy_dma_addr is not 256-bytes-aligned.\n", dev->name );
	}
	if(priv->rxdesc_phy_dma_addr & 0xff){
		printk("%s: Gigabit driver rxdesc_phy_dma_addr is not 256-bytes-aligned.\n", dev->name );
	}
	// Set tx/rx descriptor space
	priv->TxDescArray = (struct TxDesc *)priv->txdesc_space;
	priv->RxDescArray = (struct RxDesc *)priv->rxdesc_space;

	{
		int i;
		struct sk_buff *skb = NULL;

		for(i=0;i<NUM_RX_DESC;i++){
			skb = RTL8169_ALLOC_RXSKB(MAX_RX_SKBDATA_SIZE);
			if( skb != NULL ) {
				skb_reserve (skb, 2);	// 16 byte align the IP fields. //
				priv->Rx_skbuff[i] = skb;
			}
			else{
				printk("%s: Gigabit driver failed to allocate skbuff.\n", dev->name);
				priv->drvinit_fail = 1;
			}
		}
	}

#ifdef M_RX_SKB_POOL
	// Initialize the SKB pool and allocate the memory for its entries
	priv->c_skb_pool_timer_active_flag = 0;
	fn_skb_pool_handler(dev, CMD_SKB_POOL_ALLOCATE);
	INIT_TQUEUE(&priv->rx_bh, RXbh, dev);
#endif

#ifdef M_NH230
  #ifdef RTL8169_CHECKSUM_BY_HW
  	if (M_IS_CHECKSUM_BY_HW_ENABLED(priv))  // Is Checksum by H/W enabled ?
		dev->features |= NETIF_F_SG | NETIF_F_IP_CSUM;
	else
		dev->features &= ~(NETIF_F_SG | NETIF_F_IP_CSUM);
  #endif
#endif


	//////////////////////////////////////////////////////////////////////////////
	rtl8169_init_ring (dev);
	rtl8169_hw_start (dev);


	// ------------------------------------------------------
	DBG_PRINT("FIX PCS -> rtl8169_request_timer\n");
	priv->expire_time = RTL8169_TIMER_EXPIRE_TIME;
	rtl8169_request_timer( (&priv->r8169_timer), priv->expire_time, rtl8169_timer_handler, ((void *)dev) );  //in open()


	DBG_PRINT("%s: %s() alloc_rxskb_cnt = %d\n", dev->name, __FUNCTION__, alloc_rxskb_cnt );

	return 0;

}//end of rtl8169_open (struct net_device *dev)








//======================================================================================================
static void rtl8169_hw_PHY_reset(struct net_device *dev)
{
	int val, phy_reset_expiretime = 50;
	struct rtl8169_private *priv = dev->priv;
	unsigned long ioaddr = priv->ioaddr;

	DBG_PRINT("%s: Reset RTL8169s PHY\n", dev->name);

	val = ( RTL8169_READ_GMII_REG( ioaddr, 0 ) | 0x8000 ) & 0xffff;
	RTL8169_WRITE_GMII_REG( ioaddr, 0, val );

	do //waiting for phy reset
	{
		if( RTL8169_READ_GMII_REG( ioaddr, 0 ) & 0x8000 ){
			phy_reset_expiretime --;
			udelay(100);
		}
		else{
			break;
		}
	}while( phy_reset_expiretime >= 0 );

	assert( phy_reset_expiretime > 0 );
}




//======================================================================================================
static void rtl8169_hw_PHY_config (struct net_device *dev)
{
	struct rtl8169_private *priv = dev->priv;
	void *ioaddr = (void*)priv->ioaddr;

	DBG_PRINT("priv->mcfg=%d, priv->pcfg=%d\n",priv->mcfg,priv->pcfg);

#ifdef M_REALTEK_20050225_IPERF
 	if( priv->mcfg == MCFG_METHOD_4 ){
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x1F, 0x0001 );
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x1b, 0x841e );
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x0e, 0x7bfb );
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x09, 0x273a );

		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x1F, 0x0002 );
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x01, 0x90D0 );
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x1F, 0x0000 );
	}else if((priv->mcfg == MCFG_METHOD_2)||(priv->mcfg == MCFG_METHOD_3)){
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x1F, 0x0001 );
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x15, 0x1000 );
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x18, 0x65C7 );
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x04, 0x0000 );
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x03, 0x00A1 );
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x02, 0x0008 );
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x01, 0x1020 );
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x00, 0x1000 );
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x04, 0x0800 );
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x04, 0x0000 );
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x04, 0x7000 );
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x03, 0xFF41 );
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x02, 0xDE60 );
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x01, 0x0140 );
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x00, 0x0077 );
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x04, 0x7800 );
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x04, 0x7000 );
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x04, 0xA000 );
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x03, 0xDF01 );
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x02, 0xDF20 );
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x01, 0xFF95 );
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x00, 0xFA00 );
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x04, 0xA800 );
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x04, 0xA000 );
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x04, 0xB000 );
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x03, 0xFF41 );
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x02, 0xDE20 );
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x01, 0x0140 );
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x00, 0x00BB );
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x04, 0xB800 );
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x04, 0xB000 );
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x04, 0xF000 );
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x03, 0xDF01 );
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x02, 0xDF20 );
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x01, 0xFF95 );
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x00, 0xBF00 );
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x04, 0xF800 );
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x04, 0xF000 );
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x04, 0x0000 );
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x1F, 0x0000 );
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x0B, 0x0000 );
	}
	else{
		DBG_PRINT("priv->mcfg=%d. Discard hw PHY config.\n",priv->mcfg);
	}
#else
	if( priv->mcfg == MCFG_METHOD_4 ){
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x1F, 0x0001 );
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x09, 0x273a );
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x0e, 0x7bfb );
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x1b, 0x841e );

		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x1F, 0x0002 );
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x01, 0x90D0 );
		RTL8169_WRITE_GMII_REG( (unsigned long)ioaddr, 0x1F, 0x0000 );
	}
	else{
		DBG_PRINT("priv->mcfg=%d. Discard hw PHY config.\n",priv->mcfg);
	}	
#endif
}










//======================================================================================================
static void rtl8169_hw_start (struct net_device *dev)
{
	struct rtl8169_private *priv = dev->priv;
	unsigned long ioaddr = priv->ioaddr;
	u32 i;


	/* Soft reset the chip. */
	RTL_W8 ( ChipCmd, CmdReset);

	/* Check that the chip has finished the reset. */
	for (i = 1000; i > 0; i--){
		if ((RTL_R8( ChipCmd ) & CmdReset) == 0) break;
		else udelay (10);
	}

	RTL_W8 ( Cfg9346, Cfg9346_Unlock);
	RTL_W8 ( ChipCmd, CmdTxEnb | CmdRxEnb);
	RTL_W8 ( ETThReg, ETTh);

	// For gigabit rtl8169
	RTL_W16	( RxMaxSize, (unsigned short)priv->hw_rx_pkt_len );

	// Set Rx Config register
	i = rtl8169_rx_config | ( RTL_R32( RxConfig ) & rtl_chip_info[priv->chipset].RxConfigMask);
	RTL_W32 ( RxConfig, i);


	/* Set DMA burst size and Interframe Gap Time */
	RTL_W32 ( TxConfig, (TX_DMA_BURST << TxDMAShift) | (InterFrameGap << TxInterFrameGapShift) );

#ifdef M_NH230
  #ifdef RTL8169_CHECKSUM_BY_HW
        // Rx Checksum Offload enable
	if (M_IS_CHECKSUM_BY_HW_ENABLED(priv))  // Is Checksum by H/W enabled ?
        	RTL_W16( CPlusCmd, RTL_R16(CPlusCmd) | RxChkSum);
	else
		RTL_W16( CPlusCmd, RTL_R16(CPlusCmd) & (~RxChkSum));
  #else
  	RTL_W16( CPlusCmd, RTL_R16(CPlusCmd) );
  #endif
#else
	RTL_W16( CPlusCmd, RTL_R16(CPlusCmd) );
#endif

	if(	priv->mcfg == MCFG_METHOD_2 ||
		priv->mcfg == MCFG_METHOD_3)
	{
		// Bit[3]: PCI Multiple Read/Write Enable, Bit[14]: un-documented
		RTL_W16( CPlusCmd, (RTL_R16(CPlusCmd)|(1<<14)|(1<<3)) );
		DBG_PRINT("Set MAC Reg C+CR Offset 0xE0: bit-3 and bit-14\n");
	}
	else
	{
		RTL_W16( CPlusCmd, (RTL_R16(CPlusCmd)|(1<<3)) );
		DBG_PRINT("Set MAC Reg C+CR Offset 0xE0: bit-3.\n");
	}

	{
		//RTL_W16(0xE2, 0x1517);
		//RTL_W16(0xE2, 0x152a);
		//RTL_W16(0xE2, 0x282a);
		RTL_W16(0xE2, 0x0000);
	}

	priv->cur_rx = 0;

	RTL_W32 ( TxDescStartAddr, priv->txdesc_phy_dma_addr);
	RTL_W32 ( TxDescStartAddr + 4, 0x00);
	RTL_W32 ( RxDescStartAddr, priv->rxdesc_phy_dma_addr);
	RTL_W32 ( RxDescStartAddr + 4, 0x00);

	RTL_W8 ( Cfg9346, Cfg9346_Lock );
	udelay (10);

	RTL_W32 ( RxMissed, 0 );

	rtl8169_set_rx_mode (dev);

#ifdef M_NH230
    // Before over-write the IDR0 ~ 5 registers :
    //   93C46 Command Register[6:7]: Operating Mode = Config register write enable
    RTL_W8 (Cfg9346, Cfg9346_Unlock);
    
    // Configure the LED mode
    #define CONFIG1_REG_LED_MODE_3  0x0C0   // bit[7:6] is LEDS1-0
    u8 c_reg = RTL_R8(Config1);
    RTL_W8(Config1, c_reg & (~CONFIG1_REG_LED_MODE_3));
    M_DBG(KERN_NOTICE "RTL8169S : LEDS1-0 is configured to 0b00\n");
        
    M_DBG("%s: MAC address[%02X:%02X:%02X:%02X:%02X:%02X] from kernel command line\n", 
	   __FUNCTION__, gsz_eth_address[0], gsz_eth_address[1], gsz_eth_address[2],
	                 gsz_eth_address[3], gsz_eth_address[4], gsz_eth_address[5]);
   
    // Overwrite the given MAC address into registers
    u32 dw_value = gsz_eth_address[3];
    dw_value <<= 8;
    dw_value += gsz_eth_address[2];
    dw_value <<= 8;
    dw_value += gsz_eth_address[1];
    dw_value <<= 8;
    dw_value += gsz_eth_address[0];
    RTL_W32( MAC0 + 0, dw_value);           // Write 4-byte value to IDR0~3
        
    dw_value = gsz_eth_address[5];
    dw_value <<= 8;
    dw_value += gsz_eth_address[4];
        
    u32 dw_temp = RTL_R8(MAC0 + 6);         // Preserve register (offset = MAC0 + 6)
    dw_value |= dw_temp << 16;
    dw_temp = RTL_R8(MAC0 + 7);             // Preserve register (offset = MAC0 + 7)
    dw_value |= dw_temp << 24;
    RTL_W32( MAC0 + 4, dw_value);           // Write 4-byte to IDR4~5 and Reserved (0x6, 0x7)
        
    // After over-write the IDR0 ~ 5 registers :
    //   93C46 Command Register[6:7]: Operating Mode = Normal communication mode
    RTL_W8 (Cfg9346, Cfg9346_Lock);
        
	M_DBG("%s: MAC address[%02X:%02X:%02X:%02X:%02X:%02X] in ID0~5 registers\n", 
	        __FUNCTION__, RTL_R8(MAC0+0), RTL_R8(MAC0+1), RTL_R8(MAC0+2),
	        RTL_R8(MAC0+3), RTL_R8(MAC0+4), RTL_R8(MAC0+5));
#endif

	RTL_W16 ( MultiIntr, RTL_R16(MultiIntr) & 0xF000);

	RTL_W16 ( IntrMask, rtl8169_intr_mask);

	netif_start_queue (dev);

}//end of rtl8169_hw_start (struct net_device *dev)







//======================================================================================================
static void rtl8169_init_ring (struct net_device *dev)
{
	struct rtl8169_private *priv = dev->priv;
#ifdef M_NH230
#else
	struct pci_dev *pdev = priv->pci_dev;
#endif
	int i;
	struct sk_buff	*skb;
	

	priv->cur_rx = 0;
	priv->cur_tx = 0;
	priv->dirty_tx = 0;
	memset(priv->TxDescArray, 0x0, NUM_TX_DESC*sizeof(struct TxDesc));
	memset(priv->RxDescArray, 0x0, NUM_RX_DESC*sizeof(struct RxDesc));


	for (i=0 ; i<NUM_TX_DESC ; i++){
		priv->Tx_skbuff[i]=NULL;
#ifdef M_NH230
		priv->txdesc_array_dma_addr[i] = virt_to_bus(&priv->TxDescArray[i]);
#else
		priv->txdesc_array_dma_addr[i] = pci_map_single(pdev, &priv->TxDescArray[i], sizeof(struct TxDesc), PCI_DMA_TODEVICE);
#endif
	}

	for (i=0; i <NUM_RX_DESC; i++) {
		if(i==(NUM_RX_DESC-1)){
			priv->RxDescArray[i].status = cpu_to_le32((OWNbit | EORbit) | (unsigned long)priv->hw_rx_pkt_len);
		}
		else{
			priv->RxDescArray[i].status = cpu_to_le32(OWNbit | (unsigned long)priv->hw_rx_pkt_len);
		}

		{//-----------------------------------------------------------------------
			skb = priv->Rx_skbuff[i];
#ifdef M_NH230
			priv->rx_skbuff_dma_addr[i] = virt_to_bus(skb->data);
#else
			priv->rx_skbuff_dma_addr[i] = pci_map_single(pdev, skb->data, MAX_RX_SKBDATA_SIZE, PCI_DMA_FROMDEVICE);
#endif

			if( skb != NULL ){
				priv->RxDescArray[i].buf_addr = cpu_to_le32(priv->rx_skbuff_dma_addr[i]);
				priv->RxDescArray[i].buf_Haddr = 0;
			}
			else{
				DBG_PRINT("%s: %s() Rx_skbuff == NULL\n", dev->name, __FUNCTION__);
				priv->drvinit_fail = 1;
			}
		}//-----------------------------------------------------------------------
#ifdef M_NH230
		priv->rxdesc_array_dma_addr[i] = virt_to_bus(&priv->RxDescArray[i]);
#else
		priv->rxdesc_array_dma_addr[i] = pci_map_single(pdev, &priv->RxDescArray[i], sizeof(struct RxDesc), PCI_DMA_TODEVICE);
		pci_dma_sync_single(pdev, priv->rxdesc_array_dma_addr[i], sizeof(struct RxDesc), PCI_DMA_TODEVICE);
#endif
	}
}







//======================================================================================================
static void rtl8169_tx_clear (struct rtl8169_private *priv)
{
	int i;

	priv->cur_tx = 0;
	for ( i = 0 ; i < NUM_TX_DESC ; i++ ){
		if ( priv->Tx_skbuff[i] != NULL ) {
			dev_kfree_skb ( priv->Tx_skbuff[i] );
			priv->Tx_skbuff[i] = NULL;
			priv->stats.tx_dropped++;
		}
	}
}







//======================================================================================================
static void rtl8169_tx_timeout (struct net_device *dev)
{
	struct rtl8169_private *priv = dev->priv;
	unsigned long ioaddr = priv->ioaddr;
	u8 tmp8;

	/* disable Tx, if not already */
	tmp8 = RTL_R8( ChipCmd );
	if (tmp8 & CmdTxEnb){
		RTL_W8 ( ChipCmd, tmp8 & ~CmdTxEnb);
	}

	/* Disable interrupts by clearing the interrupt mask. */
	RTL_W16 ( IntrMask, 0x0000);

	/* Stop a shared interrupt from scavenging while we are. */
	spin_lock_irq (&priv->lock);
	rtl8169_tx_clear (priv);
	spin_unlock_irq (&priv->lock);


	rtl8169_hw_start (dev);

	netif_wake_queue (dev);
}







//======================================================================================================
static int rtl8169_start_xmit (struct sk_buff *skb, struct net_device *dev)
{
	struct rtl8169_private *priv = dev->priv;
	unsigned long ioaddr = priv->ioaddr;
#ifdef M_NH230
#else
	struct pci_dev *pdev = priv->pci_dev;
#endif
	int entry = priv->cur_tx % NUM_TX_DESC;
	int buf_len = 60;
	dma_addr_t txbuf_dma_addr;

	spin_lock_irq (&priv->lock);

	if( (le32_to_cpu(priv->TxDescArray[entry].status) & OWNbit)==0 ){

		priv->Tx_skbuff[entry] = skb;

#ifdef M_NH230
		txbuf_dma_addr = virt_to_bus(skb->data);
#else
		txbuf_dma_addr = pci_map_single(pdev, skb->data, skb->len, PCI_DMA_TODEVICE);
#endif
		
		priv->TxDescArray[entry].buf_addr = cpu_to_le32(txbuf_dma_addr);
		DBG_PRINT("%s: TX pkt_size = %d\n", __FUNCTION__, skb->len);
		if( skb->len <= priv->tx_pkt_len ){
			buf_len = skb->len;
		}
		else{
			printk("%s: Error -- Tx packet size(%d) > mtu(%d)+14\n", dev->name, skb->len, dev->mtu);
			buf_len = priv->tx_pkt_len;
		}

		if( entry != (NUM_TX_DESC-1) ){
			priv->TxDescArray[entry].status = cpu_to_le32((OWNbit | FSbit | LSbit) | buf_len);
		}
		else{
			priv->TxDescArray[entry].status = cpu_to_le32((OWNbit | EORbit | FSbit | LSbit) | buf_len);
		}

#ifdef M_NH230
  #ifdef RTL8169_CHECKSUM_BY_HW
  		if (M_IS_CHECKSUM_BY_HW_ENABLED(priv)) { // Is Checksum by H/W enabled ?
			if (skb->ip_summed == CHECKSUM_HW) {
				const struct iphdr *ip = skb->nh.iph;
				if (ip->protocol == IPPROTO_TCP)
					priv->TxDescArray[entry].status |= cpu_to_le32(IPCS | TCPCS);
				else if (ip->protocol == IPPROTO_UDP)
					priv->TxDescArray[entry].status |= cpu_to_le32(IPCS | UDPCS);
				else {
					printk("%s:error: ip_summed is CHECKSUM_HW, but not TCP/UDP\n", __FUNCTION__);
					BUG();
				}
			}
		}
  #endif
#endif

#ifdef M_NH230
#else
		pci_dma_sync_single(pdev, priv->txdesc_array_dma_addr[entry], sizeof(struct TxDesc), PCI_DMA_TODEVICE);
#endif

		RTL_W8 ( TxPoll, 0x40);		//set polling bit

		dev->trans_start = jiffies;

		priv->stats.tx_bytes += ( (skb->len > ETH_ZLEN) ? skb->len : ETH_ZLEN);
		priv->cur_tx++;
	}//end of if( (priv->TxDescArray[entry].status & 0x80000000)==0 )

	spin_unlock_irq (&priv->lock);

	if ( (priv->cur_tx - NUM_TX_DESC) == priv->dirty_tx ){  // Is Tx ring full ?
		M_INC_COUNTER(priv->st_counter.dw_tx_stopped);
		M_FDBG("error: stop Tx Queue as Tx Descriptor Ring full\n");
		netif_stop_queue (dev);
	}
	else{
		if (netif_queue_stopped (dev)){
			M_INC_COUNTER(priv->st_counter.dw_tx_started);
			M_FDBG("error: start Tx Queue as Tx queue stopped\n");
			netif_wake_queue (dev);
		}
	}

	return 0;
}







//======================================================================================================
static void rtl8169_tx_interrupt (struct net_device *dev, struct rtl8169_private *priv, unsigned long ioaddr)
{
	unsigned long dirty_tx, tx_left=0;
	int entry = priv->cur_tx % NUM_TX_DESC;
    	int txloop_cnt = 0;
#ifdef CONFIG_R8169_NAPI
	unsigned long dw_index;
#endif

	assert (dev != NULL);
	assert (priv != NULL);
	assert (ioaddr != NULL);

	M_INC_COUNTER(priv->st_counter.dw_tx_poll);
	M_FDBG("called\n");

	dirty_tx = priv->dirty_tx;
	tx_left = priv->cur_tx - dirty_tx;

#ifdef M_NH230
	// We should go through all Tx Descriptors, otherwise, Tx Queue will be full.
	while( (tx_left > 0) && (txloop_cnt < NUM_TX_DESC) ) {
#else
	while( (tx_left > 0) && (txloop_cnt < max_interrupt_work) ){
#endif
		if( (le32_to_cpu(priv->TxDescArray[entry].status) & OWNbit) == 0 ){

#ifdef RTL8169_DYNAMIC_CONTROL
			r8169_callback_tx(&(priv->rt), 1, priv->Tx_skbuff[dirty_tx % NUM_TX_DESC]->len);
#endif //end #ifdef RTL8169_DYNAMIC_CONTROL

#ifdef CONFIG_R8169_NAPI
			dw_index = dirty_tx % NUM_TX_DESC;
  			dev_kfree_skb( priv->Tx_skbuff[dw_index] );
			priv->Tx_skbuff[dw_index] = NULL;
#else
			dev_kfree_skb_irq( priv->Tx_skbuff[dirty_tx % NUM_TX_DESC] );
			priv->Tx_skbuff[dirty_tx % NUM_TX_DESC] = NULL;
#endif
			priv->stats.tx_packets++;
			dirty_tx++;
			tx_left--;
			entry++;
			M_INC_COUNTER(priv->st_counter.dw_tx_pkts);
		}
		txloop_cnt ++;
	}

	if (priv->dirty_tx != dirty_tx) {
		priv->dirty_tx = dirty_tx;
		if (netif_queue_stopped (dev)) {
			M_INC_COUNTER(priv->st_counter.dw_tx_started);
			M_FDBG("error: start Tx Queue as Tx Queue stopped\n");
			netif_wake_queue (dev);
		}
	}

	M_FDBG("return with txloop_cnt = %d\n", txloop_cnt);
}






//======================================================================================================
#ifdef CONFIG_R8169_NAPI
  static int rtl8169_rx_interrupt (struct net_device *dev, struct rtl8169_private *priv, unsigned long ioaddr)
#else
  static void rtl8169_rx_interrupt (struct net_device *dev, struct rtl8169_private *priv, unsigned long ioaddr)
#endif
{
#ifdef M_NH230
#else
	struct pci_dev *pdev = priv->pci_dev;
#endif
	int cur_rx;
	int pkt_size = 0 ;
	int rxdesc_cnt = 0;
#ifdef CONFIG_R8169_NAPI
#else
	int ret;
#endif
	struct sk_buff *n_skb = NULL;
	struct sk_buff *cur_skb;
	struct sk_buff *rx_skb;
	struct	RxDesc	*rxdesc;

#ifdef CONFIG_R8169_NAPI
	int received = 0;
	int rx_work_limit = dev->quota;	
	priv->c_rx_int_done = 1;   // 1 to assume Rx Descriptors is all processed
	M_FDBG("called, rx_work_limit = %d\n", rx_work_limit);
#endif

	M_INC_COUNTER(priv->st_counter.dw_rx_poll);

	assert (dev != NULL);
	assert (priv != NULL);
	assert (ioaddr != NULL);


	cur_rx = priv->cur_rx;

	rxdesc = &priv->RxDescArray[cur_rx];

#ifdef M_NH230
#else
	pci_dma_sync_single(pdev, priv->rxdesc_array_dma_addr[cur_rx], sizeof(struct RxDesc), PCI_DMA_FROMDEVICE);
#endif

#ifdef CONFIG_R8169_NAPI
	while ( ((le32_to_cpu(rxdesc->status) & OWNbit)== 0) && (rxdesc_cnt < NUM_RX_DESC) ){
	  M_FDBG("received=%d, rx_work_limit=%d, cur_rx=%d, priv->cur_rx=%d, rxdesc_cnt=%d\n", 
	  	  received, rx_work_limit, cur_rx, priv->cur_rx, rxdesc_cnt);
#else
	while ( ((le32_to_cpu(rxdesc->status) & OWNbit)== 0) && (rxdesc_cnt < max_interrupt_work) ){
#endif
	    rxdesc_cnt++;
	    M_INC_COUNTER(priv->st_counter.dw_rx_pkts);

	    if( le32_to_cpu(rxdesc->status) & RxRES ){
			printk(KERN_INFO "%s: Rx ERROR: status=0x%X\n", dev->name, le32_to_cpu(rxdesc->status));
			priv->stats.rx_errors++;
			if ( le32_to_cpu(rxdesc->status) & (RxRWT|RxRUNT) )
				priv->stats.rx_length_errors++;
			if ( le32_to_cpu(rxdesc->status) & RxCRC)
				priv->stats.rx_crc_errors++;
	    }
	    else{

#ifdef CONFIG_R8169_NAPI
		if (--rx_work_limit < 0) { /* we got packets, but no quota */
			M_INC_COUNTER(priv->st_counter.dw_rx_no_quota);
			priv->c_rx_int_done = 0; // Rx descriptors is not all processed
			M_FDBG("A2: priv->c_rx_int_done = %d\n", priv->c_rx_int_done);
                        break;
		}	    
		received++;
#endif
			pkt_size=(int)(le32_to_cpu(rxdesc->status) & 0x00001FFF)-4;

			if( pkt_size > priv->rx_pkt_len ){
				printk("%s: Error -- Rx packet size(%d) > mtu(%d)+14\n", dev->name, pkt_size, dev->mtu);
				pkt_size = priv->rx_pkt_len;
			}

			DBG_PRINT("%s: RX pkt_size = %d\n", __FUNCTION__, pkt_size);

			{// -----------------------------------------------------
				rx_skb = priv->Rx_skbuff[cur_rx];
#ifdef M_RX_SKB_POOL
				if (priv->st_skb_pool.dw_qty && priv->st_skb_pool.lpst_head) {
				  M_DBG("%s: SKB pool: get SKB\n", __FUNCTION__);
				  // assign from SKB pool
				  n_skb = priv->st_skb_pool.lpst_head->lpst_skb;
				  
				  // point to next
				  struct st_skb_pool_entry *lpst_curr = priv->st_skb_pool.lpst_head->lpst_next;
				  kfree (priv->st_skb_pool.lpst_head);  // free entry only
				  priv->st_skb_pool.lpst_head = lpst_curr;
				  
				  // sub 1 as assigned
				  priv->st_skb_pool.dw_qty--;
				  M_DBG("%s: SKB pool: Q'ty = %d\n", __FUNCTION__, priv->st_skb_pool.dw_qty);
				}
				else {
				  M_DBG("%s: SKB pool empty: allocate SKB\n", __FUNCTION__);
				  n_skb = RTL8169_ALLOC_RXSKB(MAX_RX_SKBDATA_SIZE);
				}
#else
				n_skb = RTL8169_ALLOC_RXSKB(MAX_RX_SKBDATA_SIZE);
#endif
				if( n_skb != NULL ) {
					skb_reserve (n_skb, 2);	// 16 byte align the IP fields. //

					// Indicate rx_skb
					if( rx_skb != NULL ){
						rx_skb->dev = dev;
#ifdef M_NH230
#else
						pci_dma_sync_single(pdev, priv->rx_skbuff_dma_addr[cur_rx], sizeof(struct RxDesc), PCI_DMA_FROMDEVICE);
#endif

#ifdef M_NH230
  #ifdef RTL8169_CHECKSUM_BY_HW
						/* Handle checksum offloading for incoming packets. */
						if (M_IS_CHECKSUM_BY_HW_ENABLED(priv)) { // Is Checksum by H/W enabled ?
							if (cp_rx_csum_ok(le32_to_cpu(rxdesc->status)))
								rx_skb->ip_summed = CHECKSUM_UNNECESSARY;
							else
								rx_skb->ip_summed = CHECKSUM_NONE;
						}
  #endif
#endif
						skb_put ( rx_skb, pkt_size );
						rx_skb->protocol = eth_type_trans ( rx_skb, dev );
#ifdef CONFIG_R8169_NAPI
						if (rtl8169_rx_skb(rx_skb) != NET_RX_SUCCESS)
						  M_INC_COUNTER(priv->st_counter.dw_rx_dropped);
#else
						ret = RTL8169_NETIF_RX (rx_skb);
#endif

//						dev->last_rx = jiffies;
						priv->stats.rx_bytes += pkt_size;
						priv->stats.rx_packets++;

#ifdef RTL8169_DYNAMIC_CONTROL
						r8169_callback_rx( &(priv->rt), 1, pkt_size);
#endif //end #ifdef RTL8169_DYNAMIC_CONTROL

					}//end if( rx_skb != NULL )

					priv->Rx_skbuff[cur_rx] = n_skb;
				}
				else{
					DBG_PRINT("%s: Allocate n_skb failed!\n",__FUNCTION__ );
					priv->Rx_skbuff[cur_rx] = rx_skb;
				}


				// Update rx descriptor
				if( cur_rx == (NUM_RX_DESC-1) ){
					priv->RxDescArray[cur_rx].status  = cpu_to_le32((OWNbit | EORbit) | (unsigned long)priv->hw_rx_pkt_len);
				}
				else{
					priv->RxDescArray[cur_rx].status  = cpu_to_le32(OWNbit | (unsigned long)priv->hw_rx_pkt_len);
				}

				cur_skb = priv->Rx_skbuff[cur_rx];

				if( cur_skb != NULL ){
#ifdef M_NH230
					priv->rx_skbuff_dma_addr[cur_rx] = virt_to_bus(cur_skb->data);
#else
					priv->rx_skbuff_dma_addr[cur_rx] = pci_map_single(pdev, cur_skb->data, MAX_RX_SKBDATA_SIZE, PCI_DMA_FROMDEVICE);
#endif
					rxdesc->buf_addr = cpu_to_le32(priv->rx_skbuff_dma_addr[cur_rx]);
				}
				else{
					DBG_PRINT("%s: %s() cur_skb == NULL\n", dev->name, __FUNCTION__);
				}

			}//------------------------------------------------------------

	    }// end of if( priv->RxDescArray[cur_rx].status & RxRES )

#ifdef M_NH230
	    cur_rx += 1;
	    if (cur_rx >= NUM_RX_DESC)
	      cur_rx -= NUM_RX_DESC;
#else
	    cur_rx = (cur_rx +1) % NUM_RX_DESC;
#endif
	    
	    rxdesc = &priv->RxDescArray[cur_rx];

#ifdef M_NH230
#else
	    pci_dma_sync_single(pdev, priv->rxdesc_array_dma_addr[cur_rx], sizeof(struct RxDesc), PCI_DMA_FROMDEVICE);
#endif
	}// end of while ( (priv->RxDescArray[cur_rx].status & 0x80000000)== 0)

	if( rxdesc_cnt >= max_interrupt_work ){
		DBG_PRINT("%s: Too much work at Rx interrupt.\n", dev->name);
	}

	priv->cur_rx = cur_rx;

#ifdef M_RX_SKB_POOL
        // Hook a timer to re-allocate SKB pool if empty
	if (priv->st_skb_pool.dw_qty == 0) {             // SKB pool is empty ?
	
	  // Hoot it if timer is not active
	  if (priv->c_skb_pool_timer_active_flag == 0) { // not hooked yet ?
	    printk("%s: queu task to re-allocate SKB pool as empry\n", __FUNCTION__);
            queue_task(&priv->rx_bh, &tq_immediate);
            mark_bh(IMMEDIATE_BH);
	    priv->c_skb_pool_timer_active_flag = 1;      // set as hooked
	  }
	}
#endif
	
#ifdef CONFIG_R8169_NAPI
	if ((le32_to_cpu(rxdesc->status) & OWNbit)== 0) {	  
	  priv->c_rx_int_done = 0;
	  M_FDBG("A3: priv->c_rx_int_done = %d\n", priv->c_rx_int_done);
	}
	  
	M_FDBG("return with received=%d, priv->c_rx_int_done = %d\n", received, priv->c_rx_int_done);
	return (received);
#endif
}








//======================================================================================================
/* The interrupt handler does all of the Rx thread work and cleans up after the Tx thread. */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
static void rtl8169_interrupt (int irq, void *dev_instance, struct pt_regs *regs)
#else
static irqreturn_t rtl8169_interrupt (int irq, void *dev_instance, struct pt_regs *regs)
#endif
{
	struct net_device *dev = (struct net_device *) dev_instance;
	struct rtl8169_private *priv = dev->priv;
	int boguscnt = max_interrupt_work;
	unsigned long ioaddr = priv->ioaddr;
	int status = 0;
	irqreturn_t interrupt_handled = IRQ_NONE;

#ifdef M_HANK_PROFILE_NOT_ENABLED
	// print out the following data between current and previous interrupt :
	//  1) offset of time of interrupt triggered
	//  2) NO# bytes of packets received 
	extern int nr_active_pages, nr_inactive_pages;
	if (gdw_eth_tick != 0 && gdw_eth_rx_bytes != 0 && gdw_eth_tx_bytes != 0)
	  printk("E: %lu x 10ms, Rx/Tx  %lu/%lu packets, %lu/%lu bytes, page cache: act %d, inact %d\n", 
	  		jiffies - gdw_eth_tick, 
			priv->stats.rx_packets - gdw_eth_rx_pkts,
			priv->stats.tx_packets - gdw_eth_tx_pkts,
	  		priv->stats.rx_bytes - gdw_eth_rx_bytes,
			priv->stats.tx_bytes - gdw_eth_tx_bytes,
			nr_active_pages, nr_inactive_pages);
	
	// update the time and NO# of bytes received/sent from/to Ethernet
	gdw_eth_tick = jiffies;
	gdw_eth_rx_bytes = priv->stats.rx_bytes;
	gdw_eth_tx_bytes = priv->stats.tx_bytes;
	
	gdw_eth_rx_pkts  = priv->stats.rx_packets;
	gdw_eth_tx_pkts  = priv->stats.tx_packets;
#endif

	M_FDBG("called\n");
	M_INC_COUNTER(priv->st_counter.dw_int_in);

#ifdef CONFIG_R8169_NAPI
#else
	RTL_W16 ( IntrMask, 0x0000);
#endif

	do {
		status = RTL_R16(IntrStatus);

		if (status == 0xFFFF)
			break;

#ifdef M_HANK_PROFILE
		if (status & RxOK)
			M_INC_COUNTER(priv->st_counter.dw_RxOk_int);
		if (status & RxOverflow)
			M_INC_COUNTER(priv->st_counter.dw_RxOverFlow_int);
		if (status & RxFIFOOver)
			M_INC_COUNTER(priv->st_counter.dw_RxFifoOver_int);
		if (status & TxOK)
			M_INC_COUNTER(priv->st_counter.dw_TxOk_int);
		if (status & TxErr)
			M_INC_COUNTER(priv->st_counter.dw_TxErr_int);
		if ((status & rtl8169_napi_event) == 0)
			M_INC_COUNTER(priv->st_counter.dw_other_int);
#endif


#ifdef CONFIG_R8169_NAPI
		// Do not ack the interrupts related to NAPI
		RTL_W16( IntrStatus, status & ~rtl8169_napi_event);
		////status &= priv->intr_mask;
#else
		RTL_W16( IntrStatus, status );
#endif

		if ( (status & rtl8169_intr_mask ) == 0 )
			break;
		else
			interrupt_handled = IRQ_HANDLED;

#ifdef CONFIG_R8169_NAPI
		M_FDBG("status = 0x%X\n", status);
		if (likely(netif_rx_schedule_prep(dev))) {
			RTL_W16(IntrMask, rtl8169_intr_mask & ~rtl8169_napi_event);
			priv->intr_mask = ~rtl8169_napi_event;

			M_FDBG("call __netif_rx_schedule(), status=0x%X\n", status);
			__netif_rx_schedule(dev);
		}
		else {
			M_FDBG(KERN_INFO "%s: interrupt %04x taken in poll\n", dev->name, status);
			M_INC_COUNTER(priv->st_counter.dw_inpoll_int);
		}
		break;
#else

		// Rx interrupt
//		if (status & (RxOK | RxErr /* | LinkChg | RxOverflow | RxFIFOOver*/)){
			rtl8169_rx_interrupt (dev, priv, ioaddr);
//		}

		// Tx interrupt
//		if (status & (TxOK | TxErr)) {
			spin_lock (&priv->lock);
			rtl8169_tx_interrupt (dev, priv, ioaddr);
			spin_unlock (&priv->lock);
//		}

#endif

		boguscnt--;
	} while (boguscnt > 0);

#ifdef CONFIG_R8169_NAPI
#else
	if (boguscnt <= 0) {
		DBG_PRINT("%s: Too much work at interrupt!\n", dev->name);
		RTL_W16( IntrStatus, 0xffff);	// Clear all interrupt sources
	}
#endif

#ifdef CONFIG_R8169_NAPI
#else
	RTL_W16 ( IntrMask, rtl8169_intr_mask);
#endif
	M_INC_COUNTER(priv->st_counter.dw_int_out);

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,0)
	return interrupt_handled;
#endif
}


#ifdef CONFIG_R8169_NAPI
static int rtl8169_poll(struct net_device *dev, int *budget)
 {
	unsigned int work_done = 0;
 	struct rtl8169_private *priv = dev->priv;
	unsigned long ioaddr = priv->ioaddr;
	M_INC_COUNTER(priv->st_counter.dw_poll);
 
redo_napi:
	// Ack the NAPI related interrupts here
	RTL_W16( IntrStatus, rtl8169_napi_event);
 	M_FDBG("called: budget=%d, work_done=%d, dev->quota=%d\n", *budget, work_done, dev->quota);

 	if (1)                          // pkts in Rx descriptor ring ?
	  work_done = rtl8169_rx_interrupt(dev, priv, ioaddr);
	  
	if (M_IS_PKTS_IN_TX_RING())     // pkts in Tx descriptor ring ? 
	  rtl8169_tx_interrupt(dev, priv, ioaddr);

	*budget -= work_done;
	dev->quota -= work_done;

	if (dev->quota > 0) {           // Is there any quota left ?
		if (M_IS_PKTS_IN_TX_RING() || !priv->c_rx_int_done) {
		  M_FDBG("redo_napi: dev->quota = %d, priv->c_rx_int_done = %d\n", dev->quota, priv->c_rx_int_done);
		  M_INC_COUNTER(priv->st_counter.dw_redo_poll);
		  goto redo_napi;
		}
		
		netif_rx_complete(dev);
		priv->intr_mask = 0xffff;
		/*
		 * 20040426: the barrier is not strictly required but the
		 * behavior of the irq handler could be less predictable
		 * without it. Btw, the lack of flush for the posted pci
		 * write is safe - FR
		 */
#ifdef M_ORIGINAL_CODES
		smp_wmb();
#endif

		RTL_W16(IntrMask, rtl8169_intr_mask);
	
		M_FDBG("return with 0 : dev->quota = %d\n", dev->quota);
		M_INC_COUNTER(priv->st_counter.dw_poll_done);
		return (0);             // Done
	}

	M_FDBG("quit with budget=%d, dev->quota=%d, work_done=%d\n", 
		*budget, dev->quota, work_done);
	M_FDBG("return with 1 : dev->quota = %d\n", dev->quota);

	M_INC_COUNTER(priv->st_counter.dw_poll_not_done);
	return (1);                     // Not done
}
#endif



//======================================================================================================
static int rtl8169_close (struct net_device *dev)
{
	struct rtl8169_private *priv = dev->priv;
	unsigned long ioaddr = priv->ioaddr;
	int i;

	// -----------------------------------------
	rtl8169_delete_timer( &(priv->r8169_timer) );


	netif_stop_queue (dev);

	spin_lock_irq (&priv->lock);

	/* Stop the chip's Tx and Rx processes. */
	RTL_W8 ( ChipCmd, 0x00);

	/* Disable interrupts by clearing the interrupt mask. */
	RTL_W16 ( IntrMask, 0x0000);

	/* Update the error counts. */
	priv->stats.rx_missed_errors += RTL_R32(RxMissed);
	RTL_W32( RxMissed, 0);

	spin_unlock_irq (&priv->lock);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	synchronize_irq ();
#else
	synchronize_irq (dev->irq);
#endif
	free_irq (dev->irq, dev);

	rtl8169_tx_clear (priv);
	
	//2004-05-11
	if(priv->txdesc_space != NULL){
		pci_free_consistent(
				priv->pci_dev,
				priv->sizeof_txdesc_space,
				priv->txdesc_space,
				priv->txdesc_phy_dma_addr
		);
		priv->txdesc_space = NULL;
	}

	if(priv->rxdesc_space != NULL){
		pci_free_consistent(
				priv->pci_dev,
				priv->sizeof_rxdesc_space,
				priv->rxdesc_space,
				priv->rxdesc_phy_dma_addr
		);
		priv->rxdesc_space = NULL;
	}

	priv->TxDescArray = NULL;
	priv->RxDescArray = NULL;

	{//-----------------------------------------------------------------------------
		for(i=0;i<NUM_RX_DESC;i++){
			if( priv->Rx_skbuff[i] != NULL ) {
				RTL8169_FREE_RXSKB ( priv->Rx_skbuff[i] );
			}
		}
	}//-----------------------------------------------------------------------------

#ifdef M_RX_SKB_POOL
	// Free the SKB pool and its entries
	fn_skb_pool_handler(dev, CMD_SKB_POOL_FREE);
	priv->c_skb_pool_timer_active_flag = 1;
#endif
	
	DBG_PRINT("%s: %s() alloc_rxskb_cnt = %d\n", dev->name, __FUNCTION__, alloc_rxskb_cnt );

	return 0;
}







//======================================================================================================
static unsigned const ethernet_polynomial = 0x04c11db7U;
static inline u32 ether_crc (int length, unsigned char *data)
{
	int crc = -1;

	while (--length >= 0) {
		unsigned char current_octet = *data++;
		int bit;
		for (bit = 0; bit < 8; bit++, current_octet >>= 1)
			crc = (crc << 1) ^ ((crc < 0) ^ (current_octet & 1) ? ethernet_polynomial : 0);
	}

	return crc;
}








//======================================================================================================
static void rtl8169_set_rx_mode (struct net_device *dev)
{
	struct rtl8169_private *priv = dev->priv;
	unsigned long ioaddr = priv->ioaddr;
	unsigned long flags;
	u32 mc_filter[2];	/* Multicast hash filter */
	int i, rx_mode;
	u32 tmp=0;
	
	if (dev->flags & IFF_PROMISC) {
		/* Unconditionally log net taps. */
		printk (KERN_NOTICE "%s: Promiscuous mode enabled.\n", dev->name);
		rx_mode = AcceptBroadcast | AcceptMulticast | AcceptMyPhys | AcceptAllPhys;
		mc_filter[1] = mc_filter[0] = 0xffffffff;
	} else if ((dev->mc_count > multicast_filter_limit) || (dev->flags & IFF_ALLMULTI)) {
		/* Too many to filter perfectly -- accept all multicasts. */
		rx_mode = AcceptBroadcast | AcceptMulticast | AcceptMyPhys;
		mc_filter[1] = mc_filter[0] = 0xffffffff;
	} else {
		struct dev_mc_list *mclist;
		rx_mode = AcceptBroadcast | AcceptMulticast | AcceptMyPhys;
		mc_filter[1] = mc_filter[0] = 0;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
		for (i = 0, mclist = dev->mc_list; mclist && i < dev->mc_count; i++, mclist = mclist->next)
		{
			set_bit (ether_crc (ETH_ALEN, mclist->dmi_addr) >> 26, mc_filter);
		}			
#else
		for (i = 0, mclist = dev->mc_list; mclist && i < dev->mc_count; i++, mclist = mclist->next)
		{
			int bit_nr = ether_crc(ETH_ALEN, mclist->dmi_addr) >> 26;
			                                                                                                     
			mc_filter[bit_nr >> 5] |= 1 << (bit_nr & 31);
			rx_mode |= AcceptMulticast;
		}
#endif		
	}

	spin_lock_irqsave (&priv->lock, flags);

	tmp = rtl8169_rx_config | rx_mode | (RTL_R32(RxConfig) & rtl_chip_info[priv->chipset].RxConfigMask);

	RTL_W32 ( RxConfig, tmp);
	RTL_W32 ( MAR0 + 0, mc_filter[0]);
	RTL_W32 ( MAR0 + 4, mc_filter[1]);

	spin_unlock_irqrestore (&priv->lock, flags);

}//end of rtl8169_set_rx_mode (struct net_device *dev)




//================================================================================
struct net_device_stats *rtl8169_get_stats(struct net_device *dev)

{
	struct rtl8169_private *priv = dev->priv;

#ifdef M_HANK_DEBUG
  M_DUMP_CPU_REG();                   // dump CPU Registers
  M_DUMP_PCI_REG(dev);                // dump PCI Configuration Registers
  M_DUMP_REG(dev);                    // dump MAC Registers
#endif
	
#ifdef M_HANK_PROFILE
  // 1) Print out the profile counters
  spin_lock (&priv->lock);  
  if (1) {
    M_PRINT("\nProfile counters : for debugging only\n");
  
    M_PRINT("dw_int_in         = %10u, dw_int_out          = %10u\n", 
           priv->st_counter.dw_int_in, priv->st_counter.dw_int_out);
    M_PRINT("dw_inpoll_int     = %10u, dw_tx_int_handled   = %10u\n",  
           priv->st_counter.dw_inpoll_int, priv->st_counter.dw_tx_int_handled);
    M_PRINT("dw_other_int      = %10u, dw_RxOk_int         = %10u\n", 
           priv->st_counter.dw_other_int, priv->st_counter.dw_RxOk_int);
    M_PRINT("dw_RxOverFlow_int = %10u, dw_RxFifoOver_int   = %10u\n", 
           priv->st_counter.dw_RxOverFlow_int, priv->st_counter.dw_RxFifoOver_int);
    M_PRINT("dw_TxOk_int       = %10u, dw_TxErr_int        = %10u\n", 
           priv->st_counter.dw_TxOk_int, priv->st_counter.dw_TxErr_int); 
    M_PRINT("dw_poll           = %10u, dw_redo_poll        = %10u\n", 
           priv->st_counter.dw_poll, priv->st_counter.dw_redo_poll);
    M_PRINT("dw_tx_poll        = %10u, dw_tx_polling_timer = %10u\n", 
           priv->st_counter.dw_tx_poll, priv->st_counter.dw_tx_polling_timer);
    M_PRINT("dw_rx_poll        = %10u, dw_rx_no_quota      = %10u\n",  
           priv->st_counter.dw_rx_poll, priv->st_counter.dw_rx_no_quota);
    M_PRINT("dw_poll_done      = %10u, dw_poll_not_done    = %10u\n", 
           priv->st_counter.dw_poll_done, priv->st_counter.dw_poll_not_done);
    M_PRINT("dw_tx_pkts        = %10u, dw_rx_pkts          = %10u\n", 
           priv->st_counter.dw_tx_pkts, priv->st_counter.dw_rx_pkts);
    M_PRINT("dw_rx_dropped     = %10u, dw_tx_started       = %10u\n",
           priv->st_counter.dw_rx_dropped, priv->st_counter.dw_tx_started);
    M_PRINT("dw_tx_stopped     = %10u\n",  priv->st_counter.dw_tx_stopped);  
    M_PRINT("\n");
  }
  
  // 2) reset counters for profile
  memset(&(priv->st_counter), 0, sizeof(ST_COUNTER));
  
  spin_unlock (&priv->lock);
#endif

    return &priv->stats;
}








//================================================================================
static struct pci_driver rtl8169_pci_driver = {
#ifdef M_NH230   // 10/15/2004 : Hank Chu : modify to fix linker error
        .name        = MODULENAME,
        .id_table    = rtl8169_pci_tbl,
        .probe       = rtl8169_init_one,
        .remove      = __devexit_p(rtl8169_remove_one),
        .suspend     = NULL,
        .resume      = NULL,
#else
	name:		MODULENAME,
	id_table:	rtl8169_pci_tbl,
	probe:		rtl8169_init_one,
	remove:		rtl8169_remove_one,
	suspend:	NULL,
	resume:		NULL,
#endif
};





//======================================================================================================
static int __init rtl8169_init_module (void)
{
	return pci_module_init (&rtl8169_pci_driver);	// pci_register_driver (drv)
}




//======================================================================================================
static void __exit rtl8169_cleanup_module (void)
{
	pci_unregister_driver (&rtl8169_pci_driver);
}


#ifdef RTL8169_JUMBO_FRAME_SUPPORT
static int rtl8169_change_mtu(struct net_device *dev, int new_mtu)
{
	struct rtl8169_private *priv = dev->priv;
	unsigned long ioaddr = priv->ioaddr;

#ifdef M_NH230
	if( new_mtu > MAX_JUMBO_FRAME_MTU_ALLOWED ) {
#else
	if( new_mtu > MAX_JUMBO_FRAME_MTU ){
#endif
		printk("%s: Error -- new_mtu(%d) > MAX_JUMBO_FRAME_MTU(%d).\n", dev->name, new_mtu, MAX_JUMBO_FRAME_MTU);
		return -1;
	}

	dev->mtu = new_mtu;

#ifdef M_NH230
        gdw_mtu = dev->mtu;             // duplicate the new MTU

    #ifdef RTL8169_CHECKSUM_BY_HW
        if (new_mtu <= ETH_DATA_LEN)    // Is Jumbo frame disabled (normal MTU) ?
	  M_ENABLE_CHECKSUM_BY_HW(priv);
	else
	  M_DISABLE_CHECKSUM_BY_HW(priv);
    #endif
    
	M_FDBG("MTU set to %d, MAX_JUMBO_FRAME_MTU %d, NUM_TX_DESC=%d, NUM_RX_DESC=%d\n", 
		gdw_mtu, MAX_JUMBO_FRAME_MTU, NUM_TX_DESC, NUM_RX_DESC);
#endif
	
	priv->curr_mtu_size = new_mtu;
	priv->tx_pkt_len = new_mtu + ETH_HDR_LEN;
	priv->rx_pkt_len = new_mtu + ETH_HDR_LEN;
	priv->hw_rx_pkt_len = priv->rx_pkt_len + 8;

	RTL_W8 ( Cfg9346, Cfg9346_Unlock);
	RTL_W16	( RxMaxSize, (unsigned short)priv->hw_rx_pkt_len );
	RTL_W8 ( Cfg9346, Cfg9346_Lock);

	DBG_PRINT("-------------------------- \n");
	DBG_PRINT("dev->mtu = %d \n", dev->mtu);
	DBG_PRINT("priv->curr_mtu_size = %d \n", priv->curr_mtu_size);
	DBG_PRINT("priv->rx_pkt_len = %d \n", priv->rx_pkt_len);
	DBG_PRINT("priv->tx_pkt_len = %d \n", priv->tx_pkt_len);
	DBG_PRINT("RTL_W16( RxMaxSize, %d )\n", priv->hw_rx_pkt_len);
	DBG_PRINT("-------------------------- \n");

	rtl8169_close (dev);
	rtl8169_open (dev);

	return 0;
}
#endif //end #ifdef RTL8169_JUMBO_FRAME_SUPPORT


#ifdef M_RX_SKB_POOL

unsigned long gdw_save_flags;
static void RXbh(void *data)
{
#if 1
  struct net_device	*dev = data;
  struct rtl8169_private *priv = dev->priv;
  unsigned long ioaddr = priv->ioaddr;

  if (in_interrupt()) {                 // In interrupt ?
    printk("%s: INT# : return with INT# disabled: queue again\n", __FUNCTION__);
  
      
      local_irq_save(gdw_save_flags);
          // Lock it to avoid interrupt
    // Disable the interrupt of RTL8169S before re-allocate SKB pool
    RTL_W16(IntrMask, 0);            // disable it

    // Then queue it again
    queue_task(&priv->rx_bh, &tq_immediate);
    mark_bh(IMMEDIATE_BH);
    return;
  }
  
  
  // Re-allocate the SKB pool
  printk("%s: D0: called to re-allocate SKB pool, Q'ty = %d\n", 
  	__FUNCTION__, priv->st_skb_pool.dw_qty);
	
  fn_skb_pool_handler(dev, CMD_SKB_POOL_ALLOCATE);

  printk("%s: D1: re-allocate SKB pool : done , Q'ty = %d\n", 
  	__FUNCTION__, priv->st_skb_pool.dw_qty);
    
  // Reset flag as timer released
  priv->c_skb_pool_timer_active_flag = 0;
  
  // Enable the interrupt of RTL8169S again
  RTL_W16(IntrMask, rtl8169_intr_mask);
  local_irq_restore(gdw_save_flags);          // Lock it to avoid interrupt

#else
	struct net_device	*dev = data;
	MSP_DEVICE	*pDrvCtrl = dev->priv;
	unsigned long	recovery;

	if (pDrvCtrl == 0)	/* If shutting down, do nothing */
		return;

	recovery = xchg_u32(&pDrvCtrl->bh_recovery, 0);

	if (pDrvCtrl->reloadflag)
		return;

	MSPEthRxIntEnable(pDrvCtrl);

#ifdef ETH_LOOP_BACK
// == Hiram ==  
dev_ptr = dev;
#endif
	MSPEthReceive(pDrvCtrl);

	if (recovery)
	{
		MSPEthFdaFullIntEnable(pDrvCtrl);
		MSPEthBlFullIntEnable(pDrvCtrl);
	}

	if (! MACOwnsRxFD(pDrvCtrl->FDAPtrCur)) 
	{
		queue_task(pDrvCtrl->rx_bh, &tq_immediate);
		mark_bh(IMMEDIATE_BH);
	}
#endif
}


//
// Subject :
//   The routine handles the SKB pool, including :
//   1) allocate SKBs into the linked list
//     or
//   2) free all remaining SKBs in the linked list 
// Input :
//   *dev : pinter to net_device{}
//   c_cmd : command ID to take action
// Output :
//   0      : OK
//   others : failed
//
int fn_skb_pool_handler(struct net_device *dev, u8 c_cmd)
{
  struct rtl8169_private *priv = dev->priv;
  u32 dw_loop = 0, dw_count = 0;
  int w_result = 0;
  struct sk_buff *lpst_skb = NULL;
  ST_SKB_POOL_ENTRY *lpst_next = NULL, *lpst_curr = NULL;
 
  spin_lock_irq (&priv->lock);          // Lock it to avoid interrupt
  
  switch (c_cmd) {
    case CMD_SKB_POOL_ALLOCATE :
    
      // Allocate the entries and insert to the SKB pool for Rx untill full
      M_DBG("%s: alloc: Q'ty in SKB pool is %d\n", __FUNCTION__, priv->st_skb_pool.dw_qty);
      for (dw_loop = 0; priv->st_skb_pool.dw_qty < MAX_ENTRIES_OF_SKB_POOL; dw_loop++) {

      	// Alocate a SKB pool entry
	lpst_curr = kmalloc(sizeof(ST_SKB_POOL_ENTRY), GFP_KERNEL);
	if (NULL == lpst_curr) {
	  printk("%s: error: out of memory to allocate SKB entry\n", __FUNCTION__);
	  w_result = -ENOMEM;
	  break;
	}
	else {
	  lpst_curr->lpst_next = NULL;  // clean up
	  lpst_curr->lpst_skb = NULL;   //
	}

	// Allocate a SKB
	lpst_skb = RTL8169_ALLOC_RXSKB(MAX_RX_SKBDATA_SIZE);
	if (NULL == lpst_skb) {
	  kfree (lpst_curr);        // free SKB entry allocated above
	  printk("%s: error: out of memory to allocate SKB\n", __FUNCTION__);
	  w_result = -ENOMEM;
	  break;
	}
	else
	  lpst_curr->lpst_skb = lpst_skb;    // assign SKB into entry

	// If SKB pool is empty, assign to be head
	// If SKB pool is not empty, insert before the head for speed
        if (NULL == priv->st_skb_pool.lpst_head) { // Is SKB pool empty ?   
          priv->st_skb_pool.lpst_head = lpst_curr;
	  M_DBG("%s: loop[%d]: SKB pool is empry.\n", __FUNCTION__, dw_loop);
	}
	else {
	  // Duplicate the original head first
	  lpst_next = priv->st_skb_pool.lpst_head;
	  
	  // Change the head to new entry
	  priv->st_skb_pool.lpst_head = lpst_curr;
	  
	  // Assign the next of new head to original head
	  lpst_curr->lpst_next = lpst_next;
	  M_DBG("%s: loop[%d]: new head=%p, old head=%p\n", 
	  	__FUNCTION__, dw_loop, lpst_curr, lpst_next);
	}
	
	priv->st_skb_pool.dw_qty++;     // add 1 as entry inserted
	dw_count++;
	M_DBG("%s: loop[%d]: Q'ty=%d, head=%p\n", 
		__FUNCTION__, dw_loop, priv->st_skb_pool.dw_qty, priv->st_skb_pool.lpst_head);
      }
      
      M_DBG("%s: allocate %d entries in SKB pool.\n", __FUNCTION__, dw_count);
      break;
      
    case CMD_SKB_POOL_FREE :
    
      // Free the remaining entries in the SKB pool for Rx 
      M_DBG("%s: free: Q'ty in SKB pool is %d\n", __FUNCTION__, priv->st_skb_pool.dw_qty);
      
      if (MAX_ENTRIES_OF_SKB_POOL < priv->st_skb_pool.dw_qty)
        printk("%s: error: NO# of entries in SKB pool exceeds max. %d\n", 
		__FUNCTION__, MAX_ENTRIES_OF_SKB_POOL);
      
      for (dw_loop = 0; dw_loop < priv->st_skb_pool.dw_qty; dw_loop++) {
          if (NULL == priv->st_skb_pool.lpst_head) {
	    if (!dw_loop)
	      printk("%s: error: SKB pool is empty, but Q'ty %d\n", __FUNCTION__, priv->st_skb_pool.dw_qty);
	    break;
	  }
	  
	  M_DBG("%s: loop[%d]: head %p\n", __FUNCTION__, dw_loop, priv->st_skb_pool.lpst_head);
	  
	  // Duplicate the next entry
          lpst_next = priv->st_skb_pool.lpst_head->lpst_next;
	  
	  // Free the SKB allocated previously
	  if (priv->st_skb_pool.lpst_head->lpst_skb) {
	    RTL8169_FREE_RXSKB(priv->st_skb_pool.lpst_head->lpst_skb);
	  }
	  else
	    printk("%s: error: *skb is NULL\n", __FUNCTION__);	    	    
	  
	  // Free the entry of SKB pool
	  kfree(priv->st_skb_pool.lpst_head);
	  
	  // Point to next entry of SKB pool
	  priv->st_skb_pool.lpst_head = lpst_next;
	  dw_count++;
      }
      
      M_DBG("%s: free %d entries in SKB pool.\n", __FUNCTION__, dw_count);
      priv->st_skb_pool.dw_qty = 0;     // reset Q'ty to 0
      break;
      
    default :
      w_result = -EINVAL;
      break;
  }
  
  spin_unlock_irq (&priv->lock);        // Unlock it as done
  return (w_result);
}
#endif // end of #ifdef {M_RX_SKB_POOL}


#ifdef M_HANK_DEBUG
  #include <asm/io.h>                   // for __raw_readb()
  #include <asm/mpc10x.h>               // for MPC10X_MAPB_EUMB_BASE
  
  //
  // Subject :
  //   This routine is used for debugging only.
  //   It prints out the content of CPU registers.
  //
  void dbg_dump_CPU_reg (void)
  {
    M_DBG("**** Dump the CPU registers ****\n");

    // 1) Configuration registers  (offset: 0x00 ~ 0xFF)
    #define CFG_REG_ADDR_OFFSET  0x80000000
    u32 dw_loop = 0;
    u16 w_word = 0;
    M_DBG("Configuration registers: offset in Address Map B: 0x%X\n", CFG_REG_ADDR_OFFSET);
    
    // PCI configuration space access routines.
    unsigned int *pci_config_address = (unsigned int *)MPC10X_MAPB_CNFG_ADDR;
    unsigned char *pci_config_data   = (unsigned char *)MPC10X_MAPB_CNFG_DATA;
    u32 bus = 0, dev_fn = 0;            // PCI Bridge: bus 0, dev/func 0
   
    for (dw_loop = 0; dw_loop <= 0x0FF; dw_loop += 2) {
    
      if (dw_loop && !(dw_loop & 0x0f)) // Change line
        M_DBG("\n");     
    
      if ( !(dw_loop & 0x0f) )          // Print offset
        M_DBG("%08X: ", dw_loop);
    
      if (dw_loop && ((dw_loop & 0x0f) == 0x8)) // Insert '-' if 4 registers printed
        M_DBG(" - ");

      // Write 0x800000nn to CONFIG_ADDR port in word-aligned address
      out_le32(pci_config_address,
         (((bus & 0xff)<<16) | (dev_fn<<8) | (dw_loop & 0xfc) | CFG_REG_ADDR_OFFSET));

      // Read data from CONFIG_DAT port
      w_word = in_le16((unsigned short *)(pci_config_data + (dw_loop & 3)));

      M_DBG("%04X ", w_word);           // Print 2-digit for 1-byte value
    }

    M_DBG("\n\n");

#ifdef M_04_26_2005   // Hank: disable the dumping for EMU registers as too much
    // 2) EUMB registers (EMUB base + 0x00000 ~ 0xFFFFF)
    u32 dw_offset = MPC10X_MAPB_EUMB_BASE;
    u8 c_byte = 0;
    M_DBG("EUMB registers: offset in Address Map B: 0x%X\n", dw_offset);

    for (dw_loop = 0; dw_loop <= (u32)(0x0FFFFF); dw_loop++) {

      if (dw_loop && !(dw_loop & 0x0f)) // Change line
        M_DBG("\n");    
    
      if ( !(dw_loop & 0x0f) )          // Print offset
        M_DBG("%08X: ", dw_loop);
    
      if (dw_loop && ((dw_loop & 0x0f) == 0x8)) // Insert '-' if 8 registers printed
        M_DBG(" - ");

	
      c_byte = *((volatile unsigned char *)(dw_offset + dw_loop));
      M_DBG("%02X ", c_byte);           // Print 2-digit for 1-byte value
    }
    
    M_DBG("\n\n");    
#endif
  }


  //
  // Subject :
  //   This routine is used for debugging only.
  //   It prints out the content of RTL8169S registers in byte.
  //
  void dbg_dump_PCI_reg (struct net_device *dev)
  {
    struct rtl8169_private *priv = dev->priv;
    struct pci_dev *pdev = priv->pci_dev;
    u8 c_byte = 0;  

    // Dump the contents of registers
    M_DBG("\n RTL8169S : Dumpe the PCI Configuration Registers\n");
    int w_loop = 0;
    for (w_loop = 0; w_loop <= 0x0ff; w_loop++) {
    	
      if (w_loop && (!(w_loop & 0x0f)) )
        M_DBG("\n");
        
      if (!(w_loop & 0x0f))
        M_DBG("%02X: ", w_loop);     
     
      pci_read_config_byte(pdev, w_loop, &c_byte);
      M_DBG("%02X ", c_byte);
      
      if ((w_loop & 0x0f) == 0x07)
        M_DBG(" - ");
    }
    
    M_DBG("\n");
  }


  //
  // Subject :
  //   This routine is used for debugging only.
  //   It prints out the content of registers in byte.
  //
  void dbg_dump_reg (struct net_device *dev)
  {
    struct rtl8169_private *priv = dev->priv;
    unsigned long ioaddr = priv->ioaddr;
	
    // Dump the contents of registers
    M_DBG("\n RTL8169S : dumpe the MAC registers\n");
    int w_loop = 0;
    for (w_loop = 0; w_loop <= 0x0ff; w_loop++) {
    	
      if (w_loop && (!(w_loop & 0x0f)) )
        M_DBG("\n");
        
      if (!(w_loop & 0x0f))
        M_DBG("%02X: ", w_loop);

      M_DBG("%02X ", RTL_R8(MAC0 + w_loop));
      
      if ((w_loop & 0x0f) == 0x07)
        M_DBG(" - ");	    
    }
    
    M_DBG("\n");
  }

#endif // end of #ifdef {M_HANK_DEBUG}








//======================================================================================================
module_init(rtl8169_init_module);
module_exit(rtl8169_cleanup_module);
