//
// Subject :
//   The header file provides those constants, macro, and structures related to NH-230/231.
//
// History :
//   v1.00  01/19/2005  Hank Chu
//   1) Create the GPIO pins (GPIO bit[0:7]) for NH-230/231 NAS.
//   2) Provide the API routines to read/write GPIO pins via Extended ROM/Flash space.
//
//   v1.01  08/04/2005  Hank Chu
//   1) Define the GPIO bit-3 for FAN control.
//      GPIO bit-3 : set to low for low speed, high for high speed.
//
//   v1.02  08/05/2005  Hank Chu
//   1) Use GPIO bit[6:7] for USB port1/2 LEDs.
//
//
// Notes :
//   1) Work only under Linux kernel mode.
//
//
#ifdef __KERNEL__
  #ifndef __NH230_H__
    #define __NH230_H__


    //**************************************
    //  GPIO pins
    //**************************************

    // The bit definition of GPIO pins for writting only
#define GPIO_PIN_WR_SYS_SHUTDOWN_EN  0x01  // bit[0]: high active: perform system shutdown
#define GPIO_PIN_WR_SYS_RESET_EN     0x02  // bit[1]: high active: assert System Reset to all devices
#define GPIO_PIN_WR_STATUS_LED_DIS   0x04  // bit[2]: low active: turn on status LED
#define GPIO_PIN_WR_FAN_CTRL_HIGH    0x08  // bit[3]: low for low speed, high for high speed
#define GPIO_PIN_WR_DEBUG_MASK       0x030 // bit[4:5]: high-active: LEDs for debugging
#define GPIO_PIN_WR_USB1_LED_DIS     0x040 // bit[6]: low active to USB port-1 LED
#define GPIO_PIN_WR_USB2_LED_DIS     0x080 // bit[7]: low active to USB port-2 LED

#define GPIO_PIN_ALL   0x0ff
#define GPIO_PIN_SYS_SHUTDOWN_ON (GPIO_PIN_WR_SYS_SHUTDOWN_EN | GPIO_PIN_WR_STATUS_LED_DIS)
                                  
#define GPIO_PIN_SYS_RESET_ON    (GPIO_PIN_WR_STATUS_LED_DIS | GPIO_PIN_WR_SYS_RESET_EN)
#define GPIO_PIN_STATUS_LED_ON   (0)
#define GPIO_PIN_STATUS_LED_OFF  (GPIO_PIN_WR_STATUS_LED_DIS)
                                  

    // The bit definition of GPIO pins for reading only (Low-active)
#define GPIO_PIN_RD_SYS_POWERDOWN    0x01  // bit[0]: low active : system is in power down
#define GPIO_PIN_RD_RESET_DEFAULT    0x02  // bit[1]: reset to default button is pressed
#define GPIO_PIN_RD_HW_VERSION_MASK  0x0f0 // bit[4:7]: H/W version (for NH-230 : 0b0000)

    
    // We reserved the 4KB space on top of Extended ROM space.
    // Initialized in boot-loader (cpu/mpc824x/cpu_init.c file: cpu_init_f() routine).
    // But just use the first byte for read/write to 8-bit GPIO pins.
    //
#define NH230_GPIO_ADDRESS  0x70000000    // address located on Address Map B

    // read in byte
#define NH230_GPIO_READB()        ( *((volatile unsigned char *)NH230_GPIO_ADDRESS) )

    // write in byte
void NH230_GPIO_WRITEB(unsigned char c_byte);  // API to write a byte to GPIO pins
#define __GPIO_WB(c_byte) ( *(volatile unsigned char *)NH230_GPIO_ADDRESS = c_byte)

  // Write Only : Perform the system shutdown via GPIO pin
#define NH230_GPIO_SYS_SHUTDOWN()     ( NH230_GPIO_WRITEB(GPIO_PIN_SYS_SHUTDOWN_ON) )

  // Write Only : Assert the System Reset to all devices via GPIO pin
#define NH230_GPIO_SYS_RESET()        ( NH230_GPIO_WRITEB(GPIO_PIN_SYS_RESET_ON) )

  // Write Only : Turn on/off the status LED via GPIO pin
#define NH230_GPIO_LED_STATUS_ON()    ( NH230_GPIO_WRITEB(GPIO_PIN_STATUS_LED_ON) )
#define NH230_GPIO_LED_STATUS_OFF()   ( NH230_GPIO_WRITEB(GPIO_PIN_STATUS_LED_OFF) )

  // Write Only : Set FAN to low/high speed
#define NH230_GPIO_FAN_CTRL_LOW()     ( NH230_GPIO_WRITEB(GPIO_PIN_STATUS_LED_ON))
#define NH230_GPIO_FAN_CTRL_HIGH()    ( NH230_GPIO_WRITEB(GPIO_PIN_STATUS_LED_ON | GPIO_PIN_WR_FAN_CTRL_HIGH))



  // Read Only : Is the system in power down ?
#define NH230_GPIO_IS_POWERDOWN()     ( !(NH230_GPIO_READB() & GPIO_PIN_RD_SYS_POWERDOWN) )

  // Read Only : Is the reset to default button pressed ?
#define NH230_GPIO_IS_RESET_DEFAULT() ( !(NH230_GPIO_READB() & GPIO_PIN_RD_RESET_DEFAULT) )

  // Read Only : get the H/W version
#define NH230_GPIO_HW_VERSION()       ( NH230_GPIO_READB() & GPIO_PIN_RD_HW_VERSION_MASK )





    //**************************************
    //  I2C
    //**************************************



  #endif /* __NH230_H__ */
#endif /* __KERNEL__ */

