/*
 * arch/ppc/kernel/mpc10x_common.c
 * 
 * Common routines for the Motorola SPS MPC106, MPC107 and MPC8240 Host bridge,
 * Mem ctlr, EPIC, etc.
 *
 * Author: Mark A. Greer
 *         mgreer@mvista.com
 *
 * Copyright 2001 MontaVista Software Inc.
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 */

#include <common.h>

#ifdef CONFIG_PCI

#include <asm/processor.h>
#include <asm/io.h>
#include <pci.h>
#include <mpc824x.h>

void pci_mpc824x_init (struct pci_controller *hose)
{
	// According to PCI Local Bus Specification
	//   Bus Number      : up to 256 buses on a PCI Bridge
	//   Device Number   : up to 32 devices on a given bus
	//   Function Number : up to 8 functions on a multifunction device
	//
	hose->first_busno = 0;
	hose->last_busno = 0xff;                 

        // The maximum PCI regions is determined by MAX_PCI_REGIONS (is 7).
        // The PCI Memory Space is divided into 2 regions

	/* System memory space */
	pci_set_region(hose->regions + 0,        // Region #0
		       CHRP_PCI_MEMORY_BUS,      // 0x0
		       CHRP_PCI_MEMORY_PHYS,     // 0x0
		       CHRP_PCI_MEMORY_SIZE,     // 0x4000_0000  (1GB)
		       PCI_REGION_MEM | PCI_REGION_MEMORY); // PCI Memory Space & System Memory

	/* PCI memory space */
	pci_set_region(hose->regions + 1,        // Region #1
		       CHRP_PCI_MEM_BUS,         // 0x8000_0000
		       CHRP_PCI_MEM_PHYS,        // 0x8000_0000
		       CHRP_PCI_MEM_SIZE,        // 0x7D00_0000  (2GB - 48MB)
		       PCI_REGION_MEM);          // PCI Memory Space

	/* ISA/PCI memory space */
	// Please refer to User Manual, Chapter 3.2.3  Processor Alias Space.
	// The processor alias space optional mapping can translate processor 
	// accesses in the 16MB range, starting at 0xFD00_0000 to the first 16MB of 
	// PCI Memory Space.  The processor alias space can access devices that 
	// cannot be located above 16MB in PCI Memory Space (like ISA-compatible devices).
	// Setting AMBOR[CPU_FD_ALIAS_EN] enable it.
	//
	pci_set_region(hose->regions + 2,        // Region #2
		       CHRP_ISA_MEM_BUS,         // 0x0
		       CHRP_ISA_MEM_PHYS,        // 0xFD00_0000
		       CHRP_ISA_MEM_SIZE,        // 0x0100_0000  (16MB)
		       PCI_REGION_MEM);          // PCI Memory Space

	/* PCI I/O space */
	pci_set_region(hose->regions + 3,        // Region #3
		       CHRP_PCI_IO_BUS,          // 0x0080_0000 (bit[24:31] is masked out
		       CHRP_PCI_IO_PHYS,         // 0xFE80_0000
		       CHRP_PCI_IO_SIZE,         // 0x0040_0000  (4MB)
		       PCI_REGION_IO);           // PCI I/O Space

	/* ISA/PCI I/O space */
	pci_set_region(hose->regions + 4,        // Region #4
		       CHRP_ISA_IO_BUS,          // 0x0000_0000 (bit[24:31] is masked out
		       CHRP_ISA_IO_PHYS,         // 0xFE00_0000
		       CHRP_ISA_IO_SIZE,         // 0x0080_0000  (8MB, but only 64KB usable)
		       PCI_REGION_IO);           // PCI I/O Space

	hose->region_count = 5;                  // Total NO# of regions are 5 (0..4)

        // Please refer to User Manual (Chapter 4.1  Configuration Register Access).
        // - A 32-bit register address 0x8000_00nn, where nn is the address offset of the 
        //   desired configuration register (see Table 4-1 & Figure 4-1), is written
        //   to the CONFIG_ADDR port at any word-aligned address winthin the range 
        //   0xFEC0_0000 -- 0xFEDF_FFFF.  Then the data is accessed at the CONFIG_DATA
        //   port at 0xFEE0_0000 -- 0xFEEF_FFFF.   Note that every word within this range
        //   has an alias to the same location.
        // - For example :
        //     Offset (0x8000_00E2) is  PLL Configuration Register of PCI Bridge.  
        // 
	pci_setup_indirect(hose,
			   CHRP_REG_ADDR,        // 0xFEC0_0000
			   CHRP_REG_DATA);       // 0xFEE0_0000

	pci_register_hose(hose);

	hose->last_busno = pci_hose_scan(hose);
}

#endif
