/*
 * rtl8169.c : U-Boot driver for the RealTek RTL8169
 *
 * Masami Komiya (mkomiya@sonare.it)
 *
 * Most part is taken from r8169.c of etherboot
 *
 */

/**************************************************************************
*    r8169.c: Etherboot device driver for the RealTek RTL-8169 Gigabit
*    Written 2003 by Timothy Legge <tlegge@rogers.com>
*
*    This program is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with this program; if not, write to the Free Software
*    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*
*    Portions of this code based on:
*	r8169.c: A RealTek RTL-8169 Gigabit Ethernet driver
*		for Linux kernel 2.4.x.
*
*    Written 2002 ShuChen <shuchen@realtek.com.tw>
*	  See Linux Driver for full information
*
*    Linux Driver Version 1.27a, 10.02.2002
*
*    Thanks to:
*	Jean Chen of RealTek Semiconductor Corp. for
*	providing the evaluation NIC used to develop
*	this driver.  RealTek's support for Etherboot
*	is appreciated.
*
*    REVISION HISTORY:
*    ================
*
*    v1.0	11-26-2003	timlegge	Initial port of Linux driver
*    v1.5	01-17-2004	timlegge	Initial driver output cleanup
*
*    Indent Options: indent -kr -i8
*
*
*    History :
*       v1.5-A1  03/08/2005  Hank Chu
*       1) Copied from u-boot-1.1.2/drivers/rtl8169.c file.
*          And ported for RTL8169S-32 (32-bit Gigabit Ethernet PCI controller) 
*          on ppcboot boot-loader with Motorola MPC8241 CPU platform.
*       2) Please refer to the M_NH230 macro for details.
*
*
***************************************************************************/

#include <common.h>
#include <malloc.h>
#include <net.h>
#include <asm/io.h>
#include <pci.h>

#if (CONFIG_COMMANDS & CFG_CMD_NET) && defined(CONFIG_NET_MULTI) && \
	defined(CONFIG_NET_RTL8169S)

#undef DEBUG_RTL8169
#undef DEBUG_RTL8169_TX
#undef DEBUG_RTL8169_RX

//sp/
#ifdef M_NH230
//# undef M_NH230
#endif

#ifdef M_NH230
  #define drv_version "v1.5-A4"
  #define drv_date "03-15-2005"
#else
  #define drv_version "v1.5"
  #define drv_date "01-17-2004"
#endif

static u32 ioaddr;

//#define M_HANK_DEBUG                 // Define it to enable debugging codes
#ifdef M_HANK_DEBUG
  #define M_DBG(format, arg...)		printf(format , ## arg)
  #define M_FDBG(format, arg...)	printf("%s: " format, __FUNCTION__, ## arg)
  
  void dbg_dump_reg (struct eth_device *dev);
  #define M_DBG_DUMP_REG(dev)  dbg_dump_reg(dev)
  
#else
  #define M_DBG(format, arg...)
  #define M_FDBG(format, arg...)
  #define M_DBG_DUMP_REG(dev)
#endif

#ifdef M_NH230
  #define M_HANK_PATCH          // define it to enable the patches
 
  // Add the virt_to_bus() to convert virtual address to bus address.
  #define virt_to_bus(a)     pci_phys_to_mem((pci_dev_t)dev->priv, a)

  // Copied from the Linxu driver for RTL8169S controller.
  // These MCFG_METHOD_x is determined by bits of Transmit Confoguration Register (TCR)
  //   bit[30:26] Hardware Version ID0
  //   bit[23]    Hardware Version ID1
  //
  #define MCFG_METHOD_1		0x01
  #define MCFG_METHOD_2		0x02
  #define MCFG_METHOD_3		0x03
  #define MCFG_METHOD_4		0x04
  
  // According to the format of RTL8169S EEPROM :
  //   1) if byte[1:0] of EEPROM (ID code) not 0x8129,
  //   2) or, if EEPROM is removed, 
  // its PCI Vendor & Device IDs will be 0x10EC & 0x8129. So, we use the constant to
  // know whether the content of EEPROM is correct or not.
  #define PCI_DEVICE_ID_IF_INCORRECT_EEPROM       0x8129
  
  void rtl8169_hw_PHY_config (struct eth_device *dev);
  void rtl8169_EEPROM_handler(struct eth_device *dev);
#endif

/* Condensed operations for readability. */
#define virt_to_le32desc(addr)	cpu_to_le32(virt_to_bus(addr))
#define le32desc_to_virt(addr)	bus_to_virt(le32_to_cpu(addr))

#define currticks()	get_timer(0)
#define bus_to_phys(a)	pci_mem_to_phys((pci_dev_t)dev->priv, a)
#define phys_to_bus(a)	pci_phys_to_mem((pci_dev_t)dev->priv, a)

/* media options */
#define MAX_UNITS 8
static int media[MAX_UNITS] = { -1, -1, -1, -1, -1, -1, -1, -1 };

/* MAC address length*/
#define MAC_ADDR_LEN	6

/* max supported gigabit ethernet frame size -- must be at least (dev->mtu+14+4).*/
#define MAX_ETH_FRAME_SIZE	1536

#define TX_FIFO_THRESH 256	/* In bytes */

#define RX_FIFO_THRESH	7	/* 7 means NO threshold, Rx buffer level before first PCI xfer.	 */
#define RX_DMA_BURST	6	/* Maximum PCI burst, '6' is 1024 */
#define TX_DMA_BURST	6	/* Maximum PCI burst, '6' is 1024 */
#define EarlyTxThld	0x3F	/* 0x3F means NO early transmit */
#define RxPacketMaxSize 0x0800	/* Maximum size supported is 16K-1 */
#define InterFrameGap	0x03	/* 3 means InterFrameGap = the shortest one */

#define NUM_TX_DESC	1	/* Number of Tx descriptor registers */
#define NUM_RX_DESC	4	/* Number of Rx descriptor registers */
#define RX_BUF_SIZE	1536	/* Rx Buffer size : must be mutiple of 16 as alignment */
#define RX_BUF_LEN	8192

#define RTL_MIN_IO_SIZE 0x80
#define TX_TIMEOUT  (6*HZ)

/* write/read MMIO register */
#define RTL_W8(reg, val8)	writeb ((val8), ioaddr + (reg))
#define RTL_W16(reg, val16)	writew ((val16), ioaddr + (reg))
#define RTL_W32(reg, val32)	writel ((val32), ioaddr + (reg))
#define RTL_R8(reg)		readb (ioaddr + (reg))
#define RTL_R16(reg)		readw (ioaddr + (reg))
#define RTL_R32(reg)		((unsigned long) readl (ioaddr + (reg)))

#define ETH_FRAME_LEN	MAX_ETH_FRAME_SIZE
#define ETH_ALEN	MAC_ADDR_LEN
#define ETH_ZLEN	60

enum RTL8169_registers {
	MAC0 = 0,		/* Ethernet hardware address. */
	MAR0 = 8,		/* Multicast filter. */
	TxDescStartAddr = 0x20,
	TxHDescStartAddr = 0x28,
	FLASH = 0x30,
	ERSR = 0x36,
	ChipCmd = 0x37,
	TxPoll = 0x38,
	IntrMask = 0x3C,
	IntrStatus = 0x3E,
	TxConfig = 0x40,
	RxConfig = 0x44,
	RxMissed = 0x4C,
	Cfg9346 = 0x50,
	Config0 = 0x51,
	Config1 = 0x52,
	Config2 = 0x53,
	Config3 = 0x54,
	Config4 = 0x55,
	Config5 = 0x56,
	MultiIntr = 0x5C,
	PHYAR = 0x60,
	TBICSR = 0x64,
	TBI_ANAR = 0x68,
	TBI_LPAR = 0x6A,
	PHYstatus = 0x6C,
	RxMaxSize = 0xDA,
	CPlusCmd = 0xE0,
	RxDescStartAddr = 0xE4,
	EarlyTxThres = 0xEC,
	FuncEvent = 0xF0,
	FuncEventMask = 0xF4,
	FuncPresetState = 0xF8,
	FuncForceEvent = 0xFC,
};

enum RTL8169_register_content {
	/*InterruptStatusBits */
	SYSErr = 0x8000,
	PCSTimeout = 0x4000,
	SWInt = 0x0100,
	TxDescUnavail = 0x80,
	RxFIFOOver = 0x40,
	RxUnderrun = 0x20,
	RxOverflow = 0x10,
	TxErr = 0x08,
	TxOK = 0x04,
	RxErr = 0x02,
	RxOK = 0x01,

	/*RxStatusDesc */
	RxRES = 0x00200000,
	RxCRC = 0x00080000,
	RxRUNT = 0x00100000,
	RxRWT = 0x00400000,

	/*ChipCmdBits */
	CmdReset = 0x10,
	CmdRxEnb = 0x08,
	CmdTxEnb = 0x04,
	RxBufEmpty = 0x01,

	/*Cfg9346Bits */
	Cfg9346_Lock = 0x00,
	Cfg9346_Unlock = 0xC0,

	/*rx_mode_bits */
	AcceptErr = 0x20,
	AcceptRunt = 0x10,
	AcceptBroadcast = 0x08,
	AcceptMulticast = 0x04,
	AcceptMyPhys = 0x02,
	AcceptAllPhys = 0x01,

	/*RxConfigBits */
	RxCfgFIFOShift = 13,
	RxCfgDMAShift = 8,

	/*TxConfigBits */
	TxInterFrameGapShift = 24,
	TxDMAShift = 8,		/* DMA burst value (0-7) is shift this many bits */

	/*rtl8169_PHYstatus */
	TBI_Enable = 0x80,
	TxFlowCtrl = 0x40,
	RxFlowCtrl = 0x20,
	_1000bpsF = 0x10,
	_100bps = 0x08,
	_10bps = 0x04,
	LinkStatus = 0x02,
	FullDup = 0x01,

	/*GIGABIT_PHY_registers */
	PHY_CTRL_REG = 0,
	PHY_STAT_REG = 1,
	PHY_AUTO_NEGO_REG = 4,
	PHY_1000_CTRL_REG = 9,

	/*GIGABIT_PHY_REG_BIT */
	PHY_Restart_Auto_Nego = 0x0200,
	PHY_Enable_Auto_Nego = 0x1000,

	/* PHY_STAT_REG = 1; */
	PHY_Auto_Neco_Comp = 0x0020,

	/* PHY_AUTO_NEGO_REG = 4; */
	PHY_Cap_10_Half = 0x0020,
	PHY_Cap_10_Full = 0x0040,
	PHY_Cap_100_Half = 0x0080,
	PHY_Cap_100_Full = 0x0100,

	/* PHY_1000_CTRL_REG = 9; */
	PHY_Cap_1000_Full = 0x0200,

	PHY_Cap_Null = 0x0,

	/*_MediaType*/
	_10_Half = 0x01,
	_10_Full = 0x02,
	_100_Half = 0x04,
	_100_Full = 0x08,
	_1000_Full = 0x10,

	/*_TBICSRBit*/
	TBILinkOK = 0x02000000,
};

static struct {
	const char *name;
	u8 version;		/* depend on RTL8169 docs */
	u32 RxConfigMask;	/* should clear the bits supported by this chip */
} rtl_chip_info[] = {
	{"RTL-8169", 0x00, 0xff7e1880,},
#ifdef M_NH230
	{"RTL-8169S",  MCFG_METHOD_2,  0xff7e1880 },  // copied from RTL8169S Linux driver
#endif
	{"RTL-8169", 0x04, 0xff7e1880,},
};

enum _DescStatusBit {
	OWNbit = 0x80000000,
	EORbit = 0x40000000,
	FSbit = 0x20000000,
	LSbit = 0x10000000,
};

struct TxDesc {
	u32 status;
	u32 vlan_tag;
	u32 buf_addr;
	u32 buf_Haddr;
};

struct RxDesc {
	u32 status;
	u32 vlan_tag;
	u32 buf_addr;
	u32 buf_Haddr;
};


/* Define the TX Descriptor (each descriptor must be 256-byte alignment) */
static u8 tx_ring[NUM_TX_DESC * sizeof(struct TxDesc) + 256];
/*	__attribute__ ((aligned(256))); */

/* Create a static buffer of size RX_BUF_SZ for each
TX Descriptor.	All descriptors point to a
part of this buffer */
static unsigned char txb[NUM_TX_DESC * RX_BUF_SIZE];

/* Define the RX Descriptor (each descriptor must be 256-byte alignment) */
static u8 rx_ring[NUM_RX_DESC * sizeof(struct TxDesc) + 256];
  /*  __attribute__ ((aligned(256))); */

/* Create a static buffer of size RX_BUF_SZ for each
RX Descriptor	All descriptors point to a
part of this buffer */
#ifdef M_NH230
  // According to the datasheet, Rx buffers in Rx descriptors should be at least 8 bytes 
  // and 8-byte alignment in memory.  So, we aligne it on 16-byte boundary.
  static unsigned char rxb[NUM_RX_DESC * RX_BUF_SIZE]  __attribute__ ((aligned(16)));
#else
  static unsigned char rxb[NUM_RX_DESC * RX_BUF_SIZE];
#endif

struct rtl8169_private {
	void *mmio_addr;	/* memory map physical address */
	int chipset;
	unsigned long cur_rx;	/* Index into the Rx descriptor buffer of next Rx pkt. */
	unsigned long cur_tx;	/* Index into the Tx descriptor buffer of next Rx pkt. */
	unsigned long dirty_tx;
	unsigned char *TxDescArrays;	/* Index of Tx Descriptor buffer */
	unsigned char *RxDescArrays;	/* Index of Rx Descriptor buffer */
	struct TxDesc *TxDescArray;	/* Index of 256-alignment Tx Descriptor buffer */
	struct RxDesc *RxDescArray;	/* Index of 256-alignment Rx Descriptor buffer */
	unsigned char *RxBufferRings;	/* Index of Rx Buffer  */
	unsigned char *RxBufferRing[NUM_RX_DESC];	/* Index of Rx Buffer array */
	unsigned char *Tx_skbuff[NUM_TX_DESC];
} tpx;

static struct rtl8169_private *tpc;

static const u16 rtl8169_intr_mask =
    SYSErr | PCSTimeout | RxUnderrun | RxOverflow | RxFIFOOver | TxErr |
    TxOK | RxErr | RxOK;
static const unsigned int rtl8169_rx_config =
    (RX_FIFO_THRESH << RxCfgFIFOShift) | (RX_DMA_BURST << RxCfgDMAShift);

static struct pci_device_id supported[] = {
	{PCI_VENDOR_ID_REALTEK, 0x8169},
#ifdef M_NH230
	// For RTL8169S, if EEPROM removed, the PCI Device ID will become 0x8129.
	{PCI_VENDOR_ID_REALTEK, PCI_DEVICE_ID_IF_INCORRECT_EEPROM},
#endif
	{}
};

void mdio_write(int RegAddr, int value)
{
	int i;

	RTL_W32(PHYAR, 0x80000000 | (RegAddr & 0xFF) << 16 | value);
	udelay(1000);

	for (i = 2000; i > 0; i--) {
		/* Check if the RTL8169 has completed writing to the specified MII register */
		if (!(RTL_R32(PHYAR) & 0x80000000)) {
			break;
		} else {
			udelay(100);
		}
	}
}

int mdio_read(int RegAddr)
{
	int i, value = -1;

	RTL_W32(PHYAR, 0x0 | (RegAddr & 0xFF) << 16);
	udelay(1000);

	for (i = 2000; i > 0; i--) {
		/* Check if the RTL8169 has completed retrieving data from the specified MII register */
		if (RTL_R32(PHYAR) & 0x80000000) {
			value = (int) (RTL_R32(PHYAR) & 0xFFFF);
			break;
		} else {
			udelay(100);
		}
	}
	return value;
}

#define ARRAY_SIZE(x) (sizeof(x) / sizeof((x)[0]))

static int rtl8169_init_board(struct eth_device *dev)
{
	int i;
	u32 tmp;

#ifdef DEBUG_RTL8169
	printf ("%s\n", __FUNCTION__);
#endif
	ioaddr = dev->iobase;

	/* Soft reset the chip. */
	RTL_W8(ChipCmd, CmdReset);

	/* Check that the chip has finished the reset. */
	for (i = 1000; i > 0; i--)
		if ((RTL_R8(ChipCmd) & CmdReset) == 0)
			break;
		else
			udelay(10);

	/* identify chip attached to board */
	tmp = RTL_R32(TxConfig);
	tmp = ((tmp & 0x7c000000) + ((tmp & 0x00800000) << 2)) >> 24;

	for (i = ARRAY_SIZE(rtl_chip_info) - 1; i >= 0; i--){
		if (tmp == rtl_chip_info[i].version) {
			tpc->chipset = i;
			goto match;
		}
	}

	/* if unknown chip, assume array element #0, original RTL-8169 in this case */
	printf("PCI device %s: unknown chip version, assuming RTL-8169\n", dev->name);
	printf("PCI device: TxConfig = 0x%lX\n", (unsigned long) RTL_R32(TxConfig));
	tpc->chipset = 0;

match:
	return 0;
}

/**************************************************************************
RECV - Receive a frame
***************************************************************************/
static int rtl_recv(struct eth_device *dev)
{
	/* return true if there's an ethernet packet ready to read */
	/* nic->packet should contain data on return */
	/* nic->packetlen should contain length of data */
	int cur_rx;
	int length = 0;

#ifdef DEBUG_RTL8169_RX
	printf ("%s\n", __FUNCTION__);
#endif
	ioaddr = dev->iobase;

	cur_rx = tpc->cur_rx;
	if (( le32_to_cpu(tpc->RxDescArray[cur_rx].status) & OWNbit) == 0) {
		if (!( le32_to_cpu(tpc->RxDescArray[cur_rx].status) & RxRES)) {
			unsigned char rxdata[RX_BUF_LEN];
			length = (int) (le32_to_cpu(tpc->RxDescArray[cur_rx].status) & 0x00001FFF) - 4;

			memcpy(rxdata, tpc->RxBufferRing[cur_rx], length);
			NetReceive(rxdata, length);

			if (cur_rx == NUM_RX_DESC - 1)
				tpc->RxDescArray[cur_rx].status = cpu_to_le32( (OWNbit | EORbit) + RX_BUF_SIZE );
			else
				tpc->RxDescArray[cur_rx].status = cpu_to_le32( OWNbit + RX_BUF_SIZE );
			tpc->RxDescArray[cur_rx].buf_addr = cpu_to_le32(virt_to_bus((u32)tpc->RxBufferRing[cur_rx]));
			tpc->RxDescArray[cur_rx].buf_Haddr = 0;
		} else {
			puts("Error Rx");
		}
		cur_rx = (cur_rx + 1) % NUM_RX_DESC;
		tpc->cur_rx = cur_rx;
		return 1;

	}
	tpc->cur_rx = cur_rx;
	return (0);		/* initially as this is called to flush the input */
}

#define HZ 1000
/**************************************************************************
SEND - Transmit a frame
***************************************************************************/
static int rtl_send(struct eth_device *dev, volatile void *packet, int length)
{
	/* send the packet to destination */

	u32 to;
	u8 *ptxb;
	int entry = tpc->cur_tx % NUM_TX_DESC;
	u32 len = length;

#ifdef DEBUG_RTL8169_TX
	int stime = currticks();
	printf ("%s\n", __FUNCTION__);
	printf("sending %d bytes\n", len);
#endif

	ioaddr = dev->iobase;

	/* point to the current txb incase multiple tx_rings are used */
	ptxb = tpc->Tx_skbuff[entry * MAX_ETH_FRAME_SIZE];
	memcpy(ptxb, (char *)packet, (int)length);

	while (len < ETH_ZLEN)
		ptxb[len++] = '\0';

	tpc->TxDescArray[entry].buf_addr = cpu_to_le32(virt_to_bus((u32)ptxb));
	tpc->TxDescArray[entry].buf_Haddr = 0;
	
	if (entry != (NUM_TX_DESC - 1)) {
		tpc->TxDescArray[entry].status = cpu_to_le32( (OWNbit | FSbit | LSbit) | ((len > ETH_ZLEN) ?
						len : ETH_ZLEN) );
	} else {
		tpc->TxDescArray[entry].status = cpu_to_le32(
		    (OWNbit | EORbit | FSbit | LSbit) |
		    ((len > ETH_ZLEN) ? length : ETH_ZLEN) );
	}

	RTL_W8(TxPoll, 0x40);	/* set polling bit */

	tpc->cur_tx++;
	to = currticks() + TX_TIMEOUT;
	
	while (( le32_to_cpu(tpc->TxDescArray[entry].status) & OWNbit) && (currticks() < to));	/* wait */

	if (currticks() >= to) {
#ifdef DEBUG_RTL8169_TX
		puts ("tx timeout/error\n");				
		printf ("%s elapsed time : %ld\n", __FUNCTION__, currticks()-stime);
#endif
		return 0;
	} else {
#ifdef DEBUG_RTL8169_TX
		puts("tx done\n");
#endif
		return length;
	}
}

static void rtl8169_set_rx_mode(struct eth_device *dev)
{
	u32 mc_filter[2];	/* Multicast hash filter */
	int rx_mode;
	u32 tmp = 0;

#ifdef DEBUG_RTL8169
	printf ("%s\n", __FUNCTION__);
#endif

	/* IFF_ALLMULTI */
	/* Too many to filter perfectly -- accept all multicasts. */
	rx_mode = AcceptBroadcast | AcceptMulticast | AcceptMyPhys;
	mc_filter[1] = mc_filter[0] = 0xffffffff;

	tmp = rtl8169_rx_config | rx_mode | (RTL_R32(RxConfig) &
				   rtl_chip_info[tpc->chipset].RxConfigMask);

	RTL_W32(RxConfig, tmp);
	RTL_W32(MAR0 + 0, mc_filter[0]);
	RTL_W32(MAR0 + 4, mc_filter[1]);
}

static void rtl8169_hw_start(struct eth_device *dev)
{
	u32 i;

#ifdef DEBUG_RTL8169
	int stime = currticks();
	printf ("%s\n", __FUNCTION__);
#endif

#if 0
	/* Soft reset the chip. */
	RTL_W8(ChipCmd, CmdReset);

	/* Check that the chip has finished the reset. */
	for (i = 1000; i > 0; i--) {
		if ((RTL_R8(ChipCmd) & CmdReset) == 0)
			break;
		else
			udelay(10);
	}
#endif

	RTL_W8(Cfg9346, Cfg9346_Unlock);
	RTL_W8(ChipCmd, CmdTxEnb | CmdRxEnb);
	RTL_W8(EarlyTxThres, EarlyTxThld);

	/* For gigabit rtl8169 */
	RTL_W16(RxMaxSize, RxPacketMaxSize);

	/* Set Rx Config register */
	i = rtl8169_rx_config | (RTL_R32(RxConfig) &
				 rtl_chip_info[tpc->chipset].RxConfigMask);
	RTL_W32(RxConfig, i);

	/* Set DMA burst size and Interframe Gap Time */
	RTL_W32(TxConfig, (TX_DMA_BURST << TxDMAShift) |
				(InterFrameGap << TxInterFrameGapShift));

#ifdef M_HANK_PATCH
  #ifdef M_NH230
    #ifdef RTL8169_CHECKSUM_BY_HW
        // Rx Checksum Offload enable
        RTL_W16( CPlusCmd, RTL_R16(CPlusCmd) | RxChkSum);
    #else
  	RTL_W16( CPlusCmd, RTL_R16(CPlusCmd) );
    #endif
  #else
	RTL_W16( CPlusCmd, RTL_R16(CPlusCmd) );
  #endif

	if(	rtl_chip_info[tpc->chipset].version == MCFG_METHOD_2 ||
		rtl_chip_info[tpc->chipset].version == MCFG_METHOD_3)
	{
		RTL_W16( CPlusCmd, (RTL_R16(CPlusCmd)|(1<<14)|(1<<3)) );
		M_DBG("%s: Set MAC Reg C+CR Offset 0xE0: bit-3 and bit-14\n", __FUNCTION__);
	}
	else
	{
		RTL_W16( CPlusCmd, (RTL_R16(CPlusCmd)|(1<<3)) );
		M_DBG("%s: Set MAC Reg C+CR Offset 0xE0: bit-3.\n", __FUNCTION__);
	}

	{
		//RTL_W16(0xE2, 0x1517);
		//RTL_W16(0xE2, 0x152a);
		//RTL_W16(0xE2, 0x282a);
		RTL_W16(0xE2, 0x0000);
	}
#endif  // end of #ifdef {M_HANK_PATCH}

	tpc->cur_rx = 0;

#ifdef M_NH230
	RTL_W32(TxDescStartAddr, virt_to_bus((u32)tpc->TxDescArray));
	RTL_W32(TxDescStartAddr + 4, 0x00);
	RTL_W32(RxDescStartAddr, virt_to_bus((u32)tpc->RxDescArray));
	RTL_W32(RxDescStartAddr + 4, 0x00);

	// Change the LED mode
	#define CONFIG1_REG_LED_MODE_3  0x0C0   // bit[7:6] is LEDS1-0 
	u8 c_reg = RTL_R8(Config1);
        RTL_W8(Config1, c_reg & (~CONFIG1_REG_LED_MODE_3));
        M_DBG("RTL8169S : LEDS1-0 is configured to 0b00\n");	
	
#else
	RTL_W32(TxDescStartAddr, virt_to_le32desc((u32)tpc->TxDescArray));
	RTL_W32(RxDescStartAddr, virt_to_le32desc((u32)tpc->RxDescArray));
#endif
	RTL_W8(Cfg9346, Cfg9346_Lock);
	udelay(10);

	RTL_W32(RxMissed, 0);

	rtl8169_set_rx_mode(dev);

	/* no early-rx interrupts */
	RTL_W16(MultiIntr, RTL_R16(MultiIntr) & 0xF000);

#ifdef DEBUG_RTL8169
	printf ("%s elapsed time : %ld\n", __FUNCTION__, currticks()-stime);
#endif
}

static void rtl8169_init_ring(struct eth_device *dev)
{
	int i;

#ifdef DEBUG_RTL8169
	int stime = currticks();
	printf ("%s\n", __FUNCTION__);
#endif

	tpc->cur_rx = 0;
	tpc->cur_tx = 0;
	tpc->dirty_tx = 0;

	memset(tpc->TxDescArray, 0x0, NUM_TX_DESC * sizeof(struct TxDesc));
	memset(tpc->RxDescArray, 0x0, NUM_RX_DESC * sizeof(struct RxDesc));

	for (i = 0; i < NUM_TX_DESC; i++) {
		tpc->Tx_skbuff[i] = &txb[i];
	}

	for (i = 0; i < NUM_RX_DESC; i++) {
		if (i == (NUM_RX_DESC - 1))
			tpc->RxDescArray[i].status = cpu_to_le32(
			    (OWNbit | EORbit) + RX_BUF_SIZE );
		else
			tpc->RxDescArray[i].status = cpu_to_le32( OWNbit + RX_BUF_SIZE );

		// Rx buffer(s) in Rx descriptors should be at least 8 bytes and 8-byte alignment in memory.
		tpc->RxBufferRing[i] = &rxb[i * RX_BUF_SIZE];
		tpc->RxDescArray[i].buf_addr = cpu_to_le32(virt_to_bus((u32)tpc->RxBufferRing[i]));
		tpc->RxDescArray[i].buf_Haddr = 0;
	}  

#ifdef DEBUG_RTL8169
	printf ("%s elapsed time : %ld\n", __FUNCTION__, currticks()-stime);
#endif
}

/**************************************************************************
RESET - Finish setting up the ethernet interface
***************************************************************************/
#ifdef M_NH230
  static int  rtl_reset(struct eth_device *dev, bd_t *bis)
#else
  static void rtl_reset(struct eth_device *dev, bd_t *bis)
#endif
{
	int i;
	u8 diff;
	u32 TxPhyAddr, RxPhyAddr;

#ifdef DEBUG_RTL8169
	int stime = currticks();
	printf ("%s\n", __FUNCTION__);
#endif

	tpc->TxDescArrays = tx_ring;
	if (tpc->TxDescArrays == 0)
		puts("Allot Error");
	/* Tx Desscriptor needs 256 bytes alignment; */
	TxPhyAddr = virt_to_bus((u32)tpc->TxDescArrays);
	diff = 256 - (TxPhyAddr - ((TxPhyAddr >> 8) << 8));
	TxPhyAddr += diff;
	tpc->TxDescArray = (struct TxDesc *) (tpc->TxDescArrays + diff);

	tpc->RxDescArrays = rx_ring;
	/* Rx Desscriptor needs 256 bytes alignment; */
	RxPhyAddr = virt_to_bus((u32)tpc->RxDescArrays);
	diff = 256 - (RxPhyAddr - ((RxPhyAddr >> 8) << 8));
	RxPhyAddr += diff;
	tpc->RxDescArray = (struct RxDesc *) (tpc->RxDescArrays + diff);

	if (tpc->TxDescArrays == NULL || tpc->RxDescArrays == NULL) {
		puts("Allocate RxDescArray or TxDescArray failed\n");
#ifdef M_NH230
	return (0);        // return 0 as error
#else
		return;
#endif
	}

	rtl8169_init_ring(dev);
	rtl8169_hw_start(dev);
	/* Construct a perfect filter frame with the mac address as first match
	 * and broadcast for all others */
	for (i = 0; i < 192; i++)
		txb[i] = 0xFF;

	txb[0] = dev->enetaddr[0];
	txb[1] = dev->enetaddr[1];
	txb[2] = dev->enetaddr[2];
	txb[3] = dev->enetaddr[3];
	txb[4] = dev->enetaddr[4];
	txb[5] = dev->enetaddr[5];

#ifdef DEBUG_RTL8169
	printf ("%s elapsed time : %ld\n", __FUNCTION__, currticks()-stime);
#endif

#ifdef M_NH230
	return (1);             // return 1 as OK
#endif
}

/**************************************************************************
HALT - Turn off ethernet interface
***************************************************************************/
static void rtl_halt(struct eth_device *dev)
{
	int i;

#ifdef DEBUG_RTL8169
	printf ("%s\n", __FUNCTION__);
#endif

	ioaddr = dev->iobase;

	/* Stop the chip's Tx and Rx DMA processes. */
	RTL_W8(ChipCmd, 0x00);

	/* Disable interrupts by clearing the interrupt mask. */
	RTL_W16(IntrMask, 0x0000);

	RTL_W32(RxMissed, 0);

	tpc->TxDescArrays = NULL;
	tpc->RxDescArrays = NULL;
	tpc->TxDescArray = NULL;
	tpc->RxDescArray = NULL;
	for (i = 0; i < NUM_RX_DESC; i++) {
		tpc->RxBufferRing[i] = NULL;
	}
}

/**************************************************************************
INIT - Look for an adapter, this routine's visible to the outside
***************************************************************************/

#define board_found 1
#define valid_link 0
static int rtl_init(struct eth_device *dev, bd_t *bis)
{
	static int board_idx = -1;
	static int printed_version = 0;
	int i, rc;
	int option = -1, Cap10_100 = 0, Cap1000 = 0;

#ifdef DEBUG_RTL8169
	printf ("%s\n", __FUNCTION__);
#endif

	ioaddr = dev->iobase;

	board_idx++;

	printed_version = 1;

	/* point to private storage */
	tpc = &tpx;

	rc = rtl8169_init_board(dev);
	if (rc)
		return rc;

#ifdef M_NH230
    // Re-program EEPROM of ETL8169S if the content of 93C46 EEPROM is empty or incorrect.
    // And then re-load the content of EEPROM to registers.
    u16 w_value = 0;
    pci_read_config_word((pci_dev_t)dev->priv, PCI_DEVICE_ID, &w_value);
    M_FDBG("w_value=%x", w_value);
    if (PCI_DEVICE_ID_IF_INCORRECT_EEPROM == w_value) {  // incorrect PCI Device ID ?
      rtl8169_EEPROM_handler(dev);
      M_FDBG("if (PCI_DEVICE_ID_IF_INCORRECT_EEPROM == w_value");
    }
    
    {
      // Overwrite with the MAC addresses loaded from flash (environment variables)
      extern int if_string_2_eth(char *lpsz_str, char *lpsz_eth);
      char gsz_eth_address[MAC_ADDR_LEN+1];
          
      char *lpsz_MAC_str = getenv("ethaddr");
      if_string_2_eth(lpsz_MAC_str, gsz_eth_address);
      memcpy(dev->enetaddr, gsz_eth_address, MAC_ADDR_LEN);
          
      M_DBG("%s: dev->enetaddr = %02X:%02X:%02X:%02X:%02X:%02X\n", 
      		__FUNCTION__, dev->enetaddr[0], dev->enetaddr[1], dev->enetaddr[2],
               			  dev->enetaddr[3], dev->enetaddr[4], dev->enetaddr[5]);
    
      // Before over-write the IDR0 ~ 5 registers :
      //   93C46 Command Register[6:7]: Operating Mode = Config register write enable
      RTL_W8 (Cfg9346, Cfg9346_Unlock);  
   
      // Overwrite the given MAC address into ID registers
      unsigned long dw_value = gsz_eth_address[3];
      dw_value <<= 8;
      dw_value += gsz_eth_address[2];
      dw_value <<= 8;
      dw_value += gsz_eth_address[1];
      dw_value <<= 8;
      dw_value += gsz_eth_address[0];
      RTL_W32( MAC0 + 0, dw_value);             // Write 4-byte value to IDR0~3
        
      dw_value = gsz_eth_address[5];
      dw_value <<= 8;
      dw_value += gsz_eth_address[4];
        
      unsigned long dw_temp = RTL_R8(MAC0 + 6); // Preserve register (offset = MAC0 + 6)
      dw_value |= dw_temp << 16;
      dw_temp = RTL_R8(MAC0 + 7);               // Preserve register (offset = MAC0 + 7)
      dw_value |= dw_temp << 24;
      RTL_W32( MAC0 + 4, dw_value);             // Write 4-byte to IDR4~5 and Reserved (0x6, 0x7)
        
      // After over-write the IDR0 ~ 5 registers :
      //   93C46 Command Register[6:7]: Operating Mode = Normal communication mode
      RTL_W8 (Cfg9346, Cfg9346_Lock);

      M_DBG("%s: MAC address[%02X:%02X:%02X:%02X:%02X:%02X] in IDR0~5 registers\n", 
	        __FUNCTION__, RTL_R8(MAC0+0), RTL_R8(MAC0+1), RTL_R8(MAC0+2),
	                      RTL_R8(MAC0+3), RTL_R8(MAC0+4), RTL_R8(MAC0+5));          
    }
#else
	/* Get MAC address.  FIXME: read EEPROM */
	for (i = 0; i < MAC_ADDR_LEN; i++)
		dev->enetaddr[i] = RTL_R8(MAC0 + i);
		
#endif  // end of #ifdef {M_NH230}


#ifdef DEBUG_RTL8169
	printf("MAC Address");
	for (i = 0; i < MAC_ADDR_LEN; i++)
		printf(":%02x", dev->enetaddr[i]);
	putc('\n');
#endif

#ifdef DEBUG_RTL8169
	/* Print out some hardware info */
	printf("%s: at ioaddr 0x%x\n", dev->name, ioaddr);
#endif


#ifdef M_HANK_PATCH
	// Config PHY
	rtl8169_hw_PHY_config(dev);

	M_DBG("%s: Set MAC Reg C+CR Offset 0x82h = 0x01h\n", __FUNCTION__);
	RTL_W8( 0x82, 0x01 );

	if( rtl_chip_info[tpc->chipset].version < MCFG_METHOD_3 ){
		M_DBG("%s: Set PCI Latency=0x40\n", __FUNCTION__);
		pci_write_config_byte((pci_dev_t)dev->priv, PCI_LATENCY_TIMER, 0x40);
	}

	if( rtl_chip_info[tpc->chipset].version == MCFG_METHOD_2 ){
		M_DBG("%s: Set MAC Reg C+CR Offset 0x82h = 0x01h\n", __FUNCTION__);
		RTL_W8( 0x82, 0x01 );
		M_DBG("Set PHY Reg 0x0bh = 0x00h\n");
		mdio_write( 0x0b, 0x0000 );	//w 0x0b 15 0 0
	}
#endif

	/* if TBI is not endbled */
	if (!(RTL_R8(PHYstatus) & TBI_Enable)) {
		int val = mdio_read(PHY_AUTO_NEGO_REG);

		option = (board_idx >= MAX_UNITS) ? 0 : media[board_idx];
		/* Force RTL8169 in 10/100/1000 Full/Half mode. */
		if (option > 0) {
#ifdef DEBUG_RTL8169
			printf("%s: Force-mode Enabled.\n", dev->name);
#endif
			Cap10_100 = 0, Cap1000 = 0;
			switch (option) {
			case _10_Half:
				Cap10_100 = PHY_Cap_10_Half;
				Cap1000 = PHY_Cap_Null;
				break;
			case _10_Full:
				Cap10_100 = PHY_Cap_10_Full;
				Cap1000 = PHY_Cap_Null;
				break;
			case _100_Half:
				Cap10_100 = PHY_Cap_100_Half;
				Cap1000 = PHY_Cap_Null;
				break;
			case _100_Full:
				Cap10_100 = PHY_Cap_100_Full;
				Cap1000 = PHY_Cap_Null;
				break;
			case _1000_Full:
				Cap10_100 = PHY_Cap_Null;
				Cap1000 = PHY_Cap_1000_Full;
				break;
			default:
				break;
			}
			mdio_write(PHY_AUTO_NEGO_REG, Cap10_100 | (val & 0x1F));	/* leave PHY_AUTO_NEGO_REG bit4:0 unchanged */
			mdio_write(PHY_1000_CTRL_REG, Cap1000);
		} else {
#ifdef DEBUG_RTL8169
			printf("%s: Auto-negotiation Enabled.\n",
			       dev->name);
#endif
			/* enable 10/100 Full/Half Mode, leave PHY_AUTO_NEGO_REG bit4:0 unchanged */
			mdio_write(PHY_AUTO_NEGO_REG,
				   PHY_Cap_10_Half | PHY_Cap_10_Full |
				   PHY_Cap_100_Half | PHY_Cap_100_Full |
				   (val & 0x1F));

			/* enable 1000 Full Mode */
			mdio_write(PHY_1000_CTRL_REG, PHY_Cap_1000_Full);

		}

		/* Enable auto-negotiation and restart auto-nigotiation */
		mdio_write(PHY_CTRL_REG,
			   PHY_Enable_Auto_Nego | PHY_Restart_Auto_Nego);
		udelay(100);

		/* wait for auto-negotiation process */
		for (i = 10000; i > 0; i--) {
			/* check if auto-negotiation complete */
			if (mdio_read(PHY_STAT_REG) & PHY_Auto_Neco_Comp) {
				udelay(100);
				option = RTL_R8(PHYstatus);
				if (option & _1000bpsF) {
#ifdef DEBUG_RTL8169
					printf("%s: 1000Mbps Full-duplex operation.\n",
					     dev->name);
#endif
				} else {
#ifdef DEBUG_RTL8169
					printf
					    ("%s: %sMbps %s-duplex operation.\n",
					     dev->name,
					     (option & _100bps) ? "100" :
					     "10",
					     (option & FullDup) ? "Full" :
					     "Half");
#endif
				}
				break;
			} else {
				udelay(100);
			}
		}		/* end for-loop to wait for auto-negotiation process */

	} else {
		udelay(100);
#ifdef DEBUG_RTL8169
		printf
		    ("%s: 1000Mbps Full-duplex operation, TBI Link %s!\n",
		     dev->name,
		     (RTL_R32(TBICSR) & TBILinkOK) ? "OK" : "Failed");
#endif
	}

	return 1;
}

int rtl8169_initialize(bd_t *bis)
{
	pci_dev_t devno;
	int card_number = 0;
	struct eth_device *dev;
	u32 iobase;
	int idx=0;

#ifdef M_NH230
	printf("\nRTL8169S driver %s  %s\n", drv_version, drv_date);
#endif

	while(1){
		/* Find RTL8169 */
		if ((devno = pci_find_devices(supported, idx++)) < 0) {
			if (card_number == 0)
			  M_DBG("%s: error: pci_find_devices() < 0\n", __FUNCTION__);
			break;
		}

		pci_read_config_dword(devno, PCI_BASE_ADDRESS_1, &iobase);
		M_DBG ("%s: PCI Base Address=0x%X, iobase=0x%X\n", __FUNCTION__, iobase, iobase & (~0xf) );
		iobase &= ~0xf;

		debug ("rtl8169: REALTEK RTL8169 @0x%x\n", iobase);
		
		dev = (struct eth_device *)malloc(sizeof *dev);

		sprintf (dev->name, "RTL8169#%d", card_number);

		dev->priv = (void *) devno;
		dev->iobase = (int)bus_to_phys(iobase);
		
		M_DBG ("%s: rtl8169: dev->iobase = 0x%X\n", __FUNCTION__, dev->iobase);

		dev->init = rtl_reset;
		dev->halt = rtl_halt;
		dev->send = rtl_send;
		dev->recv = rtl_recv;

		eth_register (dev);

		rtl_init(dev, bis);

		card_number++;
	}
	
	return card_number;
}

  #ifdef M_NH230
    //
    // Subject :
    //   The EEPROM of RTL8169S cannot be removed as Config3 register will become unstable
    //   and Interrupt# will not be asserted.
    //
    //   But if content of EEPROM of RTL8169S is empty or incorrect, reprogram EEPROM with
    //   correct content and reload into registers.
    //   About the details, please refer to the format of EEPROM for RTL8169S. 
    //
    void rtl8169_EEPROM_handler(struct eth_device *dev)
    {
        // The correct template for content of EEPROM
    	u8 sz_EEPROM[] = {0x29, 0x81, 0xEC, 0x10, 0x69, 0x81, 0xEC, 0x10,
	  		  0x69, 0x81, 0x20, 0x40, 0x01, 0xA1, 0x00, 0x08,	  		  
	  		  0x54, 0xD0, 0x0B, 0x6E, 0x15, 0x0D, 0xC2, 0xF7,
	  		  0x00, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x13,	  		  
	  		  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	  	  	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	  	  	  
	  		  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	  		  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x08,	  		  
	  		  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	  		  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	  		  
	  		  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	  		  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	  		  
	  		  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	  		  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	  		  
	  		  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	  		  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
	  		  
     #ifdef M_03_13_2005   // Hank Chu : not tested yet
	// 1) Configure RTL8169S under 93C46 programming mode
	RTL_W8(Cfg9346, 0x80);
	
	// 2) Loop to program EEPROM
	u16 w_loop = 0, w_EEPROM_size = sizeof(sz_EEPROM);
	for (w_loop = 0; w_loop < w_EEPROM_size; w_loop++) {
	  EEPROM_write_byte(sz_EEPROM[w_loop], w_loop);   // To Deb Defined ????
	}	
	
	// 3) Configure it under Auto-load mode
	RTL_W8(Cfg9346, 0x40);
	
	// 4) Loop to polling it until back to Normal communication mode
	for (w_loop = 0; w_loop < 2000; w_loop++) {  // Loop until 2 seconds
	  udelay(1000);                              // delay 1 ms
	  if (Cfg9346_Lock == (RTL_R8(Cfg9346) & 0x0C0))
	    break;
	}
	
	// 5) Configure it under Normal Communication mode
	RTL_W8(Cfg9346, Cfg9346_Lock);
	udelay(10);
     #endif
    }
  #endif

  #ifdef M_HANK_PATCH
    //
    // This routine is copied from Linux driver of RTL8169s controller.
    //
    void rtl8169_hw_PHY_config (struct eth_device *dev)
    {
      ioaddr = dev->iobase;

      // Identify the version of PHY
      #define PCFG_METHOD_1		0x01	//PHY Reg 0x03 bit0-3 == 0x0000
      #define PCFG_METHOD_2		0x02	//PHY Reg 0x03 bit0-3 == 0x0001
      #define PCFG_METHOD_3		0x03	//PHY Reg 0x03 bit0-3 == 0x0002
    
      int pcfg = 0;
      unsigned char val8 = (unsigned char)(mdio_read(3)&0x000f);
      if( val8 == 0x00 ){
	pcfg = PCFG_METHOD_1;
      }
      else if( val8 == 0x01 ){
 	pcfg = PCFG_METHOD_2;
      }
      else if( val8 == 0x02 ){
	pcfg = PCFG_METHOD_3;
      }
      else {
	pcfg = PCFG_METHOD_3;
      }

      M_DBG("RTL8169S: chipset version = %d, PHY version=%d\n",rtl_chip_info[tpc->chipset].version, pcfg);
//sp/
/*	DH: 051230
*	Parameters of RTL8169SB's phy; shouldn't apply this per Realtek FAE's comment; no doc to refer to..
*	Dnt: RTL8169S-32O.18s{,RTL8169SBO.13s{,supportq|t,ѼƳ]w]|P.
*
*/
/*
      if( rtl_chip_info[tpc->chipset].version == MCFG_METHOD_4 ){
	mdio_write(  0x1F, 0x0001 );
	mdio_write(  0x09, 0x273a );
	mdio_write(  0x0e, 0x7bfb );
	mdio_write(  0x1b, 0x841e );

	mdio_write(  0x1F, 0x0002 );
	mdio_write(  0x01, 0x90D0 );
	mdio_write(  0x1F, 0x0000 );
      }
      else{
	M_DBG("%s: Discard hw PHY config.\n", __FUNCTION__);
      }	
*/
}

    #ifdef M_HANK_DEBUG
        //
        // Subject :
        //   This routine is used for debugging only.
        //   It prints out the content of registers in byte.
        //
	void dbg_dump_reg (struct eth_device *dev)
	{
	  ioaddr = dev->iobase;
	
	  // Dump the contents of registers
	  printf("\n RTL8169S : Dumpe the registers\n");
	  int w_loop = 0;
	  for (w_loop = 0; w_loop <= 0x0ff; w_loop++) {
	    if (w_loop && (!(w_loop & 0x0f)) )
	      printf("\n");
            if (!(w_loop & 0x0f))
	      printf("%02X: ", w_loop);
	    
	    printf("%02X ", RTL_R8(MAC0 + w_loop));
	  }
	  printf("\n");
        }
    #endif  // end of #ifdef {M_HANK_DEBUG}

  #endif // end of #ifdef {M_NH230}

#endif
