/* 
   Unix SMB/CIFS implementation.
   uid/user handling
   Copyright (C) Tim Potter 2000
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "includes.h"
#include "sp_smbd_dbg.h"

extern struct current_user current_user;

struct sec_ctx {
	uid_t uid;
	uid_t gid;
	int ngroups;
	gid_t *groups;
	NT_USER_TOKEN *token;
};

/* A stack of security contexts.  We include the current context as being
   the first one, so there is room for another MAX_SEC_CTX_DEPTH more. */

static struct sec_ctx sec_ctx_stack[MAX_SEC_CTX_DEPTH + 1];
static int sec_ctx_stack_ndx;

/****************************************************************************
 Become the specified uid.
****************************************************************************/

static BOOL become_uid(uid_t uid)
{
	/* Check for dodgy uid values */
	sp_sec_ctx_dbg(0, ("%s: %s..:1\n", __FILE__, __func__));
	if (uid == (uid_t)-1 || 
	    ((sizeof(uid_t) == 2) && (uid == (uid_t)65535))) {
		static int done;
 		sp_sec_ctx_dbg(0, ("%s: %s..:2\n", __FILE__, __func__));
		if (!done) {
			DEBUG(1,("WARNING: using uid %d is a security risk\n",
				 (int)uid));
			done = 1;
		}
	}

	/* Set effective user id */
	sp_sec_ctx_dbg(0, ("%s: %s..:3\n", __FILE__, __func__));
	set_effective_uid(uid);
	sp_sec_ctx_dbg(0, ("%s: %s..:4\n", __FILE__, __func__));
	DO_PROFILE_INC(uid_changes);
	return True;
}

/****************************************************************************
 Become the specified gid.
****************************************************************************/

static BOOL become_gid(gid_t gid)
{
	/* Check for dodgy gid values */
	sp_sec_ctx_dbg(0, ("%s: %s..:1\n", __FILE__, __func__));
	if (gid == (gid_t)-1 || ((sizeof(gid_t) == 2) && 
				 (gid == (gid_t)65535))) {
		static int done;
		sp_sec_ctx_dbg(0, ("%s: %s..:2\n", __FILE__, __func__));
		if (!done) {
			DEBUG(1,("WARNING: using gid %d is a security risk\n",
				 (int)gid));  
			done = 1;
		}
	}
  
	/* Set effective group id */
	sp_sec_ctx_dbg(0, ("%s: %s..:3\n", __FILE__, __func__));
	set_effective_gid(gid);
	sp_sec_ctx_dbg(0, ("%s: %s..:4\n", __FILE__, __func__));
	return True;
}

/****************************************************************************
 Become the specified uid and gid.
****************************************************************************/

static BOOL become_id(uid_t uid, gid_t gid)
{
	BOOL i=0;
	sp_sec_ctx_dbg(0, ("%s: %s..:uid=%d gid=%d sizeof(uid)=%d sizeof(gid)=%d\n", __FILE__, __func__, 
				uid, gid, sizeof(uid), sizeof(gid))
				);
	i= become_gid(gid) && become_uid(uid);
	//return become_gid(gid) && become_uid(uid);
	return i;
}

/****************************************************************************
 Drop back to root privileges in order to change to another user.
****************************************************************************/

static void gain_root(void)
{
	//sp_sec_ctx_dbg(0, ("%s: %s..:\n", __FILE__, __func__));
	sp_sec_ctx_dbg(0, ("%s: %s..:1..pid=%d, euid=%d, egid=%d\n", __FILE__, __func__, getpid(), (int)geteuid(), (int)getegid()));
	if (non_root_mode()) {
		sp_sec_ctx_dbg(0, ("%s: %s..:if (non_root_mode()) {..pid=%d\n", __FILE__, __func__, getpid()));
		return;
	}
	sp_sec_ctx_dbg(0, ("%s: %s..:2..pid=%d\n", __FILE__, __func__, getpid()));
	if (geteuid() != 0) {
		sp_sec_ctx_dbg(0, ("%s: %s..:if (geteuid() != 0) {..pid=%d, euid=%d\n", __FILE__, __func__, getpid(), (int)geteuid() ));
		set_effective_uid(0);
		if (geteuid() != 0) {
			DEBUG(0,
			      ("Warning: You appear to have a trapdoor "
			       "uid system\n"));
		}
	}
	sp_sec_ctx_dbg(0, ("%s: %s..:3..pid=%d\n", __FILE__, __func__, getpid()));
	if (getegid() != 0) {
		sp_sec_ctx_dbg(0, ("%s: %s..:if (getegid() != 0) {..pid=%d, egid=%d\n", __FILE__, __func__, getpid(), (int)getegid() ));
		set_effective_gid(0);
		if (getegid() != 0) {
			DEBUG(0,
			      ("Warning: You appear to have a trapdoor "
			       "gid system\n"));
		}
	}
	sp_sec_ctx_dbg(0, ("%s: %s..:4..pid=%d\n", __FILE__, __func__, getpid()));
}

/****************************************************************************
 Get the list of current groups.
****************************************************************************/

int get_current_groups(gid_t gid, int *p_ngroups, gid_t **p_groups)
{
	int i;
	gid_t grp;
	int ngroups;
	gid_t *groups = NULL;
	sp_sec_ctx_dbg(0, ("%s: %s..:\n", __FILE__, __func__));
	
	sp_sec_ctx_dbg(0, ("%s: %s..:getgid()=%d getgid()=%x\n", __FILE__, __func__, getgid(), getgid()) );
    sp_sec_ctx_dbg(0, ("%s: %s..:getegid()=%d getegid()=%x\n", __FILE__, __func__, getegid(), getegid()) );
//sp
	if( !initgroups("root", &grp) ){
    	sp_server_dbg(0, ("%s: %s..:if( !initgroups(\"root\", &grp) ){", __FILE__, __func__));
	}
	    
	//sp_sec_ctx_dbg(0, ("%s: %s..:gid=%d *p_ngroups=%d **p_groups=%d\n", __FILE__, __func__, gid, *p_ngroups, **p_groups));
	(*p_ngroups) = 0;
	(*p_groups) = NULL;

	/* this looks a little strange, but is needed to cope with
	   systems that put the current egid in the group list
	   returned from getgroups() (tridge) */
	save_re_gid();
	set_effective_gid(gid);
	setgid(gid);

	ngroups = sys_getgroups(0,&grp);
	if (ngroups <= 0) {
		sp_sec_ctx_dbg(0, ("%s: %s..:if (ngroups=%d <= 0) {\n", __FILE__, __func__, ngroups));
		goto fail;
	}
	sp_sec_ctx_dbg(0, ("%s: %s..:1..ngroups=%d\n", __FILE__, __func__, ngroups));
	
	if((groups = (gid_t *)malloc(sizeof(gid_t)*(ngroups+1))) == NULL) {
		DEBUG(0,("setup_groups malloc fail !\n"));
		goto fail;
	}

	if ((ngroups = sys_getgroups(ngroups,groups)) == -1) {
		sp_sec_ctx_dbg(0, ("%s: %s..:if ((ngroups=%d = sys_getgroups(ngroups,groups)) == -1) {\n", __FILE__, __func__, ngroups));
		goto fail;
	}
	sp_sec_ctx_dbg(0, ("%s: %s..:2..ngroups=%d\n", __FILE__, __func__, ngroups));
	
	restore_re_gid();

	(*p_ngroups) = ngroups;
	(*p_groups) = groups;

	DEBUG( 3, ( "get_current_groups: user is in %u groups: ", ngroups));
	for (i = 0; i < ngroups; i++ ) {
		DEBUG( 3, ( "%s%d", (i ? ", " : ""), (int)groups[i] ) );
	}
	DEBUG( 3, ( "\n" ) );

	return ngroups;

fail:
	SAFE_FREE(groups);
	restore_re_gid();
	return -1;
}

/****************************************************************************
 Initialize the groups a user belongs to.
****************************************************************************/

BOOL initialise_groups(char *user, uid_t uid, gid_t gid)
{
	struct sec_ctx *prev_ctx_p;
	BOOL result = True;
	//sp_sec_ctx_dbg(0, ("%s: %s..:\n", __FILE__, __func__));
	sp_sec_ctx_dbg(0, ("%s: %s..:user=%s\n", __FILE__, __func__, user));
	if (non_root_mode()) {
		return True;
	}

	become_root();

	/* Call initgroups() to get user groups */

	if (winbind_initgroups(user,gid) == -1) {
		DEBUG(0,("Unable to initgroups. Error was %s\n", strerror(errno) ));
		if (getuid() == 0) {
			if (gid < 0 || gid > 32767 || uid < 0 || uid > 32767) {
				DEBUG(0,("This is probably a problem with the account %s\n", user));
			}
		}
		result = False;
		goto done;
	}

	/* Store groups in previous user's security context.  This will
	   always work as the become_root() call increments the stack
	   pointer. */

	prev_ctx_p = &sec_ctx_stack[sec_ctx_stack_ndx - 1];

	SAFE_FREE(prev_ctx_p->groups);
	prev_ctx_p->ngroups = 0;

	get_current_groups(gid, &prev_ctx_p->ngroups, &prev_ctx_p->groups);

 done:
	unbecome_root();

	return result;
}

/****************************************************************************
 Create a new security context on the stack.  It is the same as the old
 one.  User changes are done using the set_sec_ctx() function.
****************************************************************************/
BOOL push_sec_ctx(void){}
/*
BOOL push_sec_ctx(void)
{
	struct sec_ctx *ctx_p;
	
	// Check we don't overflow our stack 

	if (sec_ctx_stack_ndx == MAX_SEC_CTX_DEPTH) {
		DEBUG(0, ("Security context stack overflow!\n"));
		smb_panic("Security context stack overflow!\n");
	}

	// Store previous user context 

	sec_ctx_stack_ndx++;

	ctx_p = &sec_ctx_stack[sec_ctx_stack_ndx];

	ctx_p->uid = geteuid();
	ctx_p->gid = getegid();

 	DEBUG(3, ("push_sec_ctx(%u, %u) : sec_ctx_stack_ndx = %d\n", 
 		  (unsigned int)ctx_p->uid, (unsigned int)ctx_p->gid, sec_ctx_stack_ndx ));

	ctx_p->token = dup_nt_token(sec_ctx_stack[sec_ctx_stack_ndx-1].token);

	ctx_p->ngroups = sys_getgroups(0, NULL);

	if (ctx_p->ngroups != 0) {
		if (!(ctx_p->groups = malloc(ctx_p->ngroups * sizeof(gid_t)))) {
			DEBUG(0, ("Out of memory in push_sec_ctx()\n"));
			delete_nt_token(&ctx_p->token);
			return False;
		}

		sys_getgroups(ctx_p->ngroups, ctx_p->groups);
	} else {
		ctx_p->groups = NULL;
	}

	return True;
}
*/

/****************************************************************************
 Set the current security context to a given user.
****************************************************************************/
void set_sec_ctx(uid_t uid, gid_t gid, int ngroups, gid_t *groups, NT_USER_TOKEN *token){}
/*
void set_sec_ctx(uid_t uid, gid_t gid, int ngroups, gid_t *groups, NT_USER_TOKEN *token)
{
	struct sec_ctx *ctx_p = &sec_ctx_stack[sec_ctx_stack_ndx];
	sp_sec_ctx_dbg(0, ("%s: %s..:0..pid=%d uid=%d gid=%d ngroups=%d\n", __FILE__, __func__, getpid(), uid, gid, ngroups));
//sp	
	
	if( uid!=0 || gid!=0){
		sp_sec_ctx_dbg(0, ("%s: %s..:set uid&gid to 0\n", __FILE__, __func__));
		uid=0;
		gid=0;
	}
		
	// Set the security context 

	DEBUG(3, ("setting sec ctx (%u, %u) - sec_ctx_stack_ndx = %d\n", 
		(unsigned int)uid, (unsigned int)gid, sec_ctx_stack_ndx));

	debug_nt_user_token(DBGC_CLASS, 5, token);
	debug_unix_user_token(DBGC_CLASS, 5, uid, gid, ngroups, groups);
	sp_sec_ctx_dbg(0, ("%s: %s..:1..pid=%d\n", __FILE__, __func__, getpid()));
	gain_root();
	sp_sec_ctx_dbg(0, ("%s: %s..:2..pid=%d\n", __FILE__, __func__, getpid()));
#ifdef HAVE_SETGROUPS
	sys_setgroups(ngroups, groups);
#endif
	sp_sec_ctx_dbg(0, ("%s: %s..:2-1..pid=%d\n", __FILE__, __func__, getpid()));
	ctx_p->ngroups = ngroups;
	sp_sec_ctx_dbg(0, ("%s: %s..:before SAFE_FREE(ctx_p->groups);..pid=%d\n", __FILE__, __func__, getpid()));
	SAFE_FREE(ctx_p->groups);
	sp_sec_ctx_dbg(0, ("%s: %s..:after SAFE_FREE(ctx_p->groups);..pid=%d\n", __FILE__, __func__, getpid()));
	if (token && (token == ctx_p->token))
		smb_panic("DUPLICATE_TOKEN");
	sp_sec_ctx_dbg(0, ("%s: %s..:3..pid=%d\n", __FILE__, __func__, getpid()));
	delete_nt_token(&ctx_p->token);
	sp_sec_ctx_dbg(0, ("%s: %s..:4..pid=%d\n", __FILE__, __func__, getpid()));
	ctx_p->groups = memdup(groups, sizeof(gid_t) * ngroups);
	sp_sec_ctx_dbg(0, ("%s: %s..:5..pid=%d\n", __FILE__, __func__, getpid()));
	ctx_p->token = dup_nt_token(token);
	sp_sec_ctx_dbg(0, ("%s: %s..:6..pid=%d\n", __FILE__, __func__, getpid()));
	become_id(uid, gid);
	sp_sec_ctx_dbg(0, ("%s: %s..:7..pid=%d\n", __FILE__, __func__, getpid()));
	ctx_p->uid = uid;
	sp_sec_ctx_dbg(0, ("%s: %s..:8..pid=%d\n", __FILE__, __func__, getpid()));
	ctx_p->gid = gid;
	sp_sec_ctx_dbg(0, ("%s: %s..:9..pid=%d\n", __FILE__, __func__, getpid()));
	// Update current_user stuff 

	current_user.uid = uid;
	current_user.gid = gid;
	current_user.ngroups = ngroups;
	current_user.groups = groups;
	current_user.nt_user_token = ctx_p->token;
	sp_sec_ctx_dbg(0, ("%s: %s..:10..pid=%d\n", __FILE__, __func__, getpid()));
}
*/
/****************************************************************************
 Become root context.
****************************************************************************/

void set_root_sec_ctx(void)
{
	sp_sec_ctx_dbg(0, ("%s: %s..:\n", __FILE__, __func__));
	/* May need to worry about supplementary groups at some stage */

	set_sec_ctx(0, 0, 0, NULL, NULL);
}

/****************************************************************************
 Pop a security context from the stack.
****************************************************************************/
BOOL pop_sec_ctx(void){}
/*
BOOL pop_sec_ctx(void)
{
	struct sec_ctx *ctx_p;
	struct sec_ctx *prev_ctx_p;

	// Check for stack underflow 

	if (sec_ctx_stack_ndx == 0) {
		DEBUG(0, ("Security context stack underflow!\n"));
		smb_panic("Security context stack underflow!\n");
	}

	ctx_p = &sec_ctx_stack[sec_ctx_stack_ndx];

	// Clear previous user info 

	ctx_p->uid = (uid_t)-1;
	ctx_p->gid = (gid_t)-1;

	SAFE_FREE(ctx_p->groups);
	ctx_p->ngroups = 0;

	delete_nt_token(&ctx_p->token);

	// Pop back previous user 

	sec_ctx_stack_ndx--;

sp_sec_ctx_dbg(0, ("%s: %s.. before call gain_root():\n", __FILE__, __func__));
	gain_root();
sp_sec_ctx_dbg(0, ("%s: %s.. after call gain_root():\n", __FILE__, __func__));

	prev_ctx_p = &sec_ctx_stack[sec_ctx_stack_ndx];

#ifdef HAVE_SETGROUPS
	sys_setgroups(prev_ctx_p->ngroups, prev_ctx_p->groups);
#endif

	become_id(prev_ctx_p->uid, prev_ctx_p->gid);

	// Update current_user stuff 

	current_user.uid = prev_ctx_p->uid;
	current_user.gid = prev_ctx_p->gid;
	current_user.ngroups = prev_ctx_p->ngroups;
	current_user.groups = prev_ctx_p->groups;
	current_user.nt_user_token = prev_ctx_p->token;

	DEBUG(3, ("pop_sec_ctx (%u, %u) - sec_ctx_stack_ndx = %d\n", 
		(unsigned int)geteuid(), (unsigned int)getegid(), sec_ctx_stack_ndx));

	return True;
}
*/
/* Initialise the security context system */

void init_sec_ctx(void)
{
	int i;
	struct sec_ctx *ctx_p;
	sp_sec_ctx_dbg(0, ("%s: %s..:\n", __FILE__, __func__));
	/* Initialise security context stack */

	memset(sec_ctx_stack, 0, sizeof(struct sec_ctx) * MAX_SEC_CTX_DEPTH);

	for (i = 0; i < MAX_SEC_CTX_DEPTH; i++) {
		sec_ctx_stack[i].uid = (uid_t)-1;
		sec_ctx_stack[i].gid = (gid_t)-1;
	}

	/* Initialise first level of stack.  It is the current context */
	ctx_p = &sec_ctx_stack[0];

	ctx_p->uid = geteuid();
	ctx_p->gid = getegid();

	get_current_groups(ctx_p->gid, &ctx_p->ngroups, &ctx_p->groups);

	ctx_p->token = NULL; /* Maps to guest user. */

	/* Initialise current_user global */

	current_user.uid = ctx_p->uid;
	current_user.gid = ctx_p->gid;
	current_user.ngroups = ctx_p->ngroups;
	current_user.groups = ctx_p->groups;

	/* The conn and vuid are usually taken care of by other modules.
	   We initialise them here. */

	current_user.conn = NULL;
	current_user.vuid = UID_FIELD_INVALID;
	current_user.nt_user_token = NULL;
}
