/* 
   Unix SMB/Netbios implementation.
   Version 1.9.
   Samba utility functions
   Copyright (C) Andrew Tridgell 1992-1998
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "includes.h"

extern int DEBUGLEVEL;

static char *last_ptr=NULL;

void set_first_token(char *ptr)
{
	last_ptr = ptr;
}



/*******************************************************************
safe string copy into a known length string. maxlength does not
include the terminating zero.
********************************************************************/

char *safe_strcpy(char *dest,const char *src, size_t maxlength)
{
    size_t len;

    if (!dest) {
        DEBUG(0,("ERROR: NULL dest in safe_strcpy\n"));
        return NULL;
    }

    if (!src) {
        *dest = 0;
        return dest;
    }  

    len = strlen(src);

    if (len > maxlength) {
	    DEBUG(0,("ERROR: string overflow by %d in safe_strcpy [%.50s]\n",
		     (int)(len-maxlength), src));
	    len = maxlength;
    }
      
    memcpy(dest, src, len);
    dest[len] = 0;
    return dest;
}  

/*******************************************************************
safe string cat into a string. maxlength does not
include the terminating zero.
********************************************************************/

char *safe_strcat(char *dest, const char *src, size_t maxlength)
{
    size_t src_len, dest_len;

    if (!dest) {
        DEBUG(0,("ERROR: NULL dest in safe_strcat\n"));
        return NULL;
    }

    if (!src) {
        return dest;
    }  

    src_len = strlen(src);
    dest_len = strlen(dest);

    if (src_len + dest_len > maxlength) {
	    DEBUG(0,("ERROR: string overflow by %d in safe_strcat [%.50s]\n",
		     (int)(src_len + dest_len - maxlength), src));
	    src_len = maxlength - dest_len;
    }
      
    memcpy(&dest[dest_len], src, src_len);
    dest[dest_len + src_len] = 0;
    return dest;
}

/*******************************************************************
  convert a string to lower case
********************************************************************/
void strlower(char *s)
{
	while (*s)
    {
    	if (isupper(*s))
    		*s = tolower(*s);
        s++;
    }
}
/*
void strlower(char *s)
{
  while (*s)
  {  	
#if !defined(KANJI_WIN95_COMPATIBILITY)
  
   //For completeness we should put in equivalent code for code pages
   //949 (Korean hangul) and 950 (Big5 Traditional Chinese) here - but
   //doubt anyone wants Samba to behave differently from Win95 and WinNT
   //here. They both treat full width ascii characters as case senstive
   //filenames (ie. they don't do the work we do here).
   //JRA. 
  

    if(lp_client_code_page() == KANJI_CODEPAGE)
    {
      // Win95 treats full width ascii characters as case sensitive. 
      if (is_shift_jis (*s))
      {
        if (is_sj_upper (s[0], s[1]))
	    s[1] = sj_tolower2 (s[1]);
	else if (is_sj_ru_upper (s[0], s[1]))
	    s[1] = sj_ru_tolower2 (s[1]);
        s += 2;
      }
      else if (is_kana (*s))
      {
        s++;
      }
      else
      {
        if (isupper(*s))
          *s = tolower(*s);
        s++;
      }
    }
    else
#endif // KANJI_WIN95_COMPATIBILITY 

    {
      size_t skip = get_character_len( *s );
      if( skip != 0 )
        s += skip;
      else
      {
        if (isupper(*s))
          *s = tolower(*s);
        s++;
      }
    }
  }
}
*/

/*******************************************************************
  convert a string to upper case
********************************************************************/
void strupper(char *s){
	while (*s)
    {
    	if (islower(*s))
    		*s = toupper(*s);
        s++;
    }
}
/*
void strupper(char *s)
{
  while (*s)
  {
#if !defined(KANJI_WIN95_COMPATIBILITY)
  
  // For completeness we should put in equivalent code for code pages
  // 949 (Korean hangul) and 950 (Big5 Traditional Chinese) here - but
  // doubt anyone wants Samba to behave differently from Win95 and WinNT
  // here. They both treat full width ascii characters as case senstive
  // filenames (ie. they don't do the work we do here).
  // JRA. 
  

    if(lp_client_code_page() == KANJI_CODEPAGE)
    {
      // Win95 treats full width ascii characters as case sensitive. 
      if (is_shift_jis (*s))
      {
        if (is_sj_lower (s[0], s[1]))
	    s[1] = sj_toupper2 (s[1]);
	else if (is_sj_ru_lower (s[0], s[1]))
	    s[1] = sj_ru_toupper2 (s[1]);
        s += 2;
      }
      else if (is_kana (*s))
      {
        s++;
      }
      else
      {
        if (islower(*s))
          *s = toupper(*s);
        s++;
      }
    }
    else
#endif // KANJI_WIN95_COMPATIBILITY 
    {
      size_t skip = get_character_len( *s );
      if( skip != 0 )
        s += skip;
      else
      {
        if (islower(*s))
          *s = toupper(*s);
        s++;
      }
    }
  }
}
*/

/*******************************************************************
  compare 2 strings 
********************************************************************/
BOOL strequal(const char *s1, const char *s2)
{
  if (s1 == s2) return(True);
  if (!s1 || !s2) return(False);
  
  return(StrCaseCmp(s1,s2)==0);
}

/*******************************************************************
  case insensitive string compararison
********************************************************************/
int StrCaseCmp(const char *s, const char *t)
{
	return (strcasecmp(s,t));
}
/*
int StrCaseCmp(const char *s, const char *t)
{
   //compare until we run out of string, either t or s, or find a difference 
   //We *must* use toupper rather than tolower here due to the
   //  asynchronous upper to lower mapping.
   
#if !defined(KANJI_WIN95_COMPATIBILITY)
  
  // For completeness we should put in equivalent code for code pages
  // 949 (Korean hangul) and 950 (Big5 Traditional Chinese) here - but
  // doubt anyone wants Samba to behave differently from Win95 and WinNT
  // here. They both treat full width ascii characters as case senstive
  // filenames (ie. they don't do the work we do here).
  // JRA.
  

  if(lp_client_code_page() == KANJI_CODEPAGE)
  {
    // Win95 treats full width ascii characters as case sensitive. 
    int diff;
    for (;;)
    {
      if (!*s || !*t)
	    return toupper (*s) - toupper (*t);
      else if (is_sj_alph (*s) && is_sj_alph (*t))
      {
        diff = sj_toupper2 (*(s+1)) - sj_toupper2 (*(t+1));
        if (diff)
          return diff;
        s += 2;
        t += 2;
      }
      else if (is_sj_russian(*s) && is_sj_russian(*t))
      {
        diff = sj_ru_toupper2 (*(s+1)) - sj_ru_toupper2 (*(t+1));
	if (diff)
	      return diff;
	s += 2;
	t += 2;
      }
      else if (is_shift_jis (*s) && is_shift_jis (*t))
      {
        diff = ((int) (unsigned char) *s) - ((int) (unsigned char) *t);
        if (diff)
          return diff;
        diff = ((int) (unsigned char) *(s+1)) - ((int) (unsigned char) *(t+1));
        if (diff)
          return diff;
        s += 2;
        t += 2;
      }
      else if (is_shift_jis (*s))
        return 1;
      else if (is_shift_jis (*t))
        return -1;
      else 
      {
        diff = toupper (*s) - toupper (*t);
        if (diff)
          return diff;
        s++;
        t++;
      }
    }
  }
  else
#endif  //KANJI_WIN95_COMPATIBILITY 
  {
    while (*s && *t && toupper(*s) == toupper(*t))
    {
      s++;
      t++;
    }

    return(toupper(*s) - toupper(*t));
  }
}
*/
/*******************************************************************
  case insensitive string compararison, length limited
********************************************************************/
int StrnCaseCmp(const char *s, const char *t, size_t n)
{
	return (strncasecmp(s,t,n));
}
/*
int StrnCaseCmp(const char *s, const char *t, size_t n)
{
   //compare until we run out of string, either t or s, or chars 
   //We *must* use toupper rather than tolower here due to the
   //  asynchronous upper to lower mapping.
   
#if !defined(KANJI_WIN95_COMPATIBILITY)
  
  // For completeness we should put in equivalent code for code pages
  // 949 (Korean hangul) and 950 (Big5 Traditional Chinese) here - but
  // doubt anyone wants Samba to behave differently from Win95 and WinNT
  // here. They both treat full width ascii characters as case senstive
  // filenames (ie. they don't do the work we do here).
  // JRA. 
   

  if(lp_client_code_page() == KANJI_CODEPAGE)
  {
    // Win95 treats full width ascii characters as case sensitive. 
    int diff;
    for (;n > 0;)
    {
      if (!*s || !*t)
        return toupper (*s) - toupper (*t);
      else if (is_sj_alph (*s) && is_sj_alph (*t))
      {
        diff = sj_toupper2 (*(s+1)) - sj_toupper2 (*(t+1));
        if (diff)
          return diff;
        s += 2;
        t += 2;
        n -= 2;
      }
      else if (is_sj_russian (*s) && is_sj_russian (*t))
      {
        diff = sj_ru_toupper2 (*(s+1)) - sj_ru_toupper2 (*(t+1));
        if (diff)
          return diff;
        s += 2;
        t += 2;
        n -= 2;
      }
      else if (is_shift_jis (*s) && is_shift_jis (*t))
      {
        diff = ((int) (unsigned char) *s) - ((int) (unsigned char) *t);
        if (diff)
          return diff;
        diff = ((int) (unsigned char) *(s+1)) - ((int) (unsigned char) *(t+1));
        if (diff)
          return diff;
        s += 2;
        t += 2;
        n -= 2;
      }
      else if (is_shift_jis (*s))
        return 1;
      else if (is_shift_jis (*t))
        return -1;
      else 
      {
        diff = toupper (*s) - toupper (*t);
        if (diff)
          return diff;
        s++;
        t++;
        n--;
      }
    }
    return 0;
  }
  else
#endif // KANJI_WIN95_COMPATIBILITY 
  {
    while (n && *s && *t && toupper(*s) == toupper(*t))
    {
      s++;
      t++;
      n--;
    }

    // not run out of chars - strings are different lengths 
    if (n) 
      return(toupper(*s) - toupper(*t));

    // identical up to where we run out of chars, 
    //   and strings are same length /
    return(0);
  }
}
*/